#! /bin/sh
# \
	exec wish8.0 $0 ${1+"$@"}

#
# ECLiPSe Installation Script for Windows
#
# System:	ECLiPSe Constraint Logic Programming System
# Copyright:	(C) Imperial College London and ICL 1995-1999
# Copyright:	(C) Parc Technologies Ltd 1999
# Author/s:	Joachim Schimpf, IC-Parc
# Version:	$Id$
#

set tkecl(scriptdir) [file dirname [info script]]
set tkecl(eclipseversion) "4.2"
set tkecl(eclipsedir) $tkecl(scriptdir)
set tkecl(workdir) "C:\\"

switch $tcl_platform(platform) {
    windows {
	package require registry
	set prev [pwd]
	cd [file join $tkecl(scriptdir) lib_tcl]
	load shortcut.dll
	cd $prev
    }
    default {
	error "$tcl_platform(platform) not supported"
	exit
    }
}


wm title . "ECLiPSe Installation"
wm iconname . ECLiPSe
set tkecl(ec_image) [image create photo -format ppm -file \
	[file join $tkecl(scriptdir) lib_tcl Images eclipse_logo.ppm]]

label .logo -image $tkecl(ec_image) -relief raised
label .q1 -text "This is ECLiPSe $tkecl(eclipseversion), located in folder:"
entry .e1 -width 40 -textvariable tkecl(eclipsedir)
label .q2 -text "Where do you want your working\ndirectory when ECLiPSe starts up:"
entry .e2 -width 40 -textvariable tkecl(workdir)
frame .buttons
button .buttons.install -text "Install!" -command do_install
button .buttons.uninstall -text "Uninstall!" -command do_uninstall
button .buttons.exit -text Exit -command exit

pack .buttons.install .buttons.uninstall .buttons.exit -side left -expand 1
pack .logo -side left -padx 15 -pady 15
pack .q1 -side top -pady 5
pack .e1 -side top -padx 5
pack .q2 -side top -pady 5
pack .e2 -side top -padx 5
pack .buttons -side top -expand 1 -fill both -pady 5

update


proc do_install {} {
    global tkecl

    # make the registry entry for this version
    registry set HKEY_LOCAL_MACHINE\\SOFTWARE\\IC-Parc\\Eclipse\\$tkecl(eclipseversion) \
	ECLIPSEDIR $tkecl(eclipsedir)

    # make the registry entries for the .ecl file class
    registry set HKEY_CLASSES_ROOT\\.ecl {} \
    	ECLiPSeSource
    registry set HKEY_CLASSES_ROOT\\ECLiPSeSource {} \
    	"ECLiPSe Source"
    registry set HKEY_CLASSES_ROOT\\ECLiPSeSource\\shell\\Edit\\command {} \
    	{notepad "%1"}
    registry set HKEY_CLASSES_ROOT\\ECLiPSeSource\\shell\\Compile\\command {} \
	"$tkecl(eclipsedir)\\lib\\i386_nt\\eclipse.exe -b \"%1\""

    # create the default working directory
    file mkdir $tkecl(workdir)

    # make the start menu entries
    set startmenu [start_menu_folder]
    if {$startmenu != ""} {
	set eclmenu "$startmenu\\Programs\\ECLiPSe $tkecl(eclipseversion)"
	file mkdir $eclmenu
    }
    shortcut create $eclmenu\\DosEclipse.lnk \
    	-workingDirectory $tkecl(workdir) \
    	-objectPath $tkecl(eclipsedir)\\lib\\i386_nt\\eclipse.exe
    shortcut create $eclmenu\\TkEclipse.lnk \
    	-workingDirectory $tkecl(workdir) \
    	-objectPath $tkecl(eclipsedir)\\lib_tcl\\tkeclipse.tcl
    shortcut create $eclmenu\\Documentation.lnk \
    	-objectPath $tkecl(eclipsedir)\\doc\\index.html
    shortcut create $eclmenu\\Readme.lnk \
    	-objectPath $tkecl(eclipsedir)\\README_WIN.TXT
    shortcut create "$eclmenu\\Uninstall ECLiPSe $tkecl(eclipseversion).lnk" \
    	-workingDirectory $tkecl(eclipsedir) \
    	-objectPath $tkecl(eclipsedir)\\ecl_inst.tcl

    tk_messageBox -title "ECLiPSe Installation" \
	-message "Installation complete\nECLiPSe is now available from your Start menu" -type ok
}

proc do_uninstall {} {
    global tkecl

    # delete the entry for this version
    registry delete HKEY_LOCAL_MACHINE\\SOFTWARE\\IC-Parc\\Eclipse\\$tkecl(eclipseversion)

    # if no other versions exist, delete also the .ecl class entry
    set otherversions [registry keys HKEY_LOCAL_MACHINE\\SOFTWARE\\IC-Parc\\Eclipse]
    if {$otherversions == {}} {
	registry delete HKEY_CLASSES_ROOT\\.ecl
	registry delete HKEY_CLASSES_ROOT\\ECLiPSeSource
    }

    # delete the start menu entry
    set startmenu [start_menu_folder]
    if {$startmenu != ""} {
	file delete -force "$startmenu\\Programs\\ECLiPSe $tkecl(eclipseversion)"
    }

    tk_messageBox -title "ECLiPSe Installation" \
	-message "Uninstallation complete" -type ok
}

proc start_menu_folder {} {
    global env
    if [info exists env(WINDIR)] {
	set windir $env(WINDIR)
    } else {
	set windir $env(windir)
    }
    
    foreach f [list "$windir\\Start Menu" "$windir\\Profiles\\All Users\\Start Menu"] {
	if [file exists $f] {
	    return $f
	}
    }
    return {}
}
