% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: answer.pl,v 1.2 1999/08/07 15:49:58 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 *
 *	sccsid("%W%		%E%").
 *	sccscr("%Z%  Copyright 1990 ECRC GmbH ").
 *
 * IDENTIFICATION:	answer.pl 
 *
 * DESCRIPTION: 	Utility predicates, for user convenience.
 *
 * CONTENTS:
 *
 */


:- module(answer).

:- export print_answer/3, answer/2.

% Print the answer binding and report delayed goals.
print_answer(_, List, M) :-
    bind_vars(List, 0, Length),
    print_values(List, Length),
    delayed_goals(DL),
    (DL == [] ->
        true
    ;
        error(273, DL, M)       % report the delayed goals
    ).

% Bind each free variable to a term containing its name
% and find out the length of the longest name of a bound variable.
bind_vars([], Length, Length).
bind_vars([[Name|Value]|Rest], SoFar, Max) :-
    (var(Value) ->
        Value = '$VAR'(Name),
        L = SoFar
    ;
        % note the use of atom_length/2 as arithmetic functor
        L is max(atom_length(Name), SoFar)
    ),
    bind_vars(Rest, L, Max).

% Print the answer binding using portray/2 in printf/3.
print_values([], _).
print_values([[Name|Value]|Rest], L) :-
    (Value == '$VAR'(Name) ->
        true                                  % don't report X = X
    ;
        printf(answer_output, "%-*s = %p\n", [L, Name, Value])
    ),
    print_values(Rest, L).

% Print an input variable as its name.
portray(S, '$VAR'(Name)) :-
    atom(Name),
    write(S, Name).



answer(_, yes) :-
    ask_more.
answer(_, last_yes) :-
    write(toplevel_output, '\nyes.\n').
answer(_, no) :-
    write(toplevel_output, '\nno.\n').
answer(_, more_answers) :-
    ask_more.
answer(_, last_answer) :-
    write(toplevel_output, '\nyes.\n').
answer(_, no_answer) :-
    write(toplevel_output, '\nno.\n').



ask_more :-
    write(toplevel_output, '\tMore? '),
    flush(toplevel_output),
    tyi(toplevel_input, C),
    (C == 0'; ->
        nl(toplevel_output),
        fail
    ;
        true
    ).
