% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: cio.pl,v 1.2 1999/08/07 15:50:03 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%		%E%").
 * sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 */

/*
 * IDENTIFICATION:	cio.pl
 *
 * DESCRIPTION: 	Contains built-in predicates to handle the I/O
 *			in the C-prolog way.
 *
 *
 * CONTENTS:     
 *
 * REVISION HISTORY:
 * AUTHOR	VERSION	DATE	REASON
 * Micha Meier	2.4	14.2.90	
 */

:- module(cio).

% Make sure that the important operators are ok
:-	op(1100, xfy, ;),
	op(1050, xfy, ->).

:- global
	see/1,
	seeing/1,
	seen/0,
	skip/1,
	skip/2,
	tab/1,
	tab/2,
	tell/1,
	telling/1,
	told/0.

:- import
	is_predicate_/2,
	put0/1,
	see1/1,
	tell1/1,
	write_/2
   from sepia_kernel.

see(user_input) :- !, see(user).
see(File) :-
	expand_file(File, FileExp),
	see1(FileExp),
	(FileExp \== user ->
		set_stream(FileExp, input)
	;
		true
	).

seeing(File) :-
	current_stream(File, _, input).

seen :-
	seeing(File),
	(File \== user -> close(File); true),
	set_stream(input, stdin).

tell(user_output) :- !, tell(user).
tell(user_error) :- !, tell(error).
tell(File) :-
	expand_file(File, FileExp),
	tell1(FileExp),
	(FileExp \== user ->
		set_stream(FileExp, output)
	;
		true
	).

telling(File) :-
	current_stream(File, _, output).

told :-
	telling(File),
	(File \== user -> close(File); true),
	set_stream(output, stdout).

expand_file(File, FileExp) :-
	pathname(File, A, B),
	concat_atom([A, B], FileExp).

skip(S) :-
	skip(input, S).

skip(S, X) :-
	N is X,
	skip_to(S, N).

skip_to(S, N) :-
	(get(S, N) ->
		true
	;
		skip_to(S, N)
	).

tab(X) :-
	tab(output, X).

tab(S, X) :-
	Tab is X,
	put_spaces(S, Tab).

put_spaces(_, 0) :- !.
put_spaces(S, N) :-
	put(S, 32),
	N1 is N-1,
	put_spaces(S, N1).

:- skipped
	see/1,
	seeing/1,
	seen/0,
	skip/1,
	skip/2,
	tab/1,
	tab/2,
	tell/1,
	telling/1,
	told/0.
