% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: ech.pl,v 1.5.2.1 1999/08/20 17:50:17 ks15 Exp $
% ----------------------------------------------------------------------

%  New CHR implementation
%  Kish Shen, March - June 1998, version 1
%             March - April, 1999, some fixes
%  Partial support for multi-head (>2) rules: no multi-head propagation rules
%  Syntax changes and minor semantic changes from old CHR
%  Faster execution


%:- global []/0. % hacks round problem
%:- dynamic []/0.


%:- reset_event_handler(139).
%:- reset_event_handler(42).
%:- reset_error_handlers. %not for new debugger

:- module_interface(ech).

:- meta_attribute(ech, [
			  unify:   unify_ech/2
                          %print:   print_ech/2
   ]).

:- op(1200, fy, handler).
:- op(1000, fy, constraints).
%:- op(1200, fy, [label_with]).
:- op(1190, xfx, [==>, <=>]).
%:- op(1200, xfx, :--).
:- op(1200, xfx, ::=). % replaces @ in old syntax
:- op(1100, xfx, |).
:- op(1100, xfx, \ ).
:- op(700, xfx, flag).

:- export record_chrrule/3, record_chrprule/3, new_constraints/3, 
          ignore_handler/2, record_namedrule/3, appliedpos/2,
          suspending_priority/2, suspendindexpos/2, constraintnumpos/2,
          wrapperinpos/2, cdeletethreshold/2, varslistthreshold/2.

:- tool(record_namedrule/2, record_namedrule/3).
:- tool(record_chrrule/2, record_chrrule/3).
:- tool(record_chrprule/2, record_chrprule/3).
:- tool(new_constraints/2, new_constraints/3).


:- export suspend_constraint/3, find_partner/5, get_global_constraint/2, 
          get_constraint_list/3, kill_constraint/2, check_pairapplied/6,
          check_pairapplied_direct/6, insert_pairapplied/6, is_in_store/2,
          create_applied/2, check_samepairapplied/6, insert_samepairapplied/6, 
          check_samepairapplied_direct/6, check_samepairapplied_directmarked/8,
          check_pairapplied_directmarked/8, ffind_partner/4, in_chrstore/2, 
	  chr_get_gconstraint/2, chr_get_vconstraint/3.

:- global chr/1, chr/2, in_chrstore/1, option/2, option/3, chr_get_constraint/1, 
       chr_get_constraint/2, (constraints)/1, (constraints)/2, (handler)/1.


:- tool(suspend_constraint/3, suspend_constraint/4).
:- tool(get_global_constraint/2, get_global_constraint/3).
:- tool(get_constraint_list/3, get_constraint_list/4).
:- tool(kill_constraint/2, kill_constraint/3).
:- tool(check_pairapplied/6, check_pairapplied/7).
:- tool(insert_pairapplied/6, insert_pairapplied/7).
:- tool(check_samepairapplied/6, check_samepairapplied/7).
:- tool(insert_samepairapplied/6, insert_samepairapplied/7).
:- tool(check_pairapplied_directmarked/8, check_pairapplied_directmarked/9).
:- tool(check_samepairapplied_directmarked/8, 
        check_samepairapplied_directmarked/9).
:- tool((chr)/1, (chr)/2).
:- tool(in_chrstore/1, in_chrstore/2).
:- tool(option/2, option/3).
:- tool((constraints)/1, (constraints)/2).
:- tool(chr_get_constraint/1, chr_get_gconstraint/2).
:- tool(chr_get_constraint/2, chr_get_vconstraint/3).



% avoid magic number in code
suspending_priority(suspendpriority, 3). % priority to suspend with
realconstraintpos(constraint_in_wrapper_pos, 2). 
% arg. pos for const. in wrapper
suspendindexpos(suspendid_pos, 3). % arg. pos for Index in wrapper
% arg. pos for In arg. in wrapper
wrapperinpos(wrapper_inpos, 4).
% arg. pos for global const. no. in wrapper
constraintnumpos(constraintnum_pos, 1). 
% arg. pos for applied-list in wrapper
appliedpos(applied_pos, 5).
% threshold for general clean-up of var's constraint list
varslistthreshold(varslist_threshold, 100).
% threshold of number of kill_constraints before general cleanup
cdeletethreshold(cdelete_threshold, 15).

% macros for use when CHR syntax code is being read in

ignore_handler(_, []). 

record_chrrule(Rule, [], Module) :-
      % put in front as later we add to Processed at the front of the list
      check_if_new_or_update(Module),
      recorda('CHRcode', Rule)@Module. 

record_chrprule(Rule0, [], Module) :-
      check_if_new_or_update(Module),
      erase('CHRprule_count', count(Module,Np)),
      Np1 is Np + 1,
      (Rule0 = (Name ::= Rule1) -> 
         Rule = (Name ::= Np1-(Rule1)) ; Rule = Np1-(Rule0)
      ),
      recorda('CHRcode', Rule)@Module,
      recorda('CHRprule_count',count(Module,Np1)).

record_namedrule(NamedRule, [], Module) :-
     NamedRule = (_Name ::= Rule),
     ( Rule = (Head ==> Body) -> record_chrprule(NamedRule, [], Module)
    ;( Rule = (Head <=> Body) -> record_chrrule(NamedRule, [], Module)
    ;  writeln(error, "Syntax error, ::= is not followed by a valid rule in"),
       pretty_write(NamedRule), nl
     )).

new_constraints(constraints ConstDec, [], Module) :- % still support old syntax for now
   check_if_new_or_update(Module),
   erase('CHRconst_count', count(Module,NConst0)),
   count_and_record_constraints(ConstDec, NConst0, Count0, Module),
   recorda('CHRconst_count', count(Module,Count0)).

constraints(ConstDec, Module) :-
   check_if_new_or_update(Module),
   erase('CHRconst_count', count(Module,NConst0)),
   count_and_record_constraints(ConstDec, NConst0, Count0, Module),
   recorda('CHRconst_count', count(Module,Count0)).

handler _. % do nothing; for compatibility only



:- define_macro((handler)/2, ignore_handler/3, [clause]).
:- define_macro((==>)/2, record_chrprule/3, [clause]).
:- define_macro((<=>)/2, record_chrrule/3, [clause]).
:- define_macro((::=)/2, record_namedrule/3, [clause]).
:- define_macro((constraints)/1, new_constraints/3, [clause]).
:- define_macro(suspendpriority, suspending_priority/2, []).
:- define_macro(constraint_in_wrapper_pos, realconstraintpos/2, []).
:- define_macro(suspendid_pos, suspendindexpos/2, []).
:- define_macro(constraintnum_pos, constraintnumpos/2, []).
:- define_macro(applied_pos, appliedpos/2, []).
:- define_macro(wrapper_inpos, wrapperinpos/2, []).
:- define_macro(varslist_threshold, varslistthreshold/2, []).
:- define_macro(cdelete_threshold, cdeletethreshold/2, []).


:- begin_module(ech).


:- pragma(expand).



% global constraint count
:- current_array(constraint_number, _) -> true ; make_array(constraint_number), setval(constraint_number, 0).

:- import sepia_kernel.

:- set_flag(coroutine, on).


:- use_module(library(structures)).
:- use_module(library(lists)).
:- lib(strings).
:- lib(numbervars).


:- define_struct(
         ech(
           slists,      % indexed store for suspension lists of constraint
           count
         )
   ).

%%%%%%%%%%%%%%%%%%%%%%%

:- compile(ech2pl).



count_and_record_constraints((ConstDec1,ConstDec2), Count0, Count, Module) :-
   !,
   count_and_record_constraints(ConstDec1, Count0, Count1, Module),
   count_and_record_constraints(ConstDec2, Count1, Count, Module).
count_and_record_constraints(SingleConst, Count0, Count, Module) :-
   Count1 is Count0 + 1,
   % Count is also used as id for constraint
   recorded_list('CHRconstraints', ConList)@Module,
   (second_member(ConList, SingleConst, _) ->
       % WARNING -- need to be integrated into system errors?
       printf(user, "Warning -- constraint %w already declared\n",SingleConst),
       Count = Count0
     ; Count = Count1,
       recordz('CHRconstraints', Count-SingleConst)@Module
   ).

second_member([Nth-I1|Is], I, N) ?- 
    I == I1 -> N = Nth ; second_member(Is, I, N).

nth_member(Nth, [N-F0/A0|Constraints], F/A) :-
    Nth == N -> F = F0, A0 = A ; nth_member(Nth, Constraints, F/A).


initialise_module_for_chr(Module) :-
   % "global" (to chr in Module) chr constraint list. Note needs to 
   % be initialised to empty list when used.
   (current_array('CHRcstore',_)@Module -> 
      writeln(error, "Warning: CHRcstore counter exist in module. Erasing.."),
      erase_array('CHRcstore'/0)@Module
    ; true
   ), make_local_array('CHRcstore', global_reference)@Module,

   % count of number of chr constraints
   recorda('CHRconst_count', count(Module,0)),

   % count of number of propagation rules
   recorda('CHRprule_count', count(Module,0)),
   
   local_record('CHRcode')@Module, local_record('CHRconstraints')@Module.


redefine_cdelete_count(Error, Culprit, Module) :-
    (Culprit = make_local_array('CHRcdelete_count'(_N), integer) -> 
           true % allow silent update 
         ; error(default(Error), Culprit, Module)
    ). 



% allow cdelete_count to be redefined silently.
:- set_event_handler(42,  redefine_cdelete_count/3). 


instantiate_list([], _) :- !.
instantiate_list([E|L], E) :-
   instantiate_list(L, E).


/* split_list(+N, +List, -Front, -Back)
   splits the list List such that Front will contain the first Nth elements,
   and Back the rest
*/
split_list(0, List, [], List) :- !.
split_list(N, [E|L], [E|Front], Back) :-
   N1 is N - 1,
   split_list(N1, L, Front, Back).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* find_partner(+SuspL, -PartnerIndex, +Partner, -RestL, -Condition)
   checks the list of suspended constraint SuspL to try to find one that
   matches Partner. PartnerIndex is the internal suspension index 
   associated with the constraint. RestL is the remaining items on the
   SuspL after Partner is found. Condition returns either yes or no 
   depending on if Partner is found or not. The predicate does not simply
   fail to allow the CHR program to collect information
*/
find_partner([SIndex|Rest], SIndex, Constraint, Rest, yes) :- 
    suspension_to_goal(SIndex, Suspended, _),
    % need to extract constraint from the meta-info surrounding it
    arg(constraint_in_wrapper_pos,Suspended,Constraint).
find_partner([_|Rest], SIndex, Constraint, RestL, Cond) :-
    find_partner(Rest, SIndex, Constraint, RestL, Cond).
find_partner([], _, _, _, no).

ffind_partner([SIndex|Rest], SIndex, Constraint, Rest) :- 
    suspension_to_goal(SIndex, Suspended, _),
    % need to extract constraint from the meta-info surrounding it
    arg(constraint_in_wrapper_pos,Suspended,Constraint).
ffind_partner([_|Rest], SIndex, Constraint, RestL) :-
    ffind_partner(Rest, SIndex, Constraint, RestL).


/* get_global_constraint(+Const, -SuspL, +Module)
   get the global constraint list SuspL for constraint number Const 
*/
get_global_constraint(Const, SuspL, Module) :- 
    getval_body('CHRcstore', Store, Module),
    (Store \== 0 -> arg(Const, Store, SuspL)  ; SuspL = []).



% list is terminated by a variable. Find it
find_tail(Tail, Tail0) :-
   var(Tail), !, Tail0 = Tail.
find_tail([_|Rest], Tail) :-
   find_tail(Rest, Tail).

pretty_write(Term) :- writeln(Term).
       
count_dead([], L,L, D,D) :- !.
count_dead([S|Ss], L0, L, D0, D) :-
    (is_suspension(S) -> D1 = D0 ; D1 is D0 + 1),
    L1 is L0 + 1,
    count_dead(Ss, L1, L, D1, D).

/* get_constraint_list(?Vars, +ConstNo, -SuspL, +Module)
    returns the list of suspended constraints for constraint ConstNo in SuspL.
    The list is either those that are suspended on Vars, or the global list,
    if none of the variables in Vars remain free.
*/ 
get_constraint_list(Vars, ConstNo, SuspL, Module) :-
      nonground(Vars, Var) ->
        get_varconstraint(Var, ConstNo, SuspL, Module)

      ; get_globalconstraint(ConstNo, SuspL, Module).

get_constraint_listi(Vars, ConstNo, SuspL, Cond, Module) :-
      nonground(Vars, Var) ->
        Cond = var,
        get_varconstraint(Var, ConstNo, SuspL, Module)

      ; get_globalconstraint(ConstNo, SuspL, Module).

/*set_varconstraint(_X{ech:Attr}, ConstNo, SuspL) ?-
    setarg(ConstNo, Attr, SuspL).
*/

set_varconstraint(_X{ech:Attr}, ConstNo, SuspL) ?-
    Attr = ech with [slists: Ss],
    setarg(ConstNo, Ss, SuspL).

get_wholevarconstraint(X{ech:Attr0}, Attr, Module) ?-
    Attr = Attr0,
    get_wholevarconstraint1(X, Attr, Module).
get_wholevarconstraint(X, Attr, Module) :-
    free(X),
    new_chr_attr(X, Attr, Module).
/*NOCOUNT
get_wholevarconstraint1(_X, SuspSt, Module) :-
    var(SuspSt) -> new_cstore(SuspSt, Module); true.
*/
get_wholevarconstraint1(_X, ech with [slists: SuspSt,count:C], Module) :-
    var(SuspSt) -> new_cstore(SuspSt, Module), C = 0; true.


get_varconstraint(X{ech:Attr}, ConstNo, SuspL, Module) ?-
     get_varconstraint1(X, ConstNo, SuspL, Attr, Module).     
get_varconstraint(X, _ConstNo, SuspL, Module) :-
     free(X),
     SuspL = [],
     new_chr_attr(X, _Attr, Module).

get_varconstraint1(X, _, SuspL, Attr, Module) :-
     var(Attr), 
     SuspL = [],
     new_chr_attr(X, Attr, Module).
/*get_varconstraint1(_X, ConstNo, SuspL, Attr, _Module) :-
     arg(ConstNo, Attr, SuspL).
NOCOUNT*/
get_varconstraint1(_X, ConstNo, SuspL, Attr, Module) :-
     nonvar(Attr),
     Attr = ech with [slists: Ss,count:C],
     (var(Ss) -> new_cstore(Ss, Module), C = 0, SuspL = [] ;
          arg(ConstNo, Ss, SuspL0),
          (C == 10 -> remove_dead_suspensions(SuspL0, SuspL),
                setarg(ConstNo, Ss, SuspL), setarg(count of ech, Attr, 0)
              ; SuspL = SuspL0, C1 is C + 1,
                setarg(count of ech, Attr, C1)
          )
     ).

get_globalconstraint(ConstNo, SuspL, Module) :-
     getval_body('CHRcstore', Store, Module),
     % Store == 0 if uninitialised. 
     (Store \== 0 -> arg(ConstNo, Store, SuspL) ; SuspL = []). 


new_chr_attr(X, Attr, Module) :-  % make a new chr-variable
     Attr = ech with [slists: Ss,count: 0],
     new_cstore(Ss, Module), add_attribute(X, Attr).
/*NOCOUNT
new_chr_attr(X, Attr, Module) :-
     new_cstore(Attr, Module), add_attribute(X, Attr).
*/

new_cstore(Ss, Module) :-
     call(number_of_constraints(Size),Module), % dependent on chr prog. Need to be supplied
     functor(Ss, s, Size),
     make_empty_lists(Size, Ss).


make_empty_lists(N, Ss) :-
    N == 0 -> 
       true 
     ; N1 is N - 1,
       arg(N, Ss, List),
       (var(List) -> List = [] ; true),  
       make_empty_lists(N1, Ss).

/* insert(+Constraint, +Susp, +ConstNo)
   insert the suspension Susp into variables in the constraint Constraint.
   Optimisation should be possible which variables need to be inserted,
   and into what list
*/
insert(Constraint, Susp, ConstNo, Module) :- 
      term_variables(Constraint, VarL),
      varsuspending(VarL, Susp, ConstNo, Module).


varsuspending([], _, _, _).
/*varsuspending([V|Vs], Susp, ConstNo, Module) :-
   insert_suspension(V, Susp, constrained of suspend, suspend),
   %get_wholevarconstraint(V, Attr, Module),
   %Attr = ech with [slists: SuspSt],
   get_wholevarconstraint(V, SuspSt, Module),
   arg(ConstNo, SuspSt, SuspL0),
   cleanup(SuspL0, SuspL1),
   setarg(ConstNo, SuspSt, [Susp|SuspL1]),
   varsuspending(Vs, Susp, ConstNo, Module).
COUNT*/
varsuspending([V|Vs], Susp, ConstNo, Module) :-
   insert_suspension(V, Susp, constrained of suspend, suspend),
   get_wholevarconstraint(V, Attr, Module),
   Attr = ech with [slists: SuspSt],
   arg(ConstNo, SuspSt, SuspL0),
   cleanup(SuspL0, SuspL1),
   setarg(ConstNo, SuspSt, [Susp|SuspL1]),
   varsuspending(Vs, Susp, ConstNo, Module).


/* cleanup(+SuspLIn, -SuspLOut)
   performs some clean up (removal of dead suspensions) from SuspLIn. 
*/
cleanup(SuspL0, SuspL) :-
   SuspL0 = [Susp|SuspL1] ->
       (is_suspension(Susp) -> SuspL = SuspL0 
           ; cleanup(SuspL1, SuspL)
       )
     ; SuspL = SuspL0.



/* suspended constraints looks like:
    Name(Index,ConstraintGoal,SuspIndex,ExecInfo,Applied)
*/
suspend_constraint(ConstGoal, Constraint, ConstNo, Module) :-
    getval(constraint_number, N),
    arg(constraintnum_pos,ConstGoal,N),
    incval(constraint_number),
    arg(suspendid_pos,ConstGoal,Susp), % to be filled by suspended goal's index
    make_suspension(ConstGoal, suspendpriority, Susp, Module), % suspendpriority is a macro
    getval_body('CHRcstore',Store,Module), % need to add it to the global constraint store
    insert(Constraint, Susp, ConstNo, Module), 
    % insert into variables' constraint store 

    % insert into global constraint store
    (Store == 0 -> /* constraint store not yet initialised, initialise it */
        new_cstore(Store0, Module),
        setval_body('CHRcstore', Store0,Module),
        setarg(ConstNo, Store0, [Susp])
    ;   arg(ConstNo, Store,  SuspL), setarg(ConstNo, Store, [Susp|SuspL])
    ).

kill_constraint(ListIndex, Susp, Module) :-
    var(Susp) -> true ;
    kill_suspension(Susp),
    getval_body('CHRcdelete_count'(ListIndex), Count, Module),
    (Count == cdelete_threshold ->
       setval_body('CHRcdelete_count'(ListIndex), 0, Module), 
       cleanup_conlist(ListIndex, Module)
     ; %Count1 is Count + 1,
       %setval('CHRcdelete_count'(ListIndex), Count1)@Module
       incval_body('CHRcdelete_count'(ListIndex), Module)
    ).

cleanup_conlist(ListIndex, Module) :-
    getval_body('CHRcstore', ConStore, Module),
    arg(ListIndex, ConStore, ConList), % must be already initialised
    remove_dead_suspensions(ConList, NewConList),
    setarg(ListIndex, ConStore,  NewConList).


remove_dead_suspensions_count([], [], 0) :- !.
remove_dead_suspensions_count([Susp|ConsList0], ConsList, N) :-
   is_suspension(Susp) -> 
           ConsList = [Susp|ConsList1],
           remove_dead_suspensions_count(ConsList0, ConsList1, N)
         ; remove_dead_suspensions_count(ConsList0, ConsList, N1),
           N is N1 + 1.


remove_dead_suspensions([], []) :- !.
%remove_dead_suspensions(ConsList, 0, ConsList) :- !.
remove_dead_suspensions([Susp|ConsList0], ConsList) :-
    is_suspension(Susp) -> ConsList = [Susp|ConsList1],
          remove_dead_suspensions(ConsList0, ConsList1)
        ; %N1 is N - 1, 
          remove_dead_suspensions(ConsList0, ConsList).


    


/* check_pairapplied(+RuleNo, ?Applied, +PartnerIndex, -Tail, -Rest, -Status,
        +Module)
   check that propagation rule RuleNo, which has two heads, has not been 
   applied with partner with PartnerIndex. As the applied lists are ordered,
   checking only needs to be done to where PartnerIndex should be in the
   list. Rest is the rest of the list after this position, and Tail is
   where the position is. This is to allow for insertion of PartnerIndex
   in place if rule is fired. Status returns found or notfound.
*/
check_pairapplied(RuleNo, Applied, PartnerIndex, Tail, Rest, Status, Module) :-
     (nonvar(Applied) ->
          get_applied_list(RuleNo, Applied, AppL),
          AppL = [_|Rest0],
          check_pairapplied_direct(AppL, Rest0, PartnerIndex, Tail, Rest, Status)
        ; call(number_of_propagations(Size),Module), % Need to be supplied by chr program
          create_applied(Applied, Size),
          get_applied_list(RuleNo, Applied, Tail),
	  %Rest = [],
          Rest = Tail, /* must at least be at last element....*/
          Status = notfound
     ).

check_samepairapplied(RuleNo, Applied, PartnerIndex, Tail, Rest, Status, Module) :-
     (nonvar(Applied) ->
          get_applied_list(RuleNo, Applied, AppL),
          AppL = [_|Rest0],
          check_samepairapplied_direct(AppL, Rest0, PartnerIndex, Tail, Rest, Status)
        ; call(number_of_propagations(Size),Module), % Need to be supplied by chr program
          create_applied(Applied, Size),
          get_applied_list(RuleNo, Applied, Tail),
          /* in samepairapplied, so Rest is empty list */
          Rest = [],
          Status = notfound
     ).


check_pairapplied_direct(AppL, Rest0, PartnerIndex, Tail, Rest, Status) :-
    (Rest0 = [Index|Rest1] ->
       (Index \== PartnerIndex -> 
         (Index > PartnerIndex -> /* still need to search list */
            check_pairapplied_direct(Rest0, Rest1, PartnerIndex, Tail, Rest, Status) 
          ; Tail = AppL, Rest = Rest0, Status = notfound
         )
        ; Rest = Rest0, Tail = AppL, Status = found
       ) 
     ; /* reached end of list - not found */
       Tail = AppL, Rest = Tail, Status = notfound
       /* never return Rest as empty list (except samepairapplied) */
    ).

check_samepairapplied_direct(AppL, Rest0, PosPartnerIndex, Tail, Rest, Status) :-
    (Rest0 = [PosIndex|Rest1] ->
       (PosIndex \== PosPartnerIndex -> 
         PosIndex = Pos-Index, PosPartnerIndex = _PPos-PartnerIndex,
         (Index > PartnerIndex -> /* still need to search list */
            check_samepairapplied_direct(Rest0, Rest1, PosPartnerIndex, Tail, Rest, Status) 
          ; (Index \== PartnerIndex -> Tail = AppL, Rest = Rest0, Status = notfound
            ; % Index == PartnerIndex, so Poses must be different
              % 3 --> both pos tried
              (Pos == 3 -> Rest = Rest0, Tail = AppL, Status = found 
                ; % Pos = 1 or 2, not tried in PPos' position
                  Tail = AppL, Rest = Rest0, Status = notfound
              )
            )
         )
       ; % PosIndex == PosPartnerIndex
         Tail = AppL, Rest = Rest0, Status = found
       ) 
     ; /* reached end of list - not found */ 
       Tail = AppL, Rest = [], Status = notfound
    ).

check_pairapplied_directmarked(AppL, Rest0, PartnerIndex, Rest, ConsNumber, PGoal, RuleNo, Status, Module) :-
    (Rest0 = [Index|Rest1] ->
       (Index \== PartnerIndex -> 
         (Index > PartnerIndex -> /* still need to search list */
            check_pairapplied_directmarked(Rest0, Rest1, PartnerIndex, Rest, ConsNumber, PGoal, RuleNo, Status, Module) 
            ; % first Index < PartnerIndex, so not found and insert it
              Status = notfound,  
              setarg(2, AppL, [PartnerIndex|Rest0]),
              Rest = Rest0,
              mark_partnerapplied(ConsNumber, PGoal, RuleNo, Module)
         )
        ; %Index == PartnerIndex
          Status = found,
          Rest = Rest1
       ) 
     ; % reached end of list - not found  
       Status = notfound,
       setarg(2, AppL, [PartnerIndex]),
       mark_partnerapplied(ConsNumber, PGoal, RuleNo, Module),
       Rest = [PartnerIndex]       
    ).

mark_partnerapplied(ConsNumber, PGoal, RuleNo, Module) :-
   arg(applied_pos, PGoal, PApplied),
   (var(PApplied) -> 
      call(number_of_propagations(Size), Module),
      create_applied(PApplied, Size)
    ;true
   ),
   get_applied_list(RuleNo, PApplied, PartnerAppL),
   insert_into_partnerappliedlist(PartnerAppL, ConsNumber).

check_samepairapplied_directmarked(AppL, Rest0, PosPartnerIndex, Rest, PosConsNumber, PGoal, RuleNo, Status, Module) :-
    (Rest0 = [PosIndex|Rest1] ->
       (PosIndex \== PosPartnerIndex -> 
         PosIndex = Pos-Index, PosPartnerIndex = _PPos-PartnerIndex,
         (Index > PartnerIndex -> /* still need to search list */
            check_samepairapplied_directmarked(Rest0, Rest1, PosPartnerIndex, Rest, PosConsNumber, PGoal, RuleNo, Status, Module) 
          ; (Index \== PartnerIndex -> 
                mark_samepartnerapplied(PosConsNumber, PGoal, RuleNo, Module),
                setarg(2, AppL, [PosPartnerIndex|Rest0]), 
                Rest = Rest0, Status = notfound
            ; % Index == PartnerIndex, so Poses must be different
              % 3 --> both pos tried
              (Pos == 3 -> Rest = Rest0, Status = found 
                ; % Pos = 1 or 2, not tried in PPos' position, now both tried
                  setarg(1,PosIndex, 3), Rest = Rest0, 

                  mark_samepartnerapplied(PosConsNumber, PGoal, RuleNo, Module), 

                  Status = notfound
              )
            )
         )
       ; % PosIndex == PosPartnerIndex
         Rest = Rest0, Status = found
       ) 
     ; /* reached end of list - not found */
       mark_samepartnerapplied(PosConsNumber, PGoal, RuleNo, Module),  
       setarg(2, AppL, [PosPartnerIndex]), Rest = [], Status = notfound
    ).

mark_samepartnerapplied(PosConsNumber, PGoal, RuleNo, Module) :-
    arg(applied_pos, PGoal, PApplied),
    (var(PApplied) -> 
          call(number_of_propagations(Size), Module),
          create_applied(PApplied, Size)
         ;true
    ),
    get_applied_list(RuleNo, PApplied, PartnerAppL),
    insert_into_samepartnerappliedlist(PartnerAppL, PosConsNumber).


% like checkpairapplied_direct, except that Index known not to be in list,
% just need to find where it should be inserted.
find_pairappliedinsertpos(AppL, Rest0, Index, Tail, Rest) :-
    (Rest0 = [Index0|Rest1] ->
         (Index0 > Index -> /* still need to search list */
            check_pairapplied_direct(Rest0, Rest1, Index, Tail, Rest, _) 
          ; Tail = AppL, Rest = Rest0
         )
     ; /* reached end of list - not found */ 
       Tail = AppL, Rest = []
    ).

/*  For multiple heads not yet ready
check_multiapplied(RuleNo, Applied, Partners, Tail, Rest, Status, Module) :-
    (nonvar(Applied) ->
         get_applied_list(RuleNo, Applied, AppL),
         check_applied_direct(AppL, Partners, Tail, Rest, Status)
       ; call(number_of_propagations(Size),Module), % Need to be supplied by chr program
         create_applied(Applied, Size),
         arg(RuleNo, Applied, Tail),
         Rest = [],
         Status = notfound
     ).
%%%%%%%%%%%% write rest of code later.
*/


insert_pairapplied(PConsNumber, Tail, Rest, ConsNumber, PGoal, RuleNo, Module) :-
   setarg(2, Tail, [PConsNumber|Rest]),
   arg(applied_pos, PGoal, PApplied),
   (var(PApplied) -> 
      call(number_of_propagations(Size), Module),
      create_applied(PApplied, Size)
    ;true
   ),
   get_applied_list(RuleNo, PApplied, PartnerAppL),
   insert_into_partnerappliedlist(PartnerAppL, ConsNumber).

insert_samepairapplied(PosPConsNumber, Tail, Rest, PosConsNumber, PGoal, RuleNo, Module) :-
   insert_samepair_atpos(PosPConsNumber, Tail, Rest),
   arg(applied_pos, PGoal, PApplied),
   (var(PApplied) -> 
      call(number_of_propagations(Size), Module),
      create_applied(PApplied, Size)
    ;true
   ),
   get_applied_list(RuleNo, PApplied, PartnerAppL),
   insert_into_samepartnerappliedlist(PartnerAppL, PosConsNumber).

insert_samepair_atpos(PosPConsNumber, Tail, Rest) :-
   (Rest = [PosNum|_] ->
      PosPConsNumber = _Pos-PConsNumber,
      (PosNum = _-PConsNumber ->
         % same PConsNumber, must have tried both positions, so set Pos to 3
         setarg(1, PosNum, 3)
       ; % not same PConsNumber...
         setarg(2, Tail, [PosPConsNumber|Rest])
      )
     ; % Rest is empty list
       setarg(2, Tail, [PosPConsNumber])
   ).

insert_into_partnerappliedlist(PAppL, ConsNum) :-
   PAppL = [_|Rest0],
   find_pairappliedinsertpos(PAppL, Rest0, ConsNum, Tail, Rest),
   setarg(2, Tail, [ConsNum|Rest]).

insert_into_samepartnerappliedlist(PAppL, PosPConsNum) :-
   PAppL = [_|Rest0],
   check_samepairapplied_direct(PAppL, Rest0, PosPConsNum, Tail, Rest, _S),
   % defensive check, comment out for max. speed
   %(_S == found -> writeln('ECKKKKKK....');true),
   insert_samepair_atpos(PosPConsNum, Tail, Rest).


add_applied(PartnersNo, Tail) :-
     (nonvar(Tail) ->
        Tail = [_|Ns],
        NewTail = [PartnersNo|Ns],
        setarg(2,Tail,NewTail)
      ; Tail = [PartnersNo]
    ).

create_applied(Applied, Size) :-
   functor(Applied, a, Size).

get_applied_list(N, Applied, List) :-
    arg(N, Applied, List),
    (var(List) -> List = [1.0Inf] ; true).



% unify_ech(+Term, Attribute)
unify_ech(_, Attr) :- 
   var(Attr), !.
unify_ech(Term, _) :-
   nonvar(Term), !.
unify_ech(_Term{ech:Attr0}, Attr1) ?- !,
/* Term is var, there are chr attributes for both variables */
   arg(1, Attr0, CStore0),
   arg(1, Attr1, CStore1),
   chrstore_mergeinto(CStore0, CStore1).
unify_ech(Term, Attr) :-
/* Term does not have  chr attribute, just add it in */
   add_attribute(Term, Attr).

chrstore_mergeinto(CS0, CS1) :-
   functor(CS0, F, A),
   functor(CS1, F, A),
   % merge the lists from Attr1 into Attr0
   chrstore_merge(A, CS0, CS1).

chrstore_merge(0, _, _) :- !.
chrstore_merge(N, CS0, CS1) :-
   arg(N, CS0, List0),
   arg(N, CS1, List1),
   ordered_merge(List0, List1, List),
   setarg(N, CS0, List),
   N1 is N - 1,
   chrstore_merge(N1, CS0, CS1).

ordered_merge([], L, L) :- !.
ordered_merge(L, [], L) :- !.
ordered_merge(L0, L1, L) :-
   L0 = [SIndex0|L00],
   L1 = [SIndex1|L11],
   (suspension_to_goal(SIndex0, G0,_) ->
       (suspension_to_goal(SIndex1, G1,_) ->
           arg(constraintnum_pos, G0, CN0),
           arg(constraintnum_pos, G1, CN1),
           (CN0 =:= CN1 ->
               L = [SIndex0|L2], % identical, throw one away
               ordered_merge(L00, L11, L2)
	   ;CN0 > CN1 ->
               L = [SIndex0|L2],
               ordered_merge(L00, L1, L2)
	   ; % CN1 > CN0
               L = [SIndex1|L2],
               ordered_merge(L0, L11, L2)
           )
	 ; ordered_merge(L0, L11, L) % throw away dead suspension
       )
     ; ordered_merge(L00, L1, L) % throw away dead suspension

   ).



chr_clear :-
   recorded_list('CHRconst_count', CCountL),
   erase_all('CHRconst_count'),
   erase_all('CHRprule_count'),
   erase_all('CHRadding_code'),
   setval(constraint_number, 0),
   clean_each_module(CCountL).

clean_each_module([]) :- !.
clean_each_module([count(Module,_)|L]) :-
   abolish_record('CHRcode')@Module,
   abolish_record('CHRconstraints')@Module,
   erase_array('CHRcstore'/0)@Module,
   clean_each_module(L).


is_in_store([CIndex|L], NewCon) :-
   (suspension_to_goal(CIndex, CGoal,_) ->
       arg(constraint_in_wrapper_pos, CGoal, Con),
       (Con == NewCon -> true ; is_in_store(L, NewCon) )
     ; is_in_store(L, NewCon)
   ).


multi_append(Lists, Appended) :-
  multi_append1(Lists, [], Appended).

multi_append1([], List, List) :- !.
multi_append1([List|Ls], App1, Appended) :-
   append(List, App1, App2),
   multi_append1(Ls, App2, Appended).


chr_get_gconstraint(Constraint, Module) :-
    (nonvar(Constraint) ->
	recorded_list('CHRconstraints', ConList)@Module,
	functor(Constraint, F, A),
	second_member(ConList,  F/A, ConsNo),
	get_global_constraint(ConsNo, SuspL, Module),
	matching_constraint(SuspL, Constraint)
    ;   getval_body('CHRcstore', Store, Module),
	functor(Store, _, Size),
	get_all_sconstraints(Size, Store, Constraint)
    ).

get_all_sconstraints(0, _, _) :- !, fail.
get_all_sconstraints(N, Store, Constraint) :-
	arg(N, Store, SuspL),
	matching_constraint(SuspL, Constraint).
get_all_sconstraints(N, Store, Constraint) :-
	N1 is N - 1,
	get_all_sconstraints(N1, Store, Constraint).

matching_constraint([CIndex|_], Constraint) :-
	suspension_to_goal(CIndex, CGoal,_),
	arg(constraint_in_wrapper_pos, CGoal, Constraint).
matching_constraint([_|L], Constraint) :-
	matching_constraint(L, Constraint).


chr_get_vconstraint(_V{ech:Attr}, Constraint, Module) ?-
	nonvar(Attr),
	Attr = ech with [slists:Ss],
	nonvar(Ss),
	(nonvar(Constraint) ->
	    recorded_list('CHRconstraints', ConList)@Module,
	    functor(Constraint, F, A),
	    second_member(ConList,  F/A, ConsNo),
	    arg(ConsNo, Ss, SuspL),
	    matching_constraint(SuspL, Constraint)
	; functor(Ss, _, Size),
	  get_all_sconstraints(Size, Ss, Constraint)
        ).


in_chrstore(Constraint, Module) :-
    recorded_list('CHRconstraints', ConList)@Module,
    functor(Constraint, F, A),
    second_member(ConList,  F/A, ConsNo),
    get_global_constraint(ConsNo, SuspL, Module),
    is_in_store(SuspL, Constraint).

option(Option, State, Module) :- 
    is_valid_option(Option, OptionName) ->
       may_erase(OptionName, Module),
       update_option(State, OptionName, Module)
     ; printf(error, "'%w' is not a valid option.\n", [Option]).

update_option(on, _OptionName, _Module) ?- !.
update_option(off, OptionName, Module) ?- !,
    recorda(OptionName, Module).
update_option(State, _, _) :-
    printf(error, "'%w' is an invalid state for options.\n",[State]).

is_valid_option(check_guard_bindings, 'CHRdont_guard_bindings').
is_valid_option(already_in_store, 'CHRdont_in_store').
is_valid_option(single_symmetric_simpagation, 'CHRdont_simpa_symmetric').

may_erase(Key, Value) :-
   erase(Key, Value), !.
may_erase(_, _).

