% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: edge_finder3.pl,v 1.3 1999/08/07 14:56:37 js10 Exp $
%
% Description:		Edge-finder, cubic algorithm
%
% Author:		J.Schimpf, IC-Parc
%
% The ordering booleans: There is an array of booleans with
% one boolean for each pair of tasks. The mapping is as follows:
%
%	Pos in array		Pair of tasks
%
%	Bools[j*(j-1)//2+i+1]	i	j
% e.g.
%	Bools[1]		0	1
%	Bools[2]		0	2
%	Bools[3]		1	2
%	Bools[4]		0	3
%	Bools[5]		1	3
%	Bools[6]		2	3
%
%	Bools[idx] = 0	  <=>	task_i before task_j
%	Bools[idx] = 1	  <=>	task_i after  task_j
% ----------------------------------------------------------------------

:- module_interface(edge_finder3).
:- export disjunctive/2, disjunctive_bools/3, cumulative/4, cumulative/5.


:- begin_module(edge_finder3).
:- use_module(edge_finder_common).
:- ensure_loaded(cumulative).


disjunctive_bools(Starts, Durations, Bools) :-
	disjunctive_ef(Starts, Durations, Bools, 1).

disjunctive(Starts, Durations) :-
	disjunctive_ef(Starts, Durations, 1).

cumulative(Starts, Durations, Resources, Cap) :-
	cumulative(Starts, Durations, Resources, _Areas, Cap).

cumulative(Starts, Durations, Resources, Areas, Cap) :-
	cumulative:cumulative(Starts, Durations, Resources, Cap),
	cumulative_ef(Starts, Durations, Resources, Areas, Cap, 1).

