% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: fd_util.pl,v 1.3 1999/08/07 15:50:15 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%          %E%").
 * sccscr("%Z%  Copyright 1989, 1993 ECRC GmbH ").
 */

/*
 * FINITE DOMAINS
 *
 * IDENTIFICATION:      fd_util.pl 
 *
 * AUTHOR:		Micha Meier
 *
 * DESCRIPTION:         Various utility predicates for the finite
			domains library. They usually put together
			several calls of built-in predicates that
			occur frequently enough.

   CONTENTS:
			dvar_domain_list(Var, List)
			dvar_range(Var, Min, Max)
			dvar_size(Var, Size)

			labeling(List)
 *
 */


:- module_interface(fd_util).

:- use_module(library(fd_arith)).


:- begin_module(fd_util).

:- export
    # /3,
    dvar_domain_list/2,
    dvar_range/3,
    dvar_size/2,
    labeling/1,
    outof/2.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% 	Domain processing
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% return the list of elements in the domain of Var
%
dvar_domain_list(_{fd:(fd with domain:D)}, List) :-
    -?->
    !,
    dom_to_list(D, List).
dvar_domain_list(X, List) :-
    nonground(X),
    !,
    error(4, dvar_domain_list(X, List)).
dvar_domain_list(A, [A]).

%
% return the size of a finite term
%
dvar_size(Var, Size) :-
    dvar_domain(Var, Dom),
    dom_size(Dom, Size).

dvar_range(Var, Min, Max) :-
    dvar_domain(Var, Dom),
    dom_range(Dom, Min, Max).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% 	Composed Predicates
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

outof(_, []).
outof(X, [H|T]) :-
    X ## H,
    outof(X, T).

%
% Cardinality operator
%
#(Min, ConstrList, Max) :-
    Min #<= Max,
    bool_sum(ConstrList, Sum),
    length(ConstrList, Length),
    Sum #>= Min,
    Sum #<= Max,
    Min #>= 0,
    Max #<= Length.
    
bool_sum([], 0).
bool_sum([C|L], B+S) :-
    B isd C,
    bool_sum(L, S).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% 	Labeling
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labeling([]).
labeling([X|L]) :-
    indomain(X),
    labeling(L).
