% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: fromonto.pl,v 1.2 1999/08/07 15:50:17 js10 Exp $
% ----------------------------------------------------------------------

%
% SEPIA PROLOG LIBRARY MODULE
%
% sccsid("%W%             %E%").
% sccscr("%Z%  Copyright 1991 ECRC GmbH ").
%
% IDENTIFICATION:	fromonto.pl
%
% AUTHOR:		Joachim Schimpf
%
% CONTENTS:		Goal from_stream Stream
%			Goal onto_stream Stream
%			Goal from_string String
%			Goal onto_string String
%			Goal from_file File
%			Goal onto_file File
%
% DESCRIPTION:	Redirect input and output in a convenient way.
%		The idea is stolen from Richard O'Keefe.
%		Alternatives of the goal are discarded.


:- module_interface(fromonto).
:- export
	(from_stream)/2,
	(onto_stream)/2,
	(from_string)/2,
	(onto_string)/2,
	(from_file)/2,
	(onto_file)/2.

:- op(800, yfx, [from_string, onto_string,
		from_stream, onto_stream,
		from_file, onto_file]).

:- begin_module(fromonto).

:- tool((from_string)/2, from_string_body/3).
:- tool((from_file)/2, from_file_body/3).
:- tool((from_stream)/2, from_stream_body/3).
:- tool((onto_string)/2, onto_string_body/3).
:- tool((onto_file)/2, onto_file_body/3).
:- tool((onto_stream)/2, onto_stream_body/3).


from_stream_body(Goal, Stream, Module) :-
	get_stream(input, In),
	set_stream(input, Stream),
	(call(Goal, Module) -> Result = true ; Result = false),
	set_stream(input, In),
	Result = true.

from_file_body(Goal, File, Module) :-
	open(File, read, Stream),
	get_stream(input, In),
	set_stream(input, Stream),
	(call(Goal, Module) -> Result = true ; Result = false),
	set_stream(input, In),
	close(Stream),
	Result = true.

from_string_body(Goal, String, Module) :-
	open(String, string, Stream),
	get_stream(input, In),
	set_stream(input, Stream),
	(call(Goal, Module) -> Result = true ; Result = false),
	set_stream(input, In),
	close(Stream),
	Result = true.


onto_stream_body(Goal, Stream, Module) :-
	get_stream(output, Out),
	set_stream(output, Stream),
	(call(Goal, Module) -> Result = true ; Result = false),
	set_stream(output, Out),
	Result = true.

onto_file_body(Goal, File, Module) :-
	open(File, write, Stream),
	get_stream(output, Out),
	set_stream(output, Stream),
	(call(Goal, Module) -> Result = true ; Result = false),
	set_stream(output, Out),
	close(Stream),
	Result = true.

onto_string_body(Goal, String, Module) :-
	open("", string, Stream),
	get_stream(output, Out),
	set_stream(output, Stream),
	(call(Goal, Module) ->
	    set_stream(output, Out),
	    current_stream(String, _, Stream),
	    close(Stream)
	;
	    set_stream(output, Out),
	    close(Stream),
	    fail
	).

