% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: iso.pl,v 1.8 1999/08/07 15:50:21 js10 Exp $
% ----------------------------------------------------------------------

%
% ECLiPSe PROLOG LIBRARY MODULE
%
% $Id: iso.pl,v 1.8 1999/08/07 15:50:21 js10 Exp $
%
% IDENTIFICATION:	iso.pl
%
% AUTHOR:		Joachim Schimpf
%
% CONTENTS:		see export directive
%
% DESCRIPTION:		ISO Prolog compatibility package (incomplete)
%			It follows standard draft ISO/IEC DIS 13211-1:1995(E)
%

:- module_interface(iso).

:- set_flag(syntax_option, no_blanks),
   set_flag(syntax_option, no_array_subscripts),
   set_flag(syntax_option, limit_arg_precedence).

:- set_chtab(0'`, string_quote),
   set_chtab(0'", list_quote).

:- op(200, xfx, (**)),
   op(400, yfx, [(mod),(rem)]).

:-  op(0,   xfx, (of)),		% remove some eclipse extensions
    op(0,   xfx, (with)),
    op(0,   xfy, (do)),
    op(0,   xfx, (@)),
    op(0,   fx, (-?->)),
    define_macro((with)/2, (=)/2, [local]),
    define_macro((of)/2, (=)/2, [local]).

:- export
	(**)/3,
	assertz/1,
	at_end_of_stream/0,
	at_end_of_stream/1,
	atom_concat/3,
	atom_codes/2,
	atom_chars/2,
	catch/3,
	catch/4,
	char_conversion/2,
	close/2,
	current_char_conversion/2,
	current_input/1,
	current_output/1,
	current_prolog_flag/2,
	float/1,
	float_integer_part/2,
	float_fractional_part/2,
	floor/2,
	flush_output/0,
	flush_output/1,
	get_byte/1,
	get_byte/2,
	get_char/1,
	get_char/2,
	get_code/1,
	get_code/2,
	halt/1,
	log/2,
	multifile/1,
	multifile/2,
	number_chars/2,
	number_codes/2,
	peek_byte/1,
	peek_byte/2,
	peek_char/1,
	peek_char/2,
	peek_code/1,
	peek_code/2,
	put_byte/1,
	put_byte/2,
	put_code/1,
	put_code/2,
	read_term/2,
	read_term/3,
	rem/3,
	round/2,
	set_input/1,
	set_output/1,
	set_prolog_flag/2,
	set_stream_position/2,
	sign/2,
	stream_property/2,
	sub_atom/4,
	throw/1,
	truncate/2,
	unify_with_occurs_check/2,
	write_term/2,
	write_term/3.

:- tool(catch/3, catch/4).
:- tool(multifile/1, multifile/2).

:- use_module(scattered).	% for discontiguous/1.

:- begin_module(iso).
:- pragma(nodebug).
:- pragma(system).

:- import block/4, get_bip_error/1, set_bip_error/1 from sepia_kernel.

%-----------------------------------------------------------------------
% 7.4 Directives
%-----------------------------------------------------------------------

multifile(Preds, Module) :-
	@(dynamic(Preds), Module).

%-----------------------------------------------------------------------
% 7.8 Control constructs (ok)
%-----------------------------------------------------------------------

catch(Goal, Catcher, Recovery, Module) :-
	block(Goal, Catcher, Recovery, Module).

throw(B) :-
	exit_block(B).

%-----------------------------------------------------------------------
% 8.2 Term Unification (ok)
%-----------------------------------------------------------------------

:- set_flag(occur_check, on).
unify_with_occurs_check(X, X).			% 8.2.2
:- set_flag(occur_check, off).

%-----------------------------------------------------------------------
% 8.3 Type Testing (ok)
%-----------------------------------------------------------------------

float(X) :- real(X).				% 8.3.4

%-----------------------------------------------------------------------
% 8.9 Clause creation and destruction (ok)
%-----------------------------------------------------------------------

:- import assert_/2 from sepia_kernel.
:- tool(assertz/1, assert_/2).			% 8.9.2

% don't retract all on a subsequent dynamic/1 declaration
:- set_error_handler(64, true/0).

%-----------------------------------------------------------------------
% 8.11 Stream selection and control (incomplete)
%-----------------------------------------------------------------------

current_input(Stream) :-			% 8.11.1
	get_stream(input, Stream).

current_output(Stream) :-			% 8.11.2
	get_stream(output, Stream).

set_input(Stream) :-				% 8.11.3
	set_stream(input, Stream).

set_output(Stream) :-				% 8.11.4
	set_stream(output, Stream).

close(Stream, _) :-				% 8.11.6
	close(Stream).

flush_output :- flush(output).			% 8.11.7

flush_output(Stream) :- flush(Stream).

stream_property(Stream, Property) :-
	current_stream(Stream),
	stream_property1(Stream, Property).

stream_property1(Stream, mode(M)) :-		% 8.11.8
	get_stream_info(Stream, mode, M).
stream_property1(Stream, file_name(F)) :-
	get_stream_info(Stream, name, F).
stream_property1(Stream, position(P)) :-
	at(Stream, P).
stream_property1(_Stream, type(binary)).
stream_property1(Stream, end_of_stream(P)) :-
	(at_eof(Stream) -> P = at ; P = no).

at_end_of_stream :-
	at_eof(input).

at_end_of_stream(Stream) :-
	at_eof(Stream).

set_stream_position(Stream, P) :-		% 8.11.9
	seek(Stream, P).

get_byte(Code) :- get(Code).
get_byte(Stream, Code) :- get(Stream, Code).
get_char(Char) :- get(input, Code), char_code(Char, Code).
get_char(Stream, Char) :- get(Stream, Code), char_code(Char, Code).
get_code(Code) :- get(Code).
get_code(Stream, Code) :- get(Stream, Code).

put_byte(Code) :- put(Code).
put_byte(Stream, Code) :- put(Stream, Code).
put_code(Code) :- put(Code).
put_code(Stream, Code) :- put(Stream, Code).

peek_byte(Stream, Byte) :- error(267, peek_byte(Stream, Byte)).
peek_byte(Byte) :- peek_byte(input, Byte).
peek_char(Stream, Byte) :- error(267, peek_char(Stream, Byte)).
peek_char(Byte) :- peek_char(input, Byte).
peek_code(Stream, Byte) :- error(267, peek_code(Stream, Byte)).
peek_code(Byte) :- peek_code(input, Byte).


%-----------------------------------------------------------------------
% 8.14 Term input/output (incomplete)
%-----------------------------------------------------------------------

read_term(Stream, Term, Options) :-		% 8.14.1
	readvar(Stream, Term, Vars),
	handle_read_options(Options, Vars),
	!.
read_term(Stream, Term, Options) :-
	get_bip_error(E),
	error(E, read_term(Stream, Term, Options)).

    handle_read_options(Options, _) :- var(Options), !,
    	set_bip_error(4).
    handle_read_options([], _).
    handle_read_options([O|Os], Vars) :-
	handle_read_option(O, Vars),
	handle_read_options(Os, Vars).

    handle_read_option(Option, _) :- var(Option), !,
	set_bip_error(4).
    handle_read_option(variables(Vs), Vars) :- !,
    	vars_only(Vars, Vs).
    handle_read_option(variable_names(VNs), Vars) :- !,
    	name_eq_var(Vars, VNs).
    handle_read_option(singletons(_Ss), _) :- !,
	set_bip_error(141).
    handle_read_option(_, _) :-
	set_bip_error(6).

    vars_only([], []).
    vars_only([[_|V]|Vars], [V|Vs]) :-
	vars_only(Vars, Vs).

    name_eq_var([], []).
    name_eq_var([[N|V]|VNs], [N=V|Vs]) :-
	name_eq_var(VNs, Vs).


read_term(Term, Options) :-
	read_term(input, Term, Options).


write_term(Stream, Term, _Options) :-		% 8.14.2
	error(267, write(Stream, Term)).

write_term(Term, Options) :-
	write_term(output, Term, Options).


char_conversion(C1, C2) :-			% 8.14.5
	error(267, char_conversion(C1, C2)).

current_char_conversion(C, C).			% 8.14.6


%-----------------------------------------------------------------------
% 8.16 Constant Processing (ok)
%-----------------------------------------------------------------------

atom_concat(A, B, C) :-				% 8.16.2
	concat_atoms(A, B, C).

sub_atom(Atom, Start, Length, SubAtom) :-	% 8.16.3
	var(SubAtom),
	atom_string(Atom, String),
	substring(String, Start, Length, SubString),
	atom_string(SubAtom, SubString).
sub_atom(Atom, Start, Length, SubAtom) :-
	nonvar(SubAtom),
	atom_string(Atom, String),
	atom_string(SubAtom, SubString),
	substring(String, Start, Length, SubString).

atom_chars(Atom, Chars) :-			% 8.16.4
	var(Atom),
	concat_atom(Chars, Atom).
atom_chars(Atom, Chars) :-
	nonvar(Atom),
	atom_codes(Atom, Codes),
	chars_codes(Chars, Codes).

atom_codes(Atom, List) :-			% 8.16.5
	var(Atom),
	string_list(String, List),
	atom_string(Atom, String).
atom_codes(Atom, List) :-
	nonvar(Atom),
	atom_string(Atom, String),
	string_list(String, List).

number_chars(Number, Chars) :-			% 8.16.7
	var(Number),
	concat_atom(Chars, Atom),
	atom_string(Atom, String),
	number_string(Number, String).
number_chars(Number, Chars) :-
	nonvar(Number),
	number_string(Number, String),
	string_list(String, Codes),
	chars_codes(Chars, Codes).

number_codes(Number, Codes) :-			% 8.16.8
	var(Number),
	string_list(String, Codes),
	number_string(Number, String).
number_codes(Number, Codes) :-
	nonvar(Number),
	number_string(Number, String),
	string_list(String, Codes).

    chars_codes([], []).
    chars_codes([Char|Chars], [Code|Codes]) :-
	char_code(Char, Code),
	chars_codes(Chars, Codes).

%-----------------------------------------------------------------------
% 8.17 Implementation defined hooks (incomplete)
%-----------------------------------------------------------------------

set_prolog_flag(debug, Value) :- !,
	( Value == on -> set_flag(debugging, creep)
	; Value == off -> set_flag(debugging, nodebug)
	; error(6, set_prolog_flag(debug, Value))).
set_prolog_flag(double_quotes, Value) :- !,
	( Value == atom -> set_chtab(0'", atom_quote)
	; Value == codes -> set_chtab(0'", list_quote)
	; Value == chars -> error(141, set_prolog_flag(double_quotes, Value))
	; error(6, set_prolog_flag(double_quotes, Value))).
set_prolog_flag(Flag, Value) :-
	readonly(Flag),
	!,
	error(30, set_prolog_flag(Flag, Value)).
set_prolog_flag(Flag, Value) :-			% 8.17.1
	set_flag(Flag, Value).

    readonly(bounded).
    readonly(char_conversion).
    readonly(double_quotes).
    readonly(integer_rounding_function).
    readonly(min_integer).
    readonly(max_integer).
    readonly(max_arity).
    readonly(unknown).

current_prolog_flag(bounded, false).
current_prolog_flag(char_conversion, off).
current_prolog_flag(debug, Value) :-
	get_flag(debugging, D),
	( D = creep -> Value = on
	; D = leap -> Value = on
	; Value = off ).
current_prolog_flag(double_quotes, Value) :-
	( get_chtab(0'", atom_quote) -> Value = atom
	; get_chtab(0'", list_quote) -> Value = codes
	; Value = unknown ).
current_prolog_flag(integer_rounding_function, toward_zero).
%current_prolog_flag(min_integer, _) :- fail.
%current_prolog_flag(max_integer, _) :- fail.
current_prolog_flag(max_arity, unbounded).
current_prolog_flag(unknown, error).
current_prolog_flag(Flag, Value) :- get_flag(Flag, Value).

halt(X) :- exit(X).				% 8.17.4

%-----------------------------------------------------------------------
% 9. Evaluable functors (incomplete)
% Note: the redefinitions of floor,round,mod don't work currently, the
% arithmetic transformation in the ECLiPSe compiler always uses the
% sepia_kernel definitions for the predefined arithmetic functions!
%-----------------------------------------------------------------------

**(X,Y,Z) :- Z is eval(X)^eval(Y).
sign(X,Y) :- Y is sgn(eval(X)).
truncate(X,Y) :- Y is fix(eval(X)).
log(X,Y) :- Y is ln(eval(X)).
floor(X,Y) :- X1 is X, sepia_kernel:floor(X1,Y0), Y is fix(Y0).
round(X,Y) :- X1 is X, sepia_kernel:round(X1,Y0), Y is fix(Y0).
rem(X,Y,Z) :- Z is sepia_kernel:mod(eval(X),eval(Y)).
float_integer_part(X,Y) :- Y is float(fix(eval(X))).
float_fractional_part(X,Y) :- X1 is X, Y is X1-fix(X1).
