% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: mps.pl,v 1.6 1999/08/07 15:50:31 js10 Exp $
% ----------------------------------------------------------------------

%
%	The Message Passing System
%
%	sccsid("%W%		%E%").
%	sccscr("%Z%  Copyright 1996 Micha Meier").
%

% Available predicates:
%	mps_init(+Host)
%	mps_init(+Host, +Port)
%	mps_error(-Message)
%	mps_port_register(+Key, +Name, +Signature, +Port)
%	mps_port_lookup(+Key, +Name, -Port)
%	mps_port_deregister(+Key, +Name, +Signature)
%	mps_port_allocate(+UpcallProc, -Port)
%	mps_port_deallocate(+Port)
%	mps_send(+Port, ?Term)
%	mps_receive(+Port, ?Term)
%	mps_ping(+Host)
%	mps_ping(+Host, +Port)
%	mps_exit

:- begin_module(sepia_kernel).

:- global
	mps_port_allocate/2,
	mps_send/2,
	mps_receive/2,
	mps_init/1.

:- tool(mps_port_allocate/2, mps_port_allocate/3).
:- tool(mps_send/2, mps_send_body/3).
:- tool(mps_receive/2, mps_receive_body/3).

:- import
	bytes_to_term_/3,
	mps_str_receive/2,
	mps_str_send/2,
	term_to_bytes_/3
    from sepia_kernel.

mps_init(Host) :-
    mps_init(Host, _).

mps_send_body(Port, Term, Module) :-
    term_to_bytes_(Term, String, Module),
    mps_str_send(Port, String).

mps_receive_body(Port, Term, Module) :-
    mps_str_receive(Port, String),
    bytes_to_term_(String, Term, Module).

mps_error_handler(_, Goal) :-
    ( mps_error(Message) -> true ; Message = "Wrong usage !" ),
    printf(error, "Message passing error in %w\n%s\n%b", [Goal,Message]),
    exit_block(abort).

:- set_default_error_handler(176, mps_error_handler/2),
   reset_error_handler(176).
