:- coroutine.

:- module_interface(propia).

:- use_module(library(suspend)).
:- op(900, yfx, infers).
:- define_macro(myinfers/4, tr_propia/2, [goal,write]).

:- define_macro(myinfers/6, tr_propia/2, [goal,write]).

:- begin_module(propia).
:- pragma(nodebug).

:- Pred = ((infers)/2,
           tr_propia/2),
   export(Pred).


:- tool((infers)/2, (infers)/3).
:- import sepia_kernel.


tr_propia(myinfers(G, L, _M, _S), infers(G, L)).
tr_propia(myinfers(G, L, _G, _T, _M, _S), infers(G, L)).

:- mode infers(+, +, ++).
%  Check the syntax, and handle any errors.
infers(Constraint, Language, Module) :-
        check_errors(Constraint, Language, Module).
%  If there are no errors, 
%  and the Constraint is ground, then, ignore the propagation language,
%  and just check the constraint
infers(Constraint,_AnyLang,Module) :-
	not nonground(Constraint), !,
	once(Constraint,Module).
%  Finally, see if the user has specified a priority and wakeup condition
%  using the syntax "suspend(Constraint,Priority,Condition) infers Language"
%  If not, default Priority=3, one lower than FD constraints, and 
%  Condition is the most general wakeup condition, 'constrained'.
infers(InGoal,Language,Module) :- !,
        ( InGoal = suspend(Goal,Prior,Cond) ->
	     not propia_prior_error(Prior,(InGoal infers Language)) 
	  ;  InGoal = Goal, Prior=3, Cond= (Goal->constrained)
	),
	infers2(Goal,Language,Prior,Cond,Module).

:- mode infers2(+,+,++,+,++).
infers2(Goal,Language,Prior,Cond,Module) :- 
        most_type(Language,Type), !,
	term_variables(Goal,Vars),
        copy_term((Goal,Vars),(CopyGoal,CopyVars)),
	init_domains(CopyGoal, most(CopyVars), GlobVar, Type, Module),
        subcall(extend(CopyGoal,most(CopyVars),most(Vars),most(NewVars),GlobVar,Type,Module),_),
        suspend(myinfers(Goal, most(Vars), GlobVar, Type, Module,Susp),Prior,Cond,Susp),
        arcConsNewDoms(most(Vars),most(NewVars), Type, Susp).


% Unfortunately the Priority and waking conditions for ac
% are those of the element (fd) constraint, which can't be overridden
infers2(Goal,ac,_Prior,_Cond,Module) :- !,
	term_variables(Goal,Vars),
	ac(Goal,Vars,Module).
% For languages 'consistent' and 'unique'
infers2(Goal,Language,Prior,Cond,Module) :-
	suspend(myinfers(Goal,Language,Module,Susp), Prior, Cond, Susp),
        myinfers(Goal,Language,Module,Susp).



%%%%%%%% Processing language most, fd, range         %%%%%%%%%%%
%%%%%%%%                                             %%%%%%%%%%%

most_type(most,Type) :-
        (current_module(fd) -> Type=[fd|More] ; Type=More),
        (current_module(range) -> More=[range] ; More=[]).
most_type(fd,[fd]).
most_type(range,[range]).

init_domains(Goal,Most,GlobVar,Type,Module) :-
        gensym(propiaVar,GlobVar),
        not not (call(Goal,Module),
                 newmsg(Most,Most,Type,FirstSol),
                 setval(GlobVar,FirstSol)).

:- demon myinfers/6.
:- mode myinfers(+, +, ++, ++, ++, +).
myinfers(Goal,Most,_GlobVar,_Type,Module,Susp) :-
        not nonground(Most), !,
        once(Goal,Module),
        kill_suspension(Susp).
myinfers(Goal,Most,GlobVar,Type,Module,Susp) :- 
        copy_term((Goal,Most),(CopyGoal,CopyMost)),
        not not (call(Goal,Module),
                 setval(GlobVar,Most)),
        subcall(extend(CopyGoal,CopyMost,Most,NewMost,GlobVar,Type,Module),_),
        arcConsNewDoms(Most,NewMost,Type,Susp).


extend(Goal,Most,OrigMost,NewMost,GlobVar,Type,Module) :-
        mysusp(Most,GlobVar),
        call(Goal,Module),
        getval(GlobVar,OldMost),
        newmsg(Most,OldMost,Type,NewMost),
        setval(GlobVar,NewMost),
        instance(OrigMost,NewMost), !.
extend(_,_,_,NewMost,GlobVar,_Type,_Module) :-
        getval(GlobVar,NewMost).


mysusp(most(Term),GlobVar) :- !,
    suspend(cons_diff(most(Term),GlobVar,constrained),2,most(Term)->constrained).


cons_diff(Most,GlobVar,Trigger) :-
        getval(GlobVar,OldMost),
        checkNotInst(Most,OldMost),
        ( nonground(Most) -> 
                  suspend(cons_diff(Most,GlobVar,Trigger),2,Most->Trigger)
         ;        true ).


checkNotInst(Most,OldMost) :-
    not instance(Most,OldMost).


newmsg(most(T1),most(T2),Type,most(T3)) :-
        msgList(T1,T2,T3,Type,[],_).

msgList([],[],[],_,Map,Map) :- !.
msgList([V1|V1T],[V2|V2T],[MSGV|MSGT],Type,Map,OutMap) :-
        msg(V1,V2,MSGV,Type,Map,NewMap),
        msgList(V1T,V2T,MSGT,Type,NewMap,OutMap).

msg(T1,T2,T3,Type,Map,OutMap) :- 
        nonvar(T1), nonvar(T2),
        T1=..[F|Args1], T2=..[F|Args2], !,
        msgList(Args1,Args2,Args3,Type,Map,OutMap),
        T3=..[F|Args3].
msg(T1,T2,T3,_,Map,Map) :-
        lookUpMap2(Map,T1,T2,T3), !.
msg(V1,V2,MSGV,Type,Map,[map(V1,V2,MSGV)|Map]) :-
        (ranges(Type,V1,V2) -> call(range_msg(V1,V2,MSGV),range) ; true),
        (fds(Type,V1,V2)    -> call(dvar_msg(V1,V2,MSGV),fd) ; true).


lookUpMap2([map(V1,V2,V3)|_],V11,V22,V33) :-
        V11==V1, V22==V2, !, V33=V3.
lookUpMap2([_|Map],V1,V2,V3) :-
        lookUpMap2(Map,V1,V2,V3).


ranges(Type,V1,V2) :-
        memberchk(range,Type),
        testrange(V1),
        testrange(V2).

testrange(Var) :-
        var(Var), !,
        call(is_range(Var),range).
testrange(Number) :-
        integer(Number), !.
testrange(Number) :-
        real(Number).

fds(Type,V1,V2) :-
        memberchk(fd,Type),
        groundOrDomain(V1),
        groundOrDomain(V2).

groundOrDomain(Var) :-
        var(Var), !, call(is_domain(Var),fd).
groundOrDomain(Val) :-
        not nonground(Val).


arcConsNewDoms(Vars,OldVars,Type,Susp) :-
    ( arcConsAnswer(OldVars,Type)
     ->
      kill_suspension(Susp)
     ;
      true
    ),
    (variant(Vars,OldVars) -> true 
     ; Vars=OldVars ).

arcConsAnswer(most([]),_Type) :- !.
arcConsAnswer(most([Var]),Type) :-
          var(Var),
          memberchk(fd,Type).




%%%%%%%% Processing languages consistent and unique  %%%%%%%%%%%
%%%%%%%%                                             %%%%%%%%%%%


:- demon myinfers/4.
:- mode myinfers(+, ++, ++, +).

myinfers(Goal,_AnyLang,Module,Susp) :-
	not nonground(Goal), !,
	kill_suspension(Susp),
	once(Goal,Module).

myinfers(Goal,consistent,Module,_Susp) :- !,
	copy_term(Goal,CopyGoal),
	subcall(once(CopyGoal,Module),_).

myinfers(Goal,unique,Module,Susp) :- !,
	copy_term(Goal,CopyGoal),
	subcall(once(CopyGoal,Module),Del), 
	switchUnique(Goal,CopyGoal,Del,Module,Susp).


switchUnique(Goal,CopyGoal,Del,Module,Susp) :-
        anotherAns(Goal,CopyGoal,Del,Module) -> true
	  ;         Goal=CopyGoal,
	            kill_suspension(Susp).

anotherAns(Goal,CopyGoal,Del,Module) :-
	copy_term(Goal,Copy2),
        subcall(call(Copy2,Module),Del2),
        not (instance(Copy2,CopyGoal), instance(Del2,Del)).

%%%%%%% Processing language ac              %%%%%%%%%%%%%%
%%%%%%%                                     %%%%%%%%%%%%%%


ac(Goal, Args,Module) :-
     copy_term((Args,Goal),(Vars,VarGoal)),
     findall(Vars,(call(VarGoal,Module),check_ground(VarGoal)),Matrix),
     transpose(Matrix,Matrix1),
     (foreach(Arg,Args), % foreach(Var,Vars), 
      foreach(List,Matrix1), param(I) do
        myelement(I,List,Arg) ).

myelement(I,List,Arg) :-
	intlist(List,1), !, I=Arg.
myelement(I,List,Arg) :-
	call(element(I,List,Arg),fd).

intlist([],_).
intlist([N|T],N) :-
	M is N+1,
	intlist(T,M).

transpose([], []).
transpose(LoL, Cols) :-
        heads_and_tails(LoL, Col, LoL1),
        ( Col == [] ->
            Cols = []
        ;
            Cols = [Col|Cols0],
            transpose(LoL1, Cols0)
        ).

heads_and_tails([], [], []).
heads_and_tails([L|Ls], Hs, Ts) :-
        ( L == [] ->
            Hs = Hs0, Ts = Ts0
        ;
            L = [H|T], Hs = [H|Hs0], Ts = [T|Ts0]
        ),
        heads_and_tails(Ls, Hs0, Ts0).



%%%%%%%%%%   Propia Syntax Check    %%%%%%%%%%%%
%%%%%%%%%%                          %%%%%%%%%%%%

%  ERRORS :  Where are their messages defined?  
%  4, 6

check_errors(Goal, Language, Module) :-
        ( var(Goal) ; var(Language) ; var(Module) ),
        !,
        error(4, infers(Goal, Language, Module)).
check_errors(Goal, Language, Module) :-
        \+ is_a_language(Language),
        error(6, infers(Goal, Language, Module)).

check_ground(Term) :-
        nonground(Term) ->
              write('The answer '), write(Term), write(' should be ground'),
              nl, abort.
check_ground(_).


propia_prior_error(Priority,Goal) :-
        (not integer(Priority) ; Priority < 4), !,
        error(4, Goal).

is_a_language(most).
is_a_language(fd) :- 
     (current_module(fd) -> true 
      ;  write("'infers fd' can't be used without loading lib(fd)"),
         abort
     ).
is_a_language(range) :-
     (current_module(range) -> true 
      ;  write("'infers range' can't be used without loading lib(range)"),
         abort
     ).
is_a_language(consistent).
is_a_language(unique).
is_a_language(ac) :-
     (current_module(fd) -> true 
      ;  write("'infers ac' can't be used without loading lib(fd)")
     ).



%%%%%%%%%  Utilities              %%%%%%%%%%%%
%%%%%%%%%                         %%%%%%%%%%%%


idmemb(X,[H|_]) :- X==H, !.
idmemb(X,[_|Rest]) :- idmemb(X,Rest).
    
idremove(X,[H|T],T) :-
	X==H, !.
idremove(X,[H|T],[H|Rest]) :-
	idremove(X,T,Rest).

:- setval(gensym,0).

gensym(Root,Syn) :-
        getval(gensym,N), 
        incval(gensym),
        integer_atom(N,NAtom),
        concat_atoms(Root,NAtom,Syn).

once(Goal,Module) :-
        call(Goal, Module), !.
