%
% ECLiPSe version of the Prolog Tk interface
%
%	Predicates to invoke the Tcl interpreter from Prolog.
%
% Author: Micha Meier
% Date:   August 95
%
%
%
% sccsid("@(#)tcl.pl	1.7          96/02/09").
% sccscr("@(#)  Copyright 1996 Micha Meier ").
%

:- coroutine.

expand_tcl(tcl(Script), Exp) :-
    nonvar(Script),
    % Take care of possible unexpected parsing when called as tcl (A, B)
    (Script = (S,VS) ->
	(Script = (S, V, R) ->
	    expand_tcl(tcl(S, V, R), Exp)
	;
	    expand_tcl(tcl(S, VS), Exp)
	)
    ;
	Exp = tcl_eval_string(Script, _)
    ).
expand_tcl(tcl(Script, Vars), Exp) :-
    nonvar(Script),
    (string(Script) ->
    	S = Script
    ;
    	atom_string(Script, S)
    ),
    merge_vars(0, S, Vars, List),
    !,
    Exp = (concat_string(List, String), tcl_eval_string(String, _)).
expand_tcl(tcl(Script, Vars, Result), Exp) :-
    nonvar(Script),
    (string(Script) ->
    	S = Script
    ;
    	atom_string(Script, S)
    ),
    merge_vars(0, S, Vars, List),
    !,
    Exp = (concat_string(List, String), tcl_eval_string(String, Result)).

merge_vars(Pos, Script, Vars, [String, Var|L]) :-
    substring(Script, Next, 2, "##"),
    Next > Pos,
    !,
    (compound(Vars) ->
    	Vars = [Var|Rest]
    ;
    	Var = Vars,
    	Rest = []
    ),
    Pos1 is Pos + 1,
    substring_pos(Script, Pos1, Next, String),
    Pos2 is Next + 1,
    merge_vars(Pos2, Script, Rest, L).
merge_vars(Pos, Script, [], [String]) :-
    Len is string_length(Script) - Pos,
    Pos1 is Pos + 1,
    substring(Script, Pos1, Len, String).

% substring from a given position to (excluding) another position
substring_pos(String, Pos1, Pos2, SubString) :-
    Length is Pos2 - Pos1,
    substring(String, Pos1, Length, SubString).

% Convert Prolog lists to Tcl lists
tcl_string(X, String) :-
    nonvar(X),
    X = [H|T],
    !,
    tcl_element(H, El),
    tcl_list_elements(T, List),
    concat_string([' {', El|List], String).
tcl_string(A, String) :-
    tcl_element(A, String).

tcl_element(Var, S) :-
    var(Var),
    !,
    error(4, tcl_string(Var, S)).
tcl_element([], "{}") :- !.
tcl_element([H|T], String) :-
    !,
    tcl_string([H|T], String).
tcl_element(H, String) :-
    rational(H),
    !,
    F is float(H),
    number_string(F, String).
tcl_element(N, String) :-
    bignum(N),
    !,
    F is float(N),
    number_string(F, String).
tcl_element(N, String) :-
    number(N),
    !,
    number_string(N, String).
tcl_element(A, String) :-
    atom(A),
    !,
    atom_string(A, S),
    tcl_element(S, String).
tcl_element(S, String) :-
    string(S),
    !,
    (substring(S, " ", _) ->
	concat_string(['{', S, '}'], String)
    ;
	String = S
    ).
tcl_element(Struct, Tcl) :-
    term_string(Struct, S),
    tcl_element(S, Tcl).

tcl_list_elements(X, S) :-
    var(X),
    !,
    error(4, tcl_string(X, S)).
tcl_list_elements([], ['}']) :- !.
tcl_list_elements([H|T], [' ', String|L]) :-
    !,
    tcl_element(H, String),
    tcl_list_elements(T, L).
tcl_list_elements(S, _) :-
    error(5, tcl_string([_|S], _)).

%
% Definition for non-expanded code and for metacalls.
%

:- pragma(silent_debug).

tcl(Args) :-
    -?->
    Args = (S, VS),
    !,
    (VS = (V, R) ->
	tcl(S, V, R)
    ;
	tcl(S, VS)
    ).
tcl(Script) :-
    tcl_eval(Script).

tcl(Script, Vars) :-
    expand_tcl(tcl(Script, Vars), Exp),
    call(Exp).

tcl(Script, Vars, Result) :-
    expand_tcl(tcl(Script, Vars, Result), Exp),
    call(Exp).

