%
% ECLiPSe version of the Prolog Tk interface
%
% Author: Micha Meier
% Date:   September 93
%

%
% sccsid("@(#)etk.pl	1.25          96/05/14").
% sccscr("@(#)  Copyright 1995 ECRC GmbH ").
%

:- module_interface(tk).

:- export
	tcl/1,
	tcl/2,
	tcl/3,
	tcl_call/2,
	tcl_command_complete/1,
	tcl_cut_fail/1,
	tcl_eval/1,
	tcl_eval/2,
	tcl_eval_string/2,
	tcl_interp/1,
	tcl_source/0,
	tcl_string/2,
	tclsh/0,
	tk_demo/0,
	tk_file/2,
	tk/1,
	tk_debug/0,
%	tk_debug_events/0,
	tk_do_one_event/1,
	tk_do_one_event/2,
	tk_file_handler/2,
	tk_main_loop/0,
	tk_next_event/1,
	tk_next_event/2,
	tk_num_main_windows/1,
	tk_get_event/1,
	tk_get_event/2,
	tk_load_libs/2,
	tk_test/0,
	tk_wait_input/1,
	wish/0.

% macros
:- export
	expand_tcl/2.
:- op(200, fx, tcl).

:- begin_module(tk).

:- define_macro(tcl/1, expand_tcl/2, [goal]).
:- define_macro(tcl/2, expand_tcl/2, [goal]).
:- define_macro(tcl/3, expand_tcl/2, [goal]).

:- import
	bignum/1,
	compile_term_/2,
	sepia_toplevel_prompt/2,		% default 153 handler
	symbol_address/2
    from sepia_kernel.

:- make_local_array(saved_input).

:- external(tcl_init/2).
:- external(tk_clear_options/0).
:- external(tk_option/2).
:- external(tk_do_one_event/2).
:- external(tcl_eval_string/2).
:- external(tcl_command_complete/1).
:- external(tcl_cut_fail/1).
:- external(tk_num_main_windows/1).
:- external(tk_fd_handler/3).
:- external(tcl_interp/1).

% We assume that files may contain relative pathnames and the loading
% may be done from another directory. Therefore, the pathnames of the
% ProTcl files have to be absolute.

absolute_objects(ObjectString, Cwd, Arch, Objects, OL) :-
    open(ObjectString, string, S),
    get_chtab(0'., OldT),
    set_chtab(0'., lower_case),
    add_path(S, Cwd, Arch, Objects, OL),
    set_chtab(0'., OldT).

add_path(S, Cwd, Arch, Objects, OL) :-
    read_token(S, File, Type),
    (Type = end_of_file ->
	close(S),
	Objects = OL
    ;
	concat_string([Cwd, "../", Arch, "/", File, " "], AbsFile),
	Objects = [AbsFile|NewO],
	add_path(S, Cwd, Arch, NewO, OL)
    ).

:-
    getcwd(Cwd),
    get_flag(hostarch, Arch),
    get_flag(object_suffix, O),
    concat_string(["../", Arch, "/dirs"], Dirs),
    get_flag(installation_directory, Inst),
    concat_atom([Inst, "/lib_graphic/tk"], ProTcl),
    compile(Dirs),
    set_error_handler(211, fail/0),
    xlibsw(XLib),
    (symbol_address(p_tcl_init, _) ->
	% it is statically linked
	LOAD = (Files = "" -> true; concat_string([Files, " ", XLib], Load), load(Load))
    ;
    O == "so" ->
	concat_string([Cwd, "../", Arch, "/tk.so"], Load),
	LOAD = (Files = "" -> load(Load); load(Files))
    ;
    O == "dll" ->
	concat_string([Cwd, "../", Arch, "/tk.", O], Load),
	LOAD = (Files = "" -> load(Load); load(Files))
    ;
%    O == "o" ->
	tcl_library(TclLib),
	tk_library(TkLib),
	tcl_version(TclVers),
	tk_version(TkVers),
	load_libs(Libs),
	protcl_objects(ObjectAtom),
	atom_string(ObjectAtom, ObjectString),
	absolute_objects(ObjectString, Cwd, Arch, Objects, OL),
	OL = [Files,
	    " -L", Cwd, "../", Arch,
	    " -L", TclLib,
	    " -L", TkLib,
	    " -linit -ltk", TkVers,
	    " -ltcl", TclVers,
	    " ", XLib, " ", Libs, " -lm"],
	LOAD = (concat_string(Objects, Load), load(Load))
    ),
    reset_error_handler(211),
    compile_term([(
tk_load_libs(Files, Module) :-
	LOAD,
	external(tcl_init/2, p_tcl_init),
	external(tk_clear_options/0, p_tk_clear_options),
	external(tk_option/2, p_tk_option),
	external(tk_do_one_event/2, p_tk_do_one_event),
	external(tcl_eval_string/2, p_tcl_eval_string),
	external(tcl_command_complete/1, p_tcl_command_complete),
	external(tcl_cut_fail/1, p_tcl_cut_fail),
	external(tcl_interp/1, p_tcl_interp),
	external(tk_num_main_windows/1, p_tk_num_main_windows),
	external(tk_fd_handler/3, p_tk_fd_handler)
    ),
protcl_source(ProTcl)
    ]).
	
:-
	(current_module(tkext) ->
	    true			% loading will be done from tkext
	;
	    tk_load_libs("", '')
	).

:- compile(tk_common).
:- compile(tcl).

tcl_eval(S, R) :-
    (atom(S); string(S)),
    !,
    tcl_eval_string(S, R).
tcl_eval(List, R) :-
    concat_string(List, S),
    tcl_eval_string(S, R).

tk(Opts) :-
    tk_init('', Opts).

% For compatibility with other systems
tk_do_one_event(Mask) :-
    tk_do_one_event(Mask, L),
    (L = [] ->
	true
    ;
	error(333, L)
    ).

tk_file_handler(Stream, What) :-
    get_stream_info(Stream, fd, Fd),
    get_stream(Stream, NS),
    tk_fd_handler(Fd, NS, What).

% Wait for input from a given stream and process Tk events in the meantime
tk_wait_input(Stream) :-
    (tk_num_main_windows(0) ->
	true					% no events if no windows
    ;
	get_stream_info(Stream, fd, Fd),
	get_stream(Stream, NS),
	tk_fd_handler(Fd, NS, on),
	tk_loop(NS),
	tk_fd_handler(Fd, NS, off)
    ).

tk_loop(Stream) :-
    tk_do_one_event(0, L),
    (tk_num_main_windows(0) ->
	error(333, ["exit"]),			% we are done
	reset_error_handler(153)
    ;
    atom(L) ->				% no Prolog event
	tk_loop(Stream)
    ;
    L = ["file", Stream] ->
	true					% there is something to read
    ;
	error(333, L),				% invoke the handler
	tk_loop(Stream)
    ).


% The new handler for the prompt event - until there is data to read,
%  we serve Tk events. Unfortunately this does not help when we block
%  in a read which is not issued from the top-level loop.
tk_prompt(_, M) :-
    sepia_toplevel_prompt(_, M),
    tk_wait_input(toplevel_input).

%
% Wait for the next Prolog event to occur, serve all other events
%
tk_next_event(List) :-
    tk_next_event(16'1e, List).

tk_next_event(Mask, List) :-
    M is Mask /\ 16'fe,
    (tk_do_one_event(M, L) ->
    	true
    ;
	% we are idle anyway, good time to update the screen
	tcl_eval_string("update idletasks", _),
	L = []
    ),
    (tk_num_main_windows(0) ->
	List = ["exit"]
    ;
    L = [] ->					% no Prolog event
	tk_next_event(Mask, List)
    ;
	List = L
    ).

%
% Process all events currently present in the queue, return the first Prolog
% one if available, else fail
%
tk_get_event(List) :-
    tk_get_event(16'1f, List).

tk_get_event(Mask, List) :-
    M is Mask \/ 1,				% don't wait
    tk_do_one_event(M, L),		% fails if no events
    (tk_num_main_windows(0) ->
	List = ["exit"]
    ;
    L = [] ->				% no Prolog event
	tk_get_event(Mask, List)
    ;
	List = L
    ).

% A wish-like facility to allow typing Tcl commands directly
wish :-
    wish([]).

tclsh :-
    wish([nodisplay]).

wish(Opts) :-
    (tcl_eval('') ->		% see if we have an interpreter
	true
    ;
	tk(Opts)
    ),
    (tk_num_main_windows(0) ->
	true
    ;
	set_error_handler(333, handle_exit/2)
    ),
    tcl_eval('set tcl_interactive 1'),
    get_prompt(toplevel_input, P, S),
    set_prompt(toplevel_input, '', S),
    block(wish_line(S, ""), Tag, rest(Tag, P, S)),
    set_prompt(toplevel_input, P, S),
    tk_file_handler(toplevel_input, off).

wish_line(S, Partial) :-
    (Partial = "" ->
	printf(S, "%% %b", [])
    ;
	true				% no prompt if continuation
    ),
    tk_wait_input(toplevel_input),
    read_string(toplevel_input, "\n", _, String),
    !,
    (String = "" ->
	NewPartial = Partial
    ;
	concat_string([Partial, "\n", String], NewCmd),
	(tcl_command_complete(NewCmd) ->
	    NewPartial = "",
	    (tcl_eval(NewCmd, Res) ->
		printf(toplevel_output, "%Dw\n%b", [Res])
	    ;
		true
	    )
	;
	    NewPartial = NewCmd
	)
    ),
    (tcl_eval('') ->		% do we still have an interpreter?
	wish_line(S, NewPartial)
    ;
	true
    ).
wish_line(_, _) :-
    tcl exit.

rest(Tag, P, S) :-
    set_prompt(toplevel_input, P, S),
    tk_file_handler(toplevel_input, off),
    tcl_eval(exit),
    exit_block(Tag).

tcl_call(String, Res) :-
    open(String, string, s),
    readvar(s, Goal, Vars),       % parse and return variables
    close(s),
    (Goal = Module:G ->
	call(G, Module)
    ;
	call(Goal, eclipse)
    ),
    return_vars(Vars, Res).

% convert [N|V] to [N, V] because pairs do not exist in Tcl
return_vars([], []).
    return_vars([[Name|Value]|L], [[Name, Value]|LR]) :-
    return_vars(L, LR).

:- tool(tcl_source/0, tcl_source/1).

tcl_source(Module) :-
    getcwd(Cwd0),
    os_file_name(Cwd0, Cwd1),		% this mess is for Windows...
    split_string(Cwd1, "\\", "\\", L),
    join_string(L, "/", Cwd2),
    concat_string([Cwd2, "/"], Cwd3),
    concat_string(['source ', Cwd3, '##'], Cwd),
    compile_term_(tcl_source(File) :- tcl(Cwd, [File]), Module).

tk_test :-
    tk_source(TkDir),
    concat_string([TkDir, "/unix/tkTest.o ", TkDir, "/unix/tkSquare.o"], Load),
    load(Load),
    tk([]),
    tcl_interp(Interp),
    call_c('Tktest_Init'(Interp), 0),
    cd(TkDir),
    cd(tests),
    read_directory(., '*.test', _, Files),
    execute(Files).

execute([]).
execute([F|L]) :-
    bad_file(F),
    !,
    execute(L).
execute([F|L]) :-
    tcl('source ##', [F]),
    execute(L).

% open |$tktest does not work
bad_file("focus.test").
bad_file("clipboard.test").
bad_file("select.test").
bad_file("send.test").

:- pragma(nodebug).

tk_debug :-
    set_error_handler(252, pending_events/0),
    get_stream(debug_input, S),
    setval(saved_input, S).

pending_events :-
    tcl_eval_string(update, _) -> true; true.

/*
 * This works, but does not really give much more than 'b' and ^D

:- debug_macro(0'], "@tk_debug_events").

tk_debug_events :-
    printf(debug_output, "handling Tk events %b", []),
    getval(saved_input, S),
    tk_wait_input(S),
    get_prompt(S, P, OutS),
    set_prompt(S, P, null),
    read_string(S, "\n", _, _),		% to ignore user input
    set_prompt(S, P, OutS).
*/

handle_exit(N, ["exit"]) :-
    !,
    (tcl exit -> true; true),
    reset_error_handler(N).
handle_exit(N, E) :-
    error(default(N), E).

% Initialize the saved debug input in case that the ']' macro is used
% without tk_debug
:- get_stream(debug_input, S), setval(saved_input, S).

:- untraceable
    tk_debug/0.
