%
% ProTcl 1.1
%
%	Definitions common to all Prologs
%
% Author: Micha Meier
%
%
% sccsid("@(#)tk_common.pl	1.15          96/05/14").
% sccscr("@(#)  Copyright 1993 ECRC GmbH ").
%

tk_demo :-
    tk_init('$tk_library/demos/widget', []).

tk_init(File, Opts) :-
    tk_clear_options,
    tk_options(Opts, OptString, OptFile),
    (File = '', nonvar(OptFile) ->
	InFile = OptFile
    ;
	InFile = File
    ),
    tcl_init(InFile, OptString),
    protcl_source(ProTcl),
    concat_atoms('global protcl_library; set protcl_library ', ProTcl, SP),
    tcl_eval(SP),
    tcl_eval('set argc [llength $argv]'),
    add_new_commands,
    (InFile = '' ->
	( get_flag(hostarch, "i386_nt") ->
	    true	% tk_fd_handler does not work
	;
	    set_error_handler(153, tk_prompt/2)
	)
    ;
	concat_atoms('source ', InFile, InitFile),
	tcl_eval(InitFile),
	tk_main_loop
    ).

tk_options(Opts, OptString, OptFile) :-
    tk_options(Opts, OptList, OptC, Argv, OptFile),
    (var(Argv) ->
	OptC = []
    ;
	OptC = ['-- ', Argv]
    ),
    concat_string(OptList, OptString).

tk_options([], L, L, _, _).
tk_options([Opt|List], [-, Name, ' ', Val, ' '|L], LC, Argv, File) :-
    Opt =.. [Name|Args],
    (Args = [Val|_] ->
	true
    ;
	Val = ''
    ),
    (Name = file, var(File) ->
	File = Val
    ;
	true
    ),
    (Name = argv ->
	Argv = Val
    ;
	true
    ),
    tk_option(Name, Val),
    tk_options(List, L, LC, Argv, File).

tk_file(File, Options) :-
    tk([file(File)|Options]).

update :-
    tcl_eval('if {[info commands update] == "update"} update').

tk_main_loop :-
    tk_do_one_event(0),
    tk_num_main_windows(X), X > 0 ->
	tk_main_loop
    ;
	true.

tcl_eval(S) :-
    tcl_eval(S, _).

add_new_commands :-
    tcl_eval_string(
	"proc call_prolog {goal} {"
	"    global var;"
	"    set res [prolog \"tcl_call $goal X\" tk];"
	"    if {$res != \"success\"} {return $res};"
	"    set list $var(X);"
	"    unset var(X);"
	"    foreach pair $list {"
	"	set var([lindex $pair 0]) [lindex $pair 1];"
	"    };"
	"    return $res"
	"}"
    , _).


tcl_quote_string(In, Out) :-
	string_list(In, InL),
	tcl_quote_chars(InL, OutL, []),
	string_list(Out, OutL).

tcl_quote_chars([], Outs, Outs).
tcl_quote_chars([In|Ins], Outs, Outs0) :-
	tcl_quote_char(In, Outs, Outs1),
	tcl_quote_chars(Ins, Outs1, Outs0).

    :- mode tcl_quote_char(+,-,?).
    tcl_quote_char(C,	Cs1, Cs0) :- integer(C), C < 32, !,
	char_to_octal(C, Cs1, Cs0).
    tcl_quote_char(C,	Cs1, Cs0) :- integer(C), C >= 127, !,
	char_to_octal(C, Cs1, Cs0).
    tcl_quote_char(0'\,	[0'\,0'\|Cs], Cs) :- !.
    tcl_quote_char(0' ,	[0'\,0' |Cs], Cs) :- !.
    tcl_quote_char(0'{,	[0'\,0'{|Cs], Cs) :- !.
    tcl_quote_char(0'},	[0'\,0'}|Cs], Cs) :- !.
    tcl_quote_char(0'[,	[0'\,0'[|Cs], Cs) :- !.
    tcl_quote_char(0'],	[0'\,0']|Cs], Cs) :- !.
    tcl_quote_char(0';,	[0'\,0';|Cs], Cs) :- !.
    tcl_quote_char(C,	[C|Cs], Cs).

    :- mode char_to_octal(+,-,?).
    char_to_octal(C, [0'\,O1,O2,O3|Cs], Cs) :-
    	O1 is (C // 64) + 0'0,
    	O2 is ((C mod 64) // 8) + 0'0,
    	O3 is (C mod 8) + 0'0.

