#
# $Id: eclipse.tcl,v 1.15 1999/07/16 14:27:52 js10 Exp $
# 
# This file contains some primitive procedures that are needed to
# embed ECLiPSe into Tcl applications.
#
# Do NOT include any development support here!
# Do NOT assume the existence of an interactive ECLiPSe toplevel!
#

package provide eclipse 1.0

#----------------------------------------------------------------------
# ec_resume ?async?
#	resume ECLiPSe execution and execute posted goals and events
#	returns success,fail,yield (or raises an error)
#	if async, ECLiPSe is run in a separate thread
#----------------------------------------------------------------------

set ec_resume_active 0

proc ec_resume {{async 0}} {
    global ec_resume_active

    if $ec_resume_active { error "ec_resume cannot be nested" }
    set $ec_resume_active 1

    while { 1 } {
	set reslist [ec_resume_ $async]
	if $async {
	    set reslist running
	    while {$reslist == "running"} {
		set reslist [ec_resume_status]
		update
	    }
	}
	set res [lindex $reslist 0]
	switch $res {
	    flushio {
		set stream [lindex $reslist 1]
		if [catch "ec_flushio_stream $stream" err] {
		    tk_messageBox -icon error -type ok -message $err
		}
		update
	    }
	    waitio {
		set stream [lindex $reslist 1]
		if [catch "ec_waitio_stream $stream" err] {
		    tk_messageBox -icon error -type ok -message $err
		}
	    }
	    success -
	    fail -
	    yield {
		set $ec_resume_active 0
		return $res
	    }
	    running {
		error "Cannot do ec_resume while another ec_resume is running"
	    }
	    default {
		set $ec_resume_active 0
		error "Unexpected return from ec_resume: $reslist"
	    }
	}
    }
}

#----------------------------------------------------------------------
# ec_handle_events
#	restricted form of ec_resume
#	execute only events (e.g. queue events, posted events)
#	returns success (or raises an error)
#----------------------------------------------------------------------

proc ec_handle_events {} {

    set reslist [ec_handle_events_]
    while { 1 } {
	set res [lindex $reslist 0]
	switch $res {
	    flushio {
		set stream [lindex $reslist 1]
		if [catch "ec_flushio_stream $stream" err] {
		    tk_messageBox -icon error -type ok -message $err
		}
		update
	    }
	    waitio {
		set stream [lindex $reslist 1]
		if [catch "ec_waitio_stream $stream" err] {
		    tk_messageBox -icon error -type ok -message $err
		}
	    }
	    success {
		return $res
	    }
	    running {
		error "Cannot do ec_handle_events while ec_resume running"
	    }
	    default {
		error "Unexpected return from ec_resume: $reslist"
	    }
	}
	# now resume the Eclipse handler
	set reslist [ec_resume_ 0]
    }
}

#----------------------------------------------------------------------
# Handling ECLiPSe queues
#----------------------------------------------------------------------

proc ec_queue_connect {eclipse_name access {command {}}} {
    global ec_queue_out_handlers
    global ec_queue_in_handlers

    set channelid [ec_queue_open_ $eclipse_name $access]
    if {$command != ""} { ec_set_queue_handler $eclipse_name $access $command }
    return $channelid
}

proc ec_write_exdr {channel data {format S}} {
    puts -nonewline $channel [ec_tcl2exdr $data $format]
}

proc ec_set_queue_handler {eclipse_name access command} {
    global ec_queue_out_handlers
    global ec_queue_in_handlers
    switch -- $access {
    r { set ec_queue_out_handlers([ec_stream_nr $eclipse_name]) $command }
    w { set ec_queue_in_handlers([ec_stream_nr $eclipse_name]) $command }
    default { error "ec_set_queue_handler: bad access mode, should be r or w" }
    }
}

proc ec_flushio_stream {stream} {
    global ec_queue_out_handlers
    if [info exists ec_queue_out_handlers($stream)] {
	eval $ec_queue_out_handlers($stream) $stream
    } else {
	ec_stream_output_popup "Output occurred on ECLiPSe stream $stream" $stream
    }
}

proc ec_waitio_stream {stream} {
    global ec_queue_in_handlers
    if [info exists ec_queue_in_handlers($stream)] {
	eval $ec_queue_in_handlers($stream) $stream
    } else {
	ec_stream_input_popup "Input expected on ECLiPSe stream $stream" $stream
    }
}


#----------------------------------------------------------------------
# Sample stream I/O handlers
#----------------------------------------------------------------------

set ec_stream_input_string {}

proc ec_stream_input_popup {Msg Stream} {
    global ec_stream_input_string

    toplevel .ec_stream_input_box
    label .ec_stream_input_box.prompt  -width 40 -text $Msg
    entry .ec_stream_input_box.input -bg white -width 40 -textvariable ec_stream_input_string
    button .ec_stream_input_box.clear -text "clear" -command {.ec_stream_input_box.input delete 0 end}
    button .ec_stream_input_box.ok -text "ok" -command {destroy .ec_stream_input_box}
    bind .ec_stream_input_box.input <Return> {destroy .ec_stream_input_box}

    pack .ec_stream_input_box.prompt -side top -fill x
    pack .ec_stream_input_box.input -side top -fill x
    pack .ec_stream_input_box.clear -side left -expand 1 -fill x
    pack .ec_stream_input_box.ok -side left -expand 1 -fill x

    focus .ec_stream_input_box.input
    tkwait window .ec_stream_input_box 
    ec_queue_write $Stream $ec_stream_input_string
}


# Sample queue_out_handler: output into text widget

proc ec_stream_to_window {Tag Window Stream} {
    set data [ec_queue_read $Stream 1000]
    while {$data != ""} {
	$Window insert end $data $Tag
	set data [ec_queue_read $Stream 1000]
    }
    $Window see end
}

# Sample queue_out_handler: output into message popup

proc ec_stream_output_popup {Msg Stream} {
    if ![winfo exists .ec_stream_output_box] {
	toplevel .ec_stream_output_box
	label .ec_stream_output_box.msg  -width 40 -text $Msg
	text .ec_stream_output_box.text -width 40 -height 5 -bg white -yscrollcommand ".ec_stream_output_box.vscroll set" -wrap none -xscrollcommand ".ec_stream_output_box.hscroll set"
	scrollbar .ec_stream_output_box.vscroll -command ".ec_stream_output_box.text yview"
	scrollbar .ec_stream_output_box.hscroll -command ".ec_stream_output_box.text xview" -orient horizontal
	button .ec_stream_output_box.ok -text "ok" -command {destroy .ec_stream_output_box}
	pack .ec_stream_output_box.msg -side top -fill x
	pack .ec_stream_output_box.ok -side bottom -fill x
	pack .ec_stream_output_box.vscroll -side left -fill y
	pack .ec_stream_output_box.hscroll -side bottom -fill x
	pack .ec_stream_output_box.text -expand 1 -fill both
    }
    ec_stream_to_window {} .ec_stream_output_box.text $Stream
}


#----------------------------------------------------------------------
# Init
#----------------------------------------------------------------------

proc ec_init {} {
    global ec_rpc_in_channel
    global ec_rpc_out_channel

    set res [ec_init_]
    set ec_rpc_in_channel [ec_queue_connect ec_rpc_in w]
    set ec_rpc_out_channel [ec_queue_connect ec_rpc_out r]
    return $res
}

#----------------------------------------------------------------------
# ec_rpc goal ?format?
#	returns: instantiated goal, "fail" or "throw"
#----------------------------------------------------------------------

proc ec_rpc {Goal {Format S}} {
    global ec_rpc_in_channel
    global ec_rpc_out_channel

    if [ec_running] {
	error "Cannot do ec_handle_events while ec_resume running"
    }
    ec_write_exdr $ec_rpc_in_channel $Goal $Format
    ec_handle_events
    ec_read_exdr $ec_rpc_out_channel
}


#----------------------------------------------------------------------
# Load the parts of the interface which are implemented in C:
#
# ec_init_
# ec_cleanup
# ec_set_option
# ec_post_goal
# ec_post_event
# ec_resume_
# ec_running
# ec_handle_events_
# ec_queue_write
# ec_queue_read
# ec_stream_nr
# ec_queue_open_
# ec_read_exdr
# ec_tcl2exdr
# ec_exdr2tcl
#----------------------------------------------------------------------

switch $tcl_platform(platform) {
    windows {
        # This mess is needed to make dll loading work on Win95
	set prev [pwd]
	cd [file join [file dirname [info script]] .. lib i386_nt]
	load tkeclipse.dll
	cd $prev
    }
    default {
	load tkeclipse[info sharedlibextension]
    }
}

#----------------------------------------------------------------------
# Set defaults
# The user can change these before calling ec_init
#----------------------------------------------------------------------

# use queues for stdin/stdout/stderr and connect them to popups for now
ec_set_option 12 2
set ec_queue_in_handlers(0) "ec_stream_input_popup {Input expected on ECLiPSe input stream:}"
set ec_queue_out_handlers(1) "ec_stream_output_popup {Output occurred on ECLiPSe output stream:}"
set ec_queue_out_handlers(2) "ec_stream_output_popup {Output occurred on ECLiPSe error stream:}"

