#! /bin/sh
# \
	exec wish8.0 $0 ${1+"$@"}

#
# ECLiPSe Development Tools in Tcl
#
# Copyright 1999 IC-Parc, Imperial College, London
#
# $Id: eclipse_tools.tcl,v 1.44 1999/08/07 22:19:37 js10 Exp $
#
# Code in this file must only rely on primitives in eclipse.tcl.
# Don't assume these tools to be embedded into a particular
# application (like the tkeclipse toplevel)!
#

#----------------------------------------------------------------------
# Find and load the eclipse package
#----------------------------------------------------------------------

switch $tcl_platform(platform) {
    unix {
	set tkecl(ECLIPSEDIR) $env(ECLIPSEDIR)
    }
    windows {
	package require registry
	set tkecl(ECLIPSEDIR) [registry get \
	    HKEY_LOCAL_MACHINE\\SOFTWARE\\IC-Parc\\Eclipse\\4.2 ECLIPSEDIR]
    }
    default {
	error "$tcl_platform(platform) not supported"
	exit
    }
}

lappend auto_path [file join $tkecl(ECLIPSEDIR) lib_tcl]


#----------------------------------------------------------------------
# Load packages and initialise global settings
#----------------------------------------------------------------------

package require eclipse
package require AllWidgets
package require tkinspect
package require tkgetdir

set tkecl(history) {}
set tkecl(press_creep) 0
set tkecl(next_trace_line_depth) 1
set tkecl(trace_refresh_dg) 1
set tkecl(creepwaitevent) 0
set tkecl(dgf_tracedonly) 1
set tkecl(dgf_spiedonly) 0
set tkecl(dgf_wakeonly) 0


set tkecl(balloonhelp) 0
balloonhelp disable .

# other global variables

set tkecl(stats_interval) 1

set tkecl(filetypes) {
	{{ECLiPSe Files} {.ecl .pl}}
	{{ECLiPSe specific Files} {.ecl}}
	{{Prolog Files} {.pl}}
	{{ECLiPSe Precompiled Files} {.sd}}
	{{All Files} {*}}
}
set tkecl(defaultextension) {.ecl}

set tkecl(editor) ""
if [info exists env(VISUAL)] { set tkecl(editor) $env(VISUAL) }
if {$tkecl(editor) == ""} {
    if [info exists env(EDITOR)] { set tkecl(editor) $env(EDITOR) }
}
if {$tkecl(editor) == ""} { set tkecl(editor) "wordpad" }


#----------------------------------------------------------------------
# Testing code
#----------------------------------------------------------------------

proc tkecl:test {} {
    ec_rpc_check {exit_block(abort)}
}

proc tkecl:rpc {} {
    global tkecl
    if ![winfo exists .ec_rpc] {
	toplevel .ec_rpc
	wm title .ec_rpc "ECLiPSe Simple Query"
	pack [label .ec_rpc.entrylabel -justify left -text "Enter a goal in ECLiPSe syntax:"] -expand 1 -fill x
	pack [entry .ec_rpc.entry -bg white -textvariable tkecl(rpc_goal)] \
		-expand 1 -fill x
	pack [label .ec_rpc.textlabel -text "Reply:"] -expand 1 -fill x
	pack [text .ec_rpc.text -bg white -height 8]
	bind .ec_rpc.entry <Return> tkecl:run_rpc
	button .ec_rpc.run -text "Run (once)" -command tkecl:run_rpc
	button .ec_rpc.close -text Close -command "destroy .ec_rpc"
	pack .ec_rpc.run .ec_rpc.close -side left -expand 1 -fill x
	focus .ec_rpc.entry
	balloonhelp .ec_rpc.run "Execute an ECLiPSe goal once at a new break level."
	bind .ec_rpc <Alt-h> {tkecl:Get_helpfileinfo rpc .ec_rpc}
    } else {
	tkinspect:RaiseWindow .ec_rpc
    }
}

proc tkecl:run_rpc {} {
    global tkecl
    .ec_rpc.text insert end [ec_rpc $tkecl(rpc_goal)]
    .ec_rpc.text insert end "\n"
}

proc ec_rpc_check {goal {format S}} {
    set result [ec_rpc $goal $format]
    switch $result {
	fail {
	    tk_messageBox -type ok -message "ECLiPSe goal failed: $goal"
	}
	throw {
	    tk_messageBox -type ok -message "ECLiPSe goal aborted: $goal"
	}
    }
    return $result
}


#----------------------------------------------------------------------
# History window
#----------------------------------------------------------------------

proc tkecl:popup_history {} {
    global tkecl

    if ![winfo exists .history] {
	toplevel .history
	listbox .history.box -width 40 -height 12 -yscrollcommand ".history.vscroll set"
	scrollbar .history.vscroll -command ".history.box yview"
	foreach goal $tkecl(history) {
	    .history.box insert 0 $goal
	}
	bind .history.box <Double-Button-1> {
	    set tkecl(goal) [selection get]
	}
	button .history.close -text Close -command {destroy .history}
	pack .history.close -side bottom -fill x
	pack .history.vscroll -side left -fill y
	pack .history.box -side left -fill both -expand 1
    }
}


#----------------------------------------------------------------------
# Predicate properties window
#----------------------------------------------------------------------

set tkecl(predproppred) ""
set tkecl(predpropmodule) ""

proc tkecl:combo_add_modules {w} {
    foreach item [tkecl:list_modules] {
	$w add $item
    }
}

proc tkecl:list_modules {} {
    lindex [ec_rpc_check {setof(X,current_module(X),L)}] 3
}

proc tkecl:popup_pred_prop {} {
    global tkecl

    if ![winfo exists .predprop] {
	toplevel .predprop
	wm title .predprop "ECLiPSe Predicate Browser"

	set tkecl(predpropwhich) defined
	frame .predprop.f1 -relief raised -bd 1
	combobox .predprop.which -click single \
		-list {defined exported imported local global visible} \
		-listheight 6 \
		-labeltext "Predicates " \
		-textvariable tkecl(predpropwhich) \
		-command tkecl:display_predicates
	pack .predprop.which -in .predprop.f1 -side left -expand 1 -fill x

	set modules [tkecl:list_modules]
	combobox .predprop.modules -list $modules -click single \
		-labeltext " in module: " \
		-listheight 6 \
		-textvariable tkecl(predpropmodule) \
		-command tkecl:display_predicates
	pack .predprop.modules -in .predprop.f1 -side left -expand 1 -fill x

	listbox .predprop.preds -width 20 \
		-yscrollcommand ".predprop.vscroll set"
	scrollbar .predprop.vscroll -command ".predprop.preds yview"
	bind .predprop.preds <ButtonRelease-1> {+tkecl:display_predprops .predprop.preds}

	button .predprop.close -text Close -command "destroy .predprop"

	frame .predprop.f2 -relief groove -bd 1
	pack [label .predprop.predlabel -text "Properties of Predicate:"] -in .predprop.f2 -side top -fill x
	pack [label .predprop.predname -relief sunken] -in .predprop.f2 -side top -fill x
	tkecl:add_rb .predprop.f2 disabled declared {off on}
	tkecl:add_rb .predprop.f2 disabled defined {off on}
	tkecl:add_rb .predprop.f2 disabled debugged {off on}
	tkecl:add_rb .predprop.f2 disabled stability {static dynamic}
	tkecl:add_rb .predprop.f2 disabled call_type {prolog external}
	tkecl:add_rb .predprop.f2 disabled type {built_in user}
	tkecl:add_rb .predprop.f2 disabled tool {off on}
#	tkecl:add_rb .predprop.f2 disabled visibility {local imported exported global}
	tkecl:add_rb .predprop.f2 disabled demon {off on}
	tkecl:add_rb .predprop.f2 disabled parallel {off on}
	tkecl:add_rb .predprop.f2 disabled protected {off on}
#	tkecl:add_rb .predprop.f2 disabled statistics {off on}
	tkecl:add_rb .predprop.f2 active leash {stop notrace}
	tkecl:add_rb .predprop.f2 active skip {off on}
	tkecl:add_rb .predprop.f2 active start_tracing {off on}
	tkecl:add_rb .predprop.f2 active spy {off on}

	button .predprop.f2.show -text "Show source" -command tkecl:display_source
	pack .predprop.f2.show -side top -fill x

	pack .predprop.f1 -side top -expand 1 -fill x
	pack .predprop.preds -side left -expand 1 -fill both
	pack .predprop.vscroll -side left -fill y
	pack .predprop.f2 -side top -expand 1 -fill x -padx 3 -pady 3 -ipadx 3 -ipady 3
	pack .predprop.close -side top -fill x

        balloonhelp .predprop.preds "Predicates list - select one to view its \
	    properties\n (see manual for details on properties)"
        balloonhelp .predprop.which "Type of predicates listed in predicates list.\n\
	    click arrow on right to change type"
        balloonhelp .predprop.modules "Module of predicates listed in predicates list.\n\
	    click arrow on right to change module"
        balloonhelp .predprop.predname "Name, operator and mode information for predicate if known"
	bind .predprop <Alt-h> "tkecl:Get_helpfileinfo pred .predprop"
        tkecl:display_predicates dummy
    } else {
	tkinspect:RaiseWindow .predprop
    }

}

proc tkecl:display_predicates {dummy} {
    global tkecl

    .predprop.preds delete 0 end
    set preds [lindex [lindex [ec_rpc_check "tracer_tcl:list_predicates($tkecl(predpropwhich),$tkecl(predpropmodule),_)"] 2] 3]
    foreach item $preds {
	.predprop.preds insert end $item
    }
}

proc tkecl:add_rb {parent state name values} {
    global tkecl
#    frame $parent.$name -relief groove -bd 1
    frame $parent.$name
    label $parent.$name.label -text $name -anchor w -width 20
    pack $parent.$name.label -side left
    foreach val $values {
	radiobutton $parent.$name.$val -text $val -variable tkecl(pp_$name) \
		-value $val -anchor w -state $state -command "tkecl:update_predprop $name"
	pack $parent.$name.$val -side left
    }
    pack $parent.$name -side top -fill x
}

proc tkecl:update_predprop {name} {
    global tkecl
    if {$tkecl(predproppred) != ""} {
	;# only update if a predicate has been selected...
	tkecl:set_pred_flag $tkecl(predproppred) $tkecl(predpropmodule) $name $tkecl(pp_$name)
    }
}

proc tkecl:display_predprops {w} {
    global tkecl

    set selected [$w curselection]
    if ![string match "" $selected] { 
	set tkecl(predproppred) [$w get $selected]
    }
    set home [tkecl:pred_flag_value $tkecl(predproppred) $tkecl(predpropmodule) definition_module]
    set mode [tkecl:pred_flag_value $tkecl(predproppred) $tkecl(predpropmodule) mode]
    .predprop.predname configure -text "$home : $mode"
    foreach name {call_type debugged declared defined leash protected \
	    skip spy stability tool type demon parallel statistics start_tracing} {
	set tkecl(pp_$name) [tkecl:pred_flag_value $tkecl(predproppred) $tkecl(predpropmodule) $name]
    }
    if [winfo exists .ec_source] {
	tkecl:display_source
    }
}

proc tkecl:pred_flag_value {pred module name} {
    set result [ec_rpc \
    	[list : tracer_tcl [list flag_value $pred $name $module _]] (()(S()()_))]
    # rpc can fail, return "" in that case
    lindex [lindex $result 2] 4
}

proc tkecl:set_pred_flag {pred module name value} {
    ec_rpc \
    	[list : tracer_tcl [list set_flag_string $pred $name $value $module]] (()(S()()()))
}

#----------------------------------------------------------------------
# Predicate source window
#----------------------------------------------------------------------

proc tkecl:display_source {} {
    global tkecl

    if ![winfo exists .ec_source] {
	toplevel .ec_source
	wm title .ec_source "ECLiPSe Predicate Source Viewer"

	text .ec_source.text -bg white -yscrollcommand ".ec_source.vscroll set" -wrap none -xscrollcommand ".ec_source.hscroll set"
	.ec_source.text tag configure highlight -foreground blue
	scrollbar .ec_source.vscroll -command ".ec_source.text yview"
	scrollbar .ec_source.hscroll -command ".ec_source.text xview" -orient horizontal

	frame .ec_source.buttons
	button .ec_source.buttons.refresh -text Refresh -command {tkecl:display_source}
	    pack .ec_source.buttons.refresh -side left -fill x -expand 1
	button .ec_source.buttons.close -text Close -command {destroy .ec_source}
	    pack .ec_source.buttons.close -side left -fill x -expand 1

	pack .ec_source.buttons -side bottom -fill x
	pack .ec_source.vscroll -side left -fill y
	pack .ec_source.hscroll -side bottom -fill x
	pack .ec_source.text -expand 1 -fill both
	balloonhelp .ec_source.buttons "Try to view source of selected predicate (from predicates or tracer window)"
	bind .ec_source <Alt-h> {tkecl:Get_helpfileinfo sour .ec_source}
    } else {
	tkinspect:RaiseWindow .ec_source
    }
   .ec_source.text delete 1.0 end
    ec_rpc tracer_tcl:print_source_string("$tkecl(predproppred)",$tkecl(predpropmodule))
    ec_stream_to_window {} .ec_source.text gui_source_output
}

proc tkecl:set_and_display_source {pred module} {
    global tkecl
    set tkecl(predproppred) $pred
    set tkecl(predpropmodule) $module
    tkecl:display_source
}

#----------------------------------------------------------------------
# Global settings window
#----------------------------------------------------------------------

proc tkecl:popup_global_state {} {
    global tkecl

    if ![winfo exists .gstate] {
	toplevel .gstate
	wm withdraw .gstate
	wm title .gstate "ECLiPSe Global Settings"
	
	tkecl:add_radiobutton .gstate all_dynamic "off on"
	tkecl:add_radiobutton .gstate coroutine "off on"
	tkecl:add_radiobutton .gstate debugging "nodebug creep leap"
	tkecl:add_radiobutton .gstate debug_compile "off on"
	tkecl:add_radiobutton .gstate dfid_compile "off on"
	tkecl:add_radiobutton .gstate enable_interrupts "off on"
	tkecl:add_radiobutton .gstate float_precision "single double"
	tkecl:add_radiobutton .gstate gc "off on verbose"
	tkecl:add_radiobutton .gstate goal_expansion "off on"
	tkecl:add_radiobutton .gstate macro_expansion "off on"
	tkecl:add_radiobutton .gstate occur_check "off on"
	tkecl:add_radiobutton .gstate prefer_rationals "off on"
	tkecl:add_radiobutton .gstate variable_names "off on check_singletons"

	tkecl:add_popupentry .gstate output_mode "tkecl:Set_global_output_mode output_mode" Change {}
	tkecl:add_entry .gstate gc_interval number I
	tkecl:add_entry .gstate gc_interval_dict number I
#    tkecl:add_entry .gstate output_mode none S
	tkecl:add_entry .gstate print_depth number I
#    tkecl:add_entry .gstate cwd none S
	tkecl:add_popupentry .gstate cwd {tkecl:get_newcwd cwd} Change S
	tkecl:add_menuentry .gstate library_path tkecl:paths_menu Change S
	button .gstate.close -text Close -command {destroy .gstate}
	pack .gstate.close -side top -fill x
	wm minsize .gstate 380 30
	wm resizable .gstate 1 0
	wm deiconify .gstate

	balloonhelp .gstate "ECLiPSe global state - see manual for descriptions of flags"
	balloonhelp .gstate.library_path "left click in entry to see all paths"
	bind .gstate <Alt-h> {tkecl:Get_helpfileinfo glob .gstate}
    } else {
	tkinspect:RaiseWindow .gstate
    }
}


proc tkecl:add_radiobutton {parent name values} {
    global tkecl
    set tkecl($name) [lindex [ec_rpc_check get_flag($name,_)] 2]
#    frame $parent.$name -relief groove -bd 1
    frame $parent.$name
    label $parent.$name.label -text $name -anchor w -width 20
    pack $parent.$name.label -side left
    foreach val $values {
	radiobutton $parent.$name.$val -text $val -variable tkecl($name) \
		-value $val -anchor w -command "tkecl:set_flag $name ()"
	pack $parent.$name.$val -side left
    }
    pack $parent.$name -side top -fill x
}

proc tkecl:add_popupentry {parent name command ctext exdr_type} {
    global tkecl

    set f [frame $parent.$name]
    pack [label $f.label -text $name -anchor w -width 20] -side left
    if {$exdr_type == ""} {
	set info [label $f.val -justify right -relief groove -textvariable tkecl($name)]
    } else {
	set info [entry $f.val -bg white -justify right -relief sunken -textvariable tkecl($name)]
	bind $f.val <Return> "tkecl:set_flag $name $exdr_type"
    }
    pack $info -side left -expand 1 -fill x
#    bind $parent.$name.val <Return> "tkecl:set_flag $name S"
    set tkecl($name) [lindex [ec_rpc_check get_flag($name,_)] 2]
    pack [button $f.b -anchor e -text $ctext -command $command] -side right 
    pack $f -side top -fill x
}

proc tkecl:add_menuentry {parent name buildmenu mtext exdr_type} {
    global tkecl

    set f [frame $parent.$name]
    pack [label $f.label -text $name -anchor w -width 20] -side left
    if {$exdr_type == ""} {
	set info [label $f.val -justify right -relief groove -textvariable tkecl($name)]
    } else {
	set info [entry $f.val -bg white -justify right -relief sunken -textvariable tkecl($name)]
	bind $f.val <Return> "tkecl:set_flag $name $exdr_type"
    }
    pack $info -side left -expand 1 -fill x
#    bind $parent.$name.val <Return> "tkecl:set_flag $name S"
    set tkecl($name) [lindex [ec_rpc_check get_flag($name,_)] 2]
    pack [menubutton $f.b -text $mtext -menu $f.b.m -relief raised] -side right 
    $buildmenu $f.b $name
    pack $f -side top -fill x
}

proc tkecl:add_entry {parent name vtype exdr_type} {
    global tkecl
#    frame $parent.$name -relief groove -bd 1
    switch -exact -- $vtype {
	number {
	    set vstring {regexp {^[0-9]*$} %P}
	}
	none {
	    set vstring {regexp {.*} %P}
	}
    }
    frame $parent.$name
    label $parent.$name.label -text $name -anchor w -width 20
    set tkecl($name) [lindex [ec_rpc_check get_flag($name,_)] 2]
    if {$exdr_type != ""} {
	ventry $parent.$name.val -bg white -justify right -relief sunken -textvariable tkecl($name) -validate key -invalidcmd bell -vcmd $vstring
	bind $parent.$name.val <Return> "tkecl:set_flag $name $exdr_type"
    } else {
	entry $parent.$name.val -relief groove -justify right -textvariable tkecl($name) 
	bind $parent.$name.val <Any-Key> {break}
	bind $parent.$name.val <Button-2> {break}
	bind $parent.$name.val <ButtonRelease-2> {break}
	bind $parent.$name.val <Button-1> {break}
    }
    pack $parent.$name.label -side left
    pack $parent.$name.val -side right -expand 1 -fill x
    pack $parent.$name -side top -fill x
}

proc tkecl:set_flag {name exdr_type} {
    global tkecl
    ec_rpc_check [list set_flag $name $tkecl($name)] (()$exdr_type)
}



#----------------------------------------------------------------------
# Files window
#----------------------------------------------------------------------

proc tkecl:compile_popup {} {
    global tkecl

    set file [tk_getOpenFile -defaultextension $tkecl(defaultextension) -filetypes $tkecl(filetypes) -title "Compile File" -initialdir [pwd]]
    tkecl:compile_file $file
}

proc tkecl:compile_file {file {module ""}} {
    if {$file != ""} {
	if {$module == ""} {
	    set module [lindex [ec_rpc_check get_flag(toplevel_module,_)] 2]
	}
	if {[file exists $file] && [file readable $file]} {
	    ec_rpc [list : tracer_tcl [list compile_os_file $file $module]] (()(S()))
	} else {
	    tk_messageBox -icon error -type ok -message "Cannot access file $file"
	}
	tkecl:refresh_file_window
    }
}

proc tkecl:edit_popup {} {
    global tkecl
    set file [tk_getOpenFile -defaultextension $tkecl(defaultextension) -filetypes $tkecl(filetypes) -title "Edit File" -initialdir [pwd]]
    tkecl:edit_file $file
}

proc tkecl:edit_file {file} {
    global tkecl
   # if {$file != ""} {
	exec $tkecl(editor) $file &
	# No effect because editor in background...
	# tkecl:refresh_file_window
   # }
}

proc tkecl:popup_file_window {} {
    global tkecl

    if ![winfo exists .ec_files] {
	toplevel .ec_files
	wm title .ec_files "ECLiPSe Source File Manager"

	listbox .ec_files.names -selectmode single -width 20 -height 25 -yscrollcommand ".ec_files.vscroll set"
	listbox .ec_files.state -selectmode browse -width 11 -height 25 -yscrollcommand ".ec_files.vscroll set"
	scrollbar .ec_files.vscroll -command tkecl:scroll_file_status
	bind .ec_files.names <Double-Button-1> {
	    tkecl:edit_file [.ec_files.names get [.ec_files.names curselection]]
	}

	frame .ec_files.buttons
	button .ec_files.buttons.browse -text "Add file" -command {
		set file [tk_getOpenFile -defaultextension $tkecl(defaultextension) -filetypes $tkecl(filetypes) -title "Add Source File" -initialdir [pwd]]
		if {$file != ""} "tkecl:add_source_file $file"
	    }
	    pack .ec_files.buttons.browse -side left -fill x -expand 1
	button .ec_files.buttons.edit -text Edit -command {
		set sel [.ec_files.names curselection]
		if {$sel != ""} {
		    tkecl:edit_file [.ec_files.names get $sel]
		} else {
		    set file [tk_getSaveFile -defaultextension $tkecl(defaultextension) -filetypes $tkecl(filetypes) -title "New Source File"]
		    if {$file != ""} {
			# Create the file (some editors require it)
			close [open $file w]
			tkecl:add_source_file $file
			tkecl:edit_file $file
		    }
		}}
	    pack .ec_files.buttons.edit -side left -fill x -expand 1
	button .ec_files.buttons.compile -text Compile -command {
		set sel [.ec_files.names curselection]
		if {$sel != ""} {
		    tkecl:compile_file [.ec_files.names get $sel]
		} else {
		    tkecl:compile_popup
		}}
	    pack .ec_files.buttons.compile -side left -fill x -expand 1
	button .ec_files.buttons.refresh -text Redisplay -command tkecl:refresh_file_window
	    pack .ec_files.buttons.refresh -side left -fill x -expand 1
	button .ec_files.buttons.make -text Make -command {
	    	ec_rpc_check make
		tkecl:refresh_file_window }
	    pack .ec_files.buttons.make -side left -fill x -expand 1
	button .ec_files.buttons.close -text Close -command {destroy .ec_files}
	    pack .ec_files.buttons.close -side left -fill x -expand 1

	pack .ec_files.buttons -side bottom -fill x
	pack .ec_files.vscroll -side left -fill y
	pack .ec_files.names -side left -fill both -expand 1
	pack .ec_files.state -side left -fill y
	balloonhelp .ec_files.names "ECLiPSe source files - files tracked by ECLiPSe for compilation by `make'"
	balloonhelp .ec_files.state "`ok' - previously compiled file\n \
		`modified' - previously compiled file that has been modified \
		(will be recompiled with `make')\n `new' - file names added to source list \
		(will not be compiled by `make' until it is explicitly compiled first)"
	balloonhelp .ec_files.buttons.browse "Add a file to list"
	balloonhelp .ec_files.buttons.edit "edit a file. If file is not in source list, it will be added."
	balloonhelp .ec_files.buttons.compile "compile selected file from source list"
	balloonhelp .ec_files.buttons.refresh "Refresh display - update status of files in source list"
	bind .ec_files <Alt-h> "tkecl:Get_helpfileinfo file .ec_files "
    } else {
	tkinspect:RaiseWindow .ec_files
    }
    tkecl:refresh_file_window
}

proc tkecl:add_source_file {file} {
    ec_rpc_check [list : tracer_tcl [list record_source_file $file]] (()(S))
    tkecl:refresh_file_window
}

proc tkecl:scroll_file_status {args} {
    eval ".ec_files.names yview $args"
    eval ".ec_files.state yview $args"
}

proc tkecl:refresh_file_window {} {
    if [winfo exists .ec_files] {
	.ec_files.names delete 0 end
	.ec_files.state delete 0 end
	set files [lindex [lindex [ec_rpc_check tracer_tcl:list_files(_)] 2] 1]
	foreach item $files {
	    .ec_files.names insert end [lindex $item 0]
	    .ec_files.state insert end [lindex $item 1]
	}
        # the following hangs on Windows
#	update idletasks; # without this update it even hangs on Unix
#	while { [lindex [.ec_files.names xview] 1] < 0.9} {
#	    .ec_files.names xview moveto [lindex [.ec_files.names xview] 1]
#	}
    }
}


#----------------------------------------------------------------------
# Help
#----------------------------------------------------------------------

proc tkecl:popup_help_window {} {
    global tkecl

    if ![winfo exists .ec_help] {
	toplevel .ec_help
	wm title .ec_help "ECLiPSe Help"

	label .ec_help.label -justify left -text "Type in a string to match, or predicate_name/arity:"

	text .ec_help.text -bg white -yscrollcommand ".ec_help.vscroll set" -wrap none -xscrollcommand ".ec_help.hscroll set"
	.ec_help.text configure -cursor left_ptr
	.ec_help.text tag configure highlight -foreground blue
	scrollbar .ec_help.vscroll -command ".ec_help.text yview"
	scrollbar .ec_help.hscroll -command ".ec_help.text xview" -orient horizontal

	entry .ec_help.input -bg white -width 40 -textvariable tkecl(help_input)
	bind .ec_help.input <Return> {tkecl:display_help}
	bind .ec_help.text <Button-1> {.ec_help.input delete 0 end; \
		.ec_help.input insert end [.ec_help.text get "current wordstart" "current wordend"]; break}
	bind .ec_help.text <Double-Button-1> {tkecl:display_help; break}

	frame .ec_help.buttons
	button .ec_help.buttons.display -text Display -command {tkecl:display_help}
	button .ec_help.buttons.clear -text Clear -command {.ec_help.text delete 1.0 end}
	button .ec_help.buttons.close -text Close -command {destroy .ec_help}

	pack .ec_help.buttons.display .ec_help.buttons.clear \
		.ec_help.buttons.close -side left -fill x -expand 1

	pack .ec_help.buttons -side bottom -fill x
	pack .ec_help.buttons -side bottom -fill x
	pack .ec_help.label -side top -fill x
	pack .ec_help.input -side top -fill x
	pack .ec_help.vscroll -side left -fill y
	pack .ec_help.hscroll -side bottom -fill x
	pack .ec_help.text -expand 1 -fill both
	balloonhelp .ec_help.buttons "Obtaining information on ECLiPSe built-ins"
	balloonhelp .ec_help.text "Left click on any word to put it in entry\nDouble left-click to look word up directly"
	bind .ec_help <Alt-h> {tkecl:Get_helpfileinfo help .ec_help}
	focus .ec_help.input
    } else {
	tkinspect:RaiseWindow .ec_help
    }
}

proc tkecl:display_help {} {
    global tkecl
    .ec_help.input selection range 0 end
    .ec_help.text tag remove highlight 1.0 end
    .ec_help.text configure -cursor watch ; update idletasks
    ec_rpc [list : tracer_tcl [list gui_help gui_help_output $tkecl(help_input)]] (()(()S))
    ec_stream_to_window highlight .ec_help.text gui_help_output
    .ec_help.text configure -cursor left_ptr
}


#----------------------------------------------------------------------
# Delayed goals
#----------------------------------------------------------------------

proc tkecl:popup_dg_window {} {
    global tkecl

    if ![winfo exists .ec_dg] {
	toplevel .ec_dg
	wm title .ec_dg "ECLiPSe Delayed Goals"

	text .ec_dg.text -bg white -yscrollcommand ".ec_dg.vscroll set" -wrap none -xscrollcommand ".ec_dg.hscroll set"
	scrollbar .ec_dg.vscroll -command ".ec_dg.text yview"
	scrollbar .ec_dg.hscroll -command ".ec_dg.text xview" -orient horizontal

	set ff [frame .ec_dg.filters]
	pack [checkbutton $ff.traced -text "traced only" -variable tkecl(dgf_tracedonly)] -side left
	pack [checkbutton $ff.spied -text "spied only" -variable tkecl(dgf_spiedonly)] -side left
	pack [checkbutton $ff.wake -text "scheduled only" -variable tkecl(dgf_wakeonly)] -side left
	frame .ec_dg.buttons
	button .ec_dg.buttons.refresh -text Refresh -command {tkecl:refresh_dg}
	    pack .ec_dg.buttons.refresh -side left -fill x -expand 1
	button .ec_dg.buttons.close -text Close -command {destroy .ec_dg}
	    pack .ec_dg.buttons.close -side left -fill x -expand 1

	pack .ec_dg.filters -side top -fill x
	pack .ec_dg.buttons -side bottom -fill x
	pack .ec_dg.vscroll -side left -fill y
	pack .ec_dg.hscroll -side bottom -fill x
	pack .ec_dg.text -expand 1 -fill both
	balloonhelp .ec_dg.text "Delayed goals are displayed here. Green indicates goal has been scheduled.\n Left click on goal for a popup menu related to that goal."
	balloonhelp .ec_dg.buttons "List of goals that are currently being delayed.\n\
		Can be set to automatically refresh at every trace line from tracer window."
	balloonhelp $ff "Filter options for filtering displayed delay goals."
	balloonhelp $ff.traced "Show only goals which can be traced when selected."
	balloonhelp $ff.spied "Show only goals which are being spied when selected."
	balloonhelp $ff.wake "Show only goals which have been scheduled when selected."
	bind .ec_dg <Alt-h> {tkecl:Get_helpfileinfo dela .ec_dg}
    } else {
	tkinspect:RaiseWindow .ec_dg
    }
    tkecl:refresh_dg
}

proc tkecl:refresh_dg {} {
    global tkecl

    if [winfo exists .ec_dg] {
	.ec_dg.text delete 1.0 end
	eval .ec_dg.text tag delete [.ec_dg.text tag names]
	.ec_dg.text tag configure highlight -foreground #00b000
	ec_rpc_check [list : tracer_tcl [list gui_dg [list dg_filter $tkecl(dgf_tracedonly) $tkecl(dgf_spiedonly) $tkecl(dgf_wakeonly)]]] (()((III)))
    }
}

proc tkecl:handle_dg_print {stream} {
    global tkecl

    set gui_dg_info ec_queue$stream
    set info [ec_read_exdr $gui_dg_info]
    while {$info != "end"} {
	set state [lindex $info 1]
	set prio [lindex $info 2]
	set invoc [lindex $info 3]
	set linelength [lindex $info 4]
	if {$state == 1} {
	    set Tag highlight
	} else {
	    set Tag {}
	}
	set line [ec_queue_read gui_dg_output $linelength]
	if [winfo exists .ec_dg] {
	    set gstart [.ec_dg.text index end]
	    .ec_dg.text insert end $line $Tag
	    .ec_dg.text tag bind g$invoc <Button-1> "tkecl:popup_delaymenu .ec_dg.text $invoc %X %Y; break"
	    .ec_dg.text tag add g$invoc $gstart "$gstart lineend"
	    .ec_dg.text tag raise g$invoc
	}
	set info [ec_read_exdr $gui_dg_info]
    }
}

proc tkecl:popup_delaymenu {w invoc x y} {
    global tkecl

    if [winfo exists $w.gpopup] {
	destroy $w.gpopup
    }
    if {$invoc != 0} {
	set m [menu $w.gpopup -tearoff 0]
	set rpc_result [ec_rpc_check [list : tracer_tcl \
    		[list get_goal_info_by_invoc $invoc _ _ _]] (()(I___))]
	set greturn [lindex $rpc_result 2]
	set spec [lindex  $greturn 2]
	set tspec [lindex $greturn 3]
	set module [lindex $greturn 4]
	set spied [tkecl:pred_flag_value $spec $module spy]
	if {$spied == "on"} {
	    set spytext "Nospy $spec"
	    set spyval off
	} else {
	    set spytext "Spy $spec"
	    set spyval on
	}
	if {$invoc != 0} {
	    set invtext "($invoc)"
	} else {
	    set invtext ""
	}
	$m add command -label "$tspec @ $module $invtext" -state disabled	
	$m add command -label $spytext -command \
		[list tkecl:set_pred_flag $spec $module spy $spyval]
	$m add command -label "Display source for this predicate" -command \
		[list tkecl:set_and_display_source $spec $module]
	$m add command -label "Inspect this goal" -command \
		"tkinspect:Inspect_term_init invoc($invoc)"
	tk_popup $m $x $y
    }
}


#----------------------------------------------------------------------
# Tracer
#----------------------------------------------------------------------

proc tkecl:set_fail_invoc {invoc} {
    global tkecl

    set tkecl(fail_invoc) $invoc
    set tkecl(tracercommand) f
}

proc tkecl:set_jumpto_invoc {invoc} {
    global tkecl

    if [regexp -- {^[0-9]+$} $invoc]  {
	set tkecl(cont_invoc) $invoc
	set tkecl(tracercommand) i
    }
}

proc tkecl:set_jumpto_depth {depth} {
    global tkecl

    if [regexp -- {^[0-9]+$} $depth]  {
	set tkecl(cont_mindepth) $depth
	set tkecl(cont_maxdepth) $depth
	set tkecl(tracercommand) j
    }
}

proc tkecl:setup_creep {} {
    global tkecl
    
    set tkecl(press_creep) 1
    set tkecl(tracercommand) c
}

proc tkecl:end_creep {} {
    global tkecl

    after cancel $tkecl(creepwaitevent)
    set tkecl(press_creep) 0
    set tkecl(creepwaitover) 1
}

proc tkecl:popup_tracer {} {
    global tkecl

    if ![winfo exists .ec_tracer] {
	toplevel .ec_tracer
	wm title .ec_tracer "ECLiPSe Tracer"

	set tmbar [menu .ec_tracer.menubar]
	.ec_tracer config -menu $tmbar
	$tmbar add cascade -label "Windows" -menu $tmbar.win -underline 0
	menu $tmbar.win
	$tmbar.win add command -label "Raise top-level" -command "tkinspect:RaiseWindow ."
	$tmbar.win add command -label "Close Tracer" -command tkecl:close_tracer
	$tmbar add cascade -label "Options" -menu $tmbar.opt -underline 0
	menu $tmbar.opt
	$tmbar.opt add command -label "Continue until ..." -command tkecl:popup_filter
	$tmbar.opt add command -label "Set Output mode ..." -command "tkecl:Set_tracer_output_mode"
	$tmbar.opt add command -label "Set Print depth ..." -command "tkecl:Set_tracer_print_depth"
	$tmbar.opt add check -label "Refresh delayed goals at every trace line" -variable tkecl(trace_refresh_dg)
	$tmbar add cascade -label "Help" -menu $tmbar.help -underline 0
	menu $tmbar.help
        $tmbar.help add command -label "Tracer Help" -command "tkecl:Get_helpfileinfo trac .ec_tracer"

	frame .ec_tracer.trace
	label .ec_tracer.trace.label -text "Trace Log"
	text .ec_tracer.trace.text -bg white -yscrollcommand ".ec_tracer.trace.vscroll set" -wrap none -xscrollcommand ".ec_tracer.trace.hscroll set"
	.ec_tracer.trace.text tag configure call_style -foreground blue
	.ec_tracer.trace.text tag configure exit_style -foreground #00b000
	.ec_tracer.trace.text tag configure fail_style -foreground red
	scrollbar .ec_tracer.trace.vscroll -command ".ec_tracer.trace.text yview"
	scrollbar .ec_tracer.trace.hscroll -command ".ec_tracer.trace.text xview" -orient horizontal
	pack .ec_tracer.trace.vscroll -side left -fill y
	pack .ec_tracer.trace.hscroll -side bottom -fill x
	pack .ec_tracer.trace.text -side bottom -expand 1 -fill both
	pack .ec_tracer.trace.label -side left -expand 1 -fill x

	bind .ec_tracer.trace.text <Any-Key> {break}
	bind .ec_tracer.trace.text <ButtonRelease-2> {break}

	frame .ec_tracer.stack
	label .ec_tracer.stack.label -text "Call Stack"
	text .ec_tracer.stack.text -height 15 -bg white -yscrollcommand ".ec_tracer.stack.vscroll set" -wrap none -xscrollcommand ".ec_tracer.stack.hscroll set"
	.ec_tracer.stack.text tag configure call_style -foreground blue
	.ec_tracer.stack.text tag configure exit_style -foreground #00b000
	.ec_tracer.stack.text tag configure fail_style -foreground red
	.ec_tracer.stack.text configure -cursor left_ptr
	scrollbar .ec_tracer.stack.vscroll -command ".ec_tracer.stack.text yview"
	scrollbar .ec_tracer.stack.hscroll -command ".ec_tracer.stack.text xview" -orient horizontal
	pack .ec_tracer.stack.vscroll -side left -fill y
	pack .ec_tracer.stack.hscroll -side bottom -fill x
	pack .ec_tracer.stack.text -side bottom -expand 1 -fill both
	pack .ec_tracer.stack.label -side left -expand 1 -fill x

	bind .ec_tracer.stack.text <Any-Key> {break}
	bind .ec_tracer.stack.text <ButtonRelease-2> {break}

	frame .ec_tracer.buttons
	bind .ec_tracer.buttons <Enter> {focus %W}
	bind .ec_tracer.buttons <Key-c> {set tkecl(tracercommand) c}
	bind .ec_tracer.buttons <Key-l> {set tkecl(tracercommand) l}
	bind .ec_tracer.buttons <Key-s> {set tkecl(tracercommand) s}
	bind .ec_tracer.buttons <Key-z> {set tkecl(tracercommand) z}
#	bind .ec_tracer.buttons <Key-plus> {set tkecl(tracercommand) +}
#	bind .ec_tracer.buttons <Key-minus> {set tkecl(tracercommand) -}
	button .ec_tracer.buttons.creep -text Creep -command {}
	bind .ec_tracer.buttons.creep <Button-1> {tkecl:setup_creep}
	bind .ec_tracer.buttons.creep <ButtonRelease-1> {tkecl:end_creep}
	    pack .ec_tracer.buttons.creep -side left -fill x -expand 1
	button .ec_tracer.buttons.skip -text Skip -command {set tkecl(tracercommand) s}
	    pack .ec_tracer.buttons.skip -side left -fill x -expand 1
	button .ec_tracer.buttons.leap -text Leap -command {set tkecl(tracercommand) l}
	    pack .ec_tracer.buttons.leap -side left -fill x -expand 1
	button .ec_tracer.buttons.nodebug -text Nodebug -command {set tkecl(tracercommand) n ; tkinspect:RaiseWindow .}
	    pack .ec_tracer.buttons.nodebug -side left -fill x -expand 1

	frame .ec_tracer.cont
	button .ec_tracer.cont.button -text "To Invoc:" -command {set tkecl(tracercommand) i}
	    pack .ec_tracer.cont.button -side left -fill x -expand 1
	ventry .ec_tracer.cont.invoc \
		-vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
		-width 8 -textvariable tkecl(cont_invoc) -bg white
	    set tkecl(cont_invoc) 0
	    pack .ec_tracer.cont.invoc -side left
	button .ec_tracer.cont.jump -text "To Depth:" -command {set tkecl(tracercommand) j}
	    pack .ec_tracer.cont.jump -side left -fill x -expand 1
	ventry .ec_tracer.cont.mindepth \
		-vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
		-width 5 -textvariable tkecl(cont_mindepth) -bg white
	    set tkecl(cont_mindepth) 0
	    pack .ec_tracer.cont.mindepth -side left
	ventry .ec_tracer.cont.maxdepth -labeltext .. \
		-vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
		-width 5 -textvariable tkecl(cont_maxdepth) -bg white
	    set tkecl(cont_maxdepth) 9999
	    pack .ec_tracer.cont.maxdepth -side left
	button .ec_tracer.cont.zap -text "To Port:" -command {set tkecl(tracercommand) z}
	    pack .ec_tracer.cont.zap -side left
	set tkecl(zap_port) {Not Current}
	set portlist [lindex [lindex [ec_rpc_check "sepia_kernel:debug_port_names(_)"] 2] 1]
	combobox .ec_tracer.cont.ports -click single -listheight 16 -bg white \
		-width 10 -list "{Not Current} $portlist" -textvariable tkecl(zap_port)
	    pack .ec_tracer.cont.ports -side left
	button .ec_tracer.close -text Close -command tkecl:close_tracer

	pack .ec_tracer.stack -side top -expand 1 -fill both
	pack .ec_tracer.buttons -side top -fill x
	pack .ec_tracer.cont -side top -fill x
	pack .ec_tracer.trace -side top -expand 1 -fill both
	pack .ec_tracer.close -side top -fill x

	ec_rpc "set_flag(debugging,creep)"


#--------------------------------------------------------------------
# Balloon Help for tracer
#--------------------------------------------------------------------
       balloonhelp .ec_tracer "Tracer for ECLiPSe execution - start execution from main window"
       balloonhelp .ec_tracer.stack.label "Execution call stack - \
   shows the current goal and its ancestors.\n \
   Calls for current goal in blue, failure in red, success in green. \ 
   Ancestors printed with non-current bindings in black\n \
   Press left mosuse button over a stack item for popup menu related to that goal/predicate"
       balloonhelp .ec_tracer.trace.label "Trace log: chronological log of traced goals.\n Calls in blue, successes in green, failures in red\n Leading indentation indicates depth"
       balloonhelp .ec_tracer.buttons.creep "Creep to next tracable goal's debug port. \
   Typing `c' will also creep\n Press and hold button for continuous creep"
       balloonhelp .ec_tracer.buttons.skip "Skip to exit/fail port of goal (creep \
   if already at port). Typing `s' will also skip"
       balloonhelp .ec_tracer.buttons.leap "Leap to next spied goal's port. \
   Typing `l' will also leap"
       balloonhelp .ec_tracer.buttons.nodebug "Turn off debugging and \
   continue execution\n (Further outputs will be displayed on main window)"
       balloonhelp .ec_tracer.cont.button "Jump to port for goal with \
   invocation number on right"
       balloonhelp .ec_tracer.cont.jump "Jump to port for next goal with \
   depth in the ranges on the right"
       balloonhelp .ec_tracer.cont.zap "Jump to port selected on the right"
       bind .ec_tracer <Alt-h> "tkecl:Get_helpfileinfo trac .ec_tracer"
   } else {
       tkinspect:RaiseWindow .ec_tracer
   }   
}

proc tkecl:handle_debug_output {stream} {
    if {![winfo exists .ec_tracer]} {
	return
    }
    ec_stream_to_window {} .ec_tracer.trace.text $stream
}

# CAUTION: text widgets positions are a bit weird: the text widget always
# has a newline at the end, and the end-index is just after that. Therefore,
# an empty text widget has a newline at 1.0 and end == 2.0

proc tkecl:handle_trace_line {stream} {
    global tkecl

    if ![winfo exists .ec_tracer] {
	tkecl:popup_tracer
    }
    set trace_info [ec_read_exdr ec_queue$stream]
    set depth [lindex $trace_info 0]
    set style [lindex $trace_info 1]
    set line [lindex $trace_info 2]
    set invoc [lindex $trace_info 3]
    set port [lindex $trace_info 4]
    set tkecl(cont_invoc) $invoc  ;# defaults to current 

    .ec_tracer.trace.text tag configure $depth -lmargin1 "$depth m"
    .ec_tracer.trace.text insert end $line "$style $depth"
    .ec_tracer.trace.text see end
    .ec_tracer.trace.text insert end "\n" $style

    set stdepth [expr $depth + 1] ;# actual depth in printed stack
    if {$style == "fail_style"} {
	.ec_tracer.stack.text tag remove call_style $stdepth.0 end
	.ec_tracer.stack.text tag add fail_style $stdepth.0 end
	.ec_tracer.stack.text see $stdepth.0
	set tkecl(next_trace_line_depth) $stdepth
    } else {
	set next_line [lindex [split [.ec_tracer.stack.text index end-1chars] .] 0]
	if {$next_line > $tkecl(next_trace_line_depth)} {
	    # delete leftover exit/fail lines
	    # and tags to goals that are no longer accessible
	    tkecl:cleanup_goal_stack_line $tkecl(next_trace_line_depth) $next_line
	    set next_line $tkecl(next_trace_line_depth)
	}
	if {$next_line < $stdepth} {
	    while {$next_line < $stdepth} {
		.ec_tracer.stack.text insert end "\n"
		incr next_line
	    }
	} elseif {$next_line > $stdepth} {
	    .ec_tracer.stack.text delete $stdepth.0 $next_line.0
	}
	.ec_tracer.stack.text tag remove call_style 1.0 end
	.ec_tracer.stack.text insert end $line $style
	.ec_tracer.stack.text insert end "\n" $style
	tkecl:set_goalpopup $depth $invoc
	.ec_tracer.stack.text see end
	if {$style == "call_style"} {
	    set tkecl(next_trace_line_depth) [.ec_tracer.stack.text index end-1chars]
	} else {
	    set tkecl(next_trace_line_depth) $stdepth
	}
    }

    # Refresh the delayed goals display
    if {$tkecl(trace_refresh_dg)} { tkecl:refresh_dg }

    # Get the next command, either repeating creep or waiting for a new one
    tkecl:configure_tracer_buttons active
    if {($tkecl(press_creep) > 0) && \
	    [string match $tkecl(tracercommand) "c"]} {
	if {$tkecl(press_creep) == 1} { ;# initial press, wait longer
	    set interval 700
        } else {
	    set interval 50
	    set tkecl(press_creep) 2
	}
	set tkecl(creepwaitevent) [after $interval {set tkecl(creepwaitover) 1}]
	vwait tkecl(creepwaitover)
	if {($tkecl(press_creep) > 0) && \
	    [string match $tkecl(tracercommand) "c"]} {
	   set tkecl(press_creep) 2
        } else {
	    if [string match $tkecl(tracercommand) "c"] {
		tkwait variable tkecl(tracercommand)
	    } ;# otherwise already have another pending command
	}
    } else {
	tkwait variable tkecl(tracercommand)
    }

    # interpret the command and configure Eclipse for continuation
    while 1 {
	switch -exact -- $tkecl(tracercommand) {
	    N {
		# caution: if tracercommand = N the window is already destroyed!
		ec_queue_write debug_input n
		break
	    }
	    i {
		if [regexp -- {^[0-9]+$} $tkecl(cont_invoc)] {
		    tkecl:configure_tracer_buttons disabled
		    ec_rpc_check "sepia_kernel:configure_prefilter($tkecl(cont_invoc),_,_,_,_)"
		    ec_queue_write debug_input $tkecl(tracercommand)
		    break
		}
	    }
	    j {
		if {[regexp -- {^[0-9]+$} $tkecl(cont_mindepth)] && \
			[regexp -- {^[0-9]+$} $tkecl(cont_mindepth)]} {
		    tkecl:configure_tracer_buttons disabled
		    ec_rpc_check "sepia_kernel:configure_prefilter(_,$tkecl(cont_mindepth)..$tkecl(cont_maxdepth),_,_,_)"
		    ec_queue_write debug_input $tkecl(tracercommand)
		    break
		}
	    }
	    f { ;# fail to $tkecl(fail_invoc)
	         tkecl:configure_tracer_buttons disabled
                 ec_queue_write debug_input $tkecl(tracercommand)
                 ec_queue_write debug_input [ec_tcl2exdr $tkecl(fail_invoc) I]
	         break
	    }
	    z { ;# zap to $tkecl(zap_port)
		tkecl:configure_tracer_buttons disabled
		if [string match {Not Current} $tkecl(zap_port)] {
		    ec_queue_write debug_input $tkecl(tracercommand)
		} else {
		    ec_rpc_check [list : sepia_kernel [list configure_prefilter _ _ $tkecl(zap_port) _ dontcare]] {(()(__()_()))}
		}
		break
	    }
	    filter {
		set wanted_ports {}
		foreach port $tkecl(portlist) {
		    if $tkecl(filter_port,$port) {
			lappend wanted_ports $port
		    }
		}
		# sepia_kernel:configure_prefilter(Invoc, Depth, Ports, Preds, Module)
		ec_rpc_check [list : sepia_kernel \
		    [list configure_prefilter \
			[list .. $tkecl(filter_mininvoc) $tkecl(filter_maxinvoc)] \
			[list .. $tkecl(filter_mindepth) $tkecl(filter_maxdepth)] \
			$wanted_ports \
			$tkecl(filter_spy) \
			dontcare]] \
		    {(()((II)(II)[()*]()()))}
		break
	    }
	    default {
		tkecl:configure_tracer_buttons disabled
		ec_queue_write debug_input $tkecl(tracercommand)
		break
	    }
	}
	tkwait variable tkecl(tracercommand)
    }

    # return from the handler - this will continue Eclipse execution
}


proc tkecl:configure_tracer_buttons {state} {
    .ec_tracer.buttons.creep configure -state $state
    .ec_tracer.buttons.leap configure -state $state
    .ec_tracer.buttons.skip configure -state $state
    .ec_tracer.buttons.nodebug configure -state $state
    .ec_tracer.cont.button configure -state $state
    .ec_tracer.cont.jump configure -state $state
    .ec_tracer.cont.zap configure -state $state
    if [winfo exists .ec_tracer.filter] {
	.ec_tracer.filter.go configure -state $state
    }
}

proc tkecl:close_tracer {} {
    global tkecl
    set tkecl(tracercommand) N
    ec_rpc set_flag(debugging,nodebug)
    destroy .ec_tracer
}

proc tkecl:Set_tracer_print_depth {} {
    global tkecl

    set tkecl(tracer_prdepth) [lindex [lindex [ec_rpc_check "tracer_tcl:get_tracer_print_depth(V)"] 2] 1]
    tkinspect:Popnumentry .ec_tracer tkecl(tracer_prdepth) $tkecl(tracer_prdepth) {print depth:} 1
    ec_rpc "tracer_tcl:set_tracer_print_depth($tkecl(tracer_prdepth))"
}

proc tkecl:Set_tracer_output_mode {} {

    set popmode .ec_tracer.omodes
    if [winfo exists $popmode] {
	tkinspect:RaiseWindow $popmode
	return
    }

    toplevel $popmode
    wm title $popmode "Set Debug Output Modes"

    set return [ec_rpc_check \
       "tracer_tcl:get_tracer_output_modes(Modes,Status,Set,UnSet,TO,TD,TS)"]
    set newmodes [tkecl:Set_output_mode $popmode $return]
    ec_rpc_check "tracer_tcl:set_tracer_output_modes($newmodes)"
}


proc tkecl:popup_goalmenu {w invoc depth x y} {
    global tkecl

    if [winfo exists $w.gpopup] {
	destroy $w.gpopup
    }
    set m [menu $w.gpopup -tearoff 0]
    set rpc_result [ec_rpc_check [list : tracer_tcl \
    		[list get_goal_info_by_invoc $invoc _ _ _]] (()(I___))]
    set greturn [lindex $rpc_result 2]
    set spec [lindex  $greturn 2]
    set tspec [lindex $greturn 3]
    set module [lindex $greturn 4]

    if {![string match unknown $spec] } {
	$m add command -label "$tspec @ $module" -state disabled
	set spied [tkecl:pred_flag_value $spec $module spy]
	if {$spied == "on"} {
	    set spytext "Nospy $spec"
	    set spyval off
	} else {
	    set spytext "Spy $spec"
	    set spyval on
	}
	$m add command -label $spytext -command \
		[list tkecl:set_pred_flag $spec $module spy $spyval]
	$m add command -label "Display source for this predicate" -command \
		[list tkecl:set_and_display_source $spec $module]
	$m add command -label "Inspect this goal" -command \
		"tkinspect:Inspect_term_init invoc($invoc)"
	$m add command -label "Force failure of this goal" -command \
		"tkecl:set_fail_invoc $invoc"
	$m add command -label "Jump to invocation number ($invoc)" -command \
		"tkecl:set_jumpto_invoc $invoc"
    }
    $m add command -label "Jump to depth $depth" -command \
    	"tkecl:set_jumpto_depth $depth"
    $m add separator
    $m add command -label "Refresh goal stack" -command \
	    "tkecl:refresh_goal_stack"

    tk_popup $m $x $y
}


proc tkecl:cleanup_goal_stack_line {depth next_line} {
    set taglist [.ec_tracer.stack.text tag names $depth.0]
    set invocidx [lsearch -regexp $taglist  {^[0-9]+$}]
    ;# tags in the lines are also deleted
    if {$invocidx >= 0} {
	.ec_tracer.stack.text tag delete [lindex $taglist $invocidx]
    }
    .ec_tracer.stack.text delete $depth.0 $next_line.0
}

proc tkecl:refresh_goal_stack {} {
    foreach anc  [lindex [lindex [ec_rpc tracer_tcl:get_ancestors(_)] 2] 1] {
	foreach {pred depth invoc line} $anc {break}
	set stdepth [expr $depth+1]
	tkecl:cleanup_goal_stack_line $stdepth [expr $stdepth+1]
	append line "\n"
	.ec_tracer.stack.text insert $stdepth.0 $line call_style
	tkecl:set_goalpopup $depth $invoc 
    }

    .ec_tracer.stack.text see end
}

proc tkecl:set_goalpopup {depth invoc} {
# print goal line in the stack display and set up the tag for it
    .ec_tracer.stack.text tag bind $invoc <Button-1> "tkecl:popup_goalmenu .ec_tracer.stack.text $invoc $depth %X %Y; break"
    set stdepth [expr $depth + 1]
    .ec_tracer.stack.text tag add $invoc $stdepth.0 $stdepth.end
    .ec_tracer.stack.text tag raise $invoc
}

proc tkecl:popup_filter {} {
    global tkecl

    if [winfo exists .ec_tracer.filter] {
	tkinspect:RaiseWindow .ec_tracer.filter
	return
    }
    
    toplevel .ec_tracer.filter
    wm title .ec_tracer.filter "Continue until"

    label .ec_tracer.filter.label -text "Continue to a port with the following properties:"
    pack .ec_tracer.filter.label -side top
    frame .ec_tracer.filter.settings -relief groove -bd 1
    pack .ec_tracer.filter.settings -side top -ipadx 3 -ipady 3 -pady 5 -padx 5

    set row 0
    set col 0
    set cols 5

    ventry .ec_tracer.filter.settings.mininvoc -labeltext Invoc: \
	    -vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
	    -width 8 -textvariable tkecl(filter_mininvoc) -bg white
	set tkecl(filter_mininvoc) 0
    ventry .ec_tracer.filter.settings.maxinvoc -labeltext .. \
	    -vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
	    -width 8 -textvariable tkecl(filter_maxinvoc) -bg white
	set tkecl(filter_maxinvoc) 9999999
    grid .ec_tracer.filter.settings.mininvoc - .ec_tracer.filter.settings.maxinvoc -sticky w
    incr row

    ventry .ec_tracer.filter.settings.mindepth -labeltext Depth: \
	    -vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
	    -width 5 -textvariable tkecl(filter_mindepth) -bg white
	set tkecl(filter_mindepth) 0
    ventry .ec_tracer.filter.settings.maxdepth -labeltext .. \
	    -vcmd {regexp {^[0-9]*$} %P} -validate key -invalidcmd bell \
	    -width 5 -textvariable tkecl(filter_maxdepth) -bg white
	set tkecl(filter_maxdepth) 9999
    grid .ec_tracer.filter.settings.mindepth - .ec_tracer.filter.settings.maxdepth - -sticky w
    incr row

    checkbutton .ec_tracer.filter.settings.spy -text "Predicate has a spy point" \
    	-variable tkecl(filter_spy) -onvalue spied -offvalue all
    grid .ec_tracer.filter.settings.spy -columnspan $cols -sticky w
    incr row

    label .ec_tracer.filter.settings.ports -text "Port is one of:"
    grid .ec_tracer.filter.settings.ports -columnspan $cols -sticky w
    incr row

    set tkecl(portlist) [lindex [lindex [ec_rpc_check "sepia_kernel:debug_port_names(_)"] 2] 1]
    foreach port $tkecl(portlist) {
	set tkecl(filter_port,$port) 1
	checkbutton .ec_tracer.filter.settings.port_$port -text $port -variable tkecl(filter_port,$port)
	grid .ec_tracer.filter.settings.port_$port -row $row -column $col -sticky w
	set col [expr ($col+1)%$cols]
	set row [expr $col?$row:$row+1]
    }

    button .ec_tracer.filter.go -text "Go" -command {set tkecl(tracercommand) filter}
    balloonhelp .ec_tracer.filter.go "Continue program execution until a matching port is encountered"
    button .ec_tracer.filter.close -text "Close" -command {destroy .ec_tracer.filter}
    pack .ec_tracer.filter.go .ec_tracer.filter.close -side left -expand 1 -fill x

    focus [.ec_tracer.filter.settings.mininvoc subwidget entry]
    return .ec_tracer.filter
}


#---------------------------------------------------------------
# Directory selection
#---------------------------------------------------------------
proc tkecl:get_newcwd {dir} {
    global tkecl


    set newdir [tk_getDirectory -initialdir [pwd] -title "Set Current Working Directory"]

    if {![string match "" $newdir]} {
	set tkecl($dir) $newdir
	cd $newdir
    }
}

proc tkecl:paths_menu {p name} {
    set menu [menu $p.m -tearoff 0 -postcommand "tkecl:build_path_menu $p.m $p $name"]
}

proc tkecl:build_path_menu {menu p name} {
    global tkecl

    $menu delete 0 end ;# get rid of old entries
    $menu add command -label "Add a new directory" -command \
	    "tkecl:add_new_path $name"
    $menu add separator

    set i 0
    foreach {item} $tkecl($name) {
	;# probably treat all spaces as breaks in name!
	$menu add command -label $item -command "tkecl:change_one_path $name $p $item $i"
	incr i
    }
}
	
proc tkecl:add_new_path {name} {
    global tkecl

    tkecl:gui_edit_one_path Insert $name [pwd] 0 

    if {[llength $tkecl($name)] != 0} {
	ec_rpc [list set_flag $name $tkecl($name)] {(()[S*])}
    }
}

proc tkecl:change_one_path {name p item i} {
    global tkecl

    set w $p.change

    if ![winfo exists $w] {
	set old [focus]
	set tkecl(path_to_change) [lindex [ec_rpc [list os_file_name $item _] \
		(S_)] 2]
	toplevel $w
	wm title $w "Change one path for $name"
	focus $w
	grab $w
	pack [entry $w.e -bg white -width 40 -textvariable tkecl(path_to_change) \
		-relief sunken] -side top -expand 1 -fill both
	bind $w.e <Return> "tkecl:perform_path_change Replace $name \
		$tkecl(path_to_change) $i"
	pack [button $w.replace -command "tkecl:gui_edit_one_path Replace $name\
		$item $i" -text Replace] -side left -expand 1 -fill both
	pack [button $w.delete -command "tkecl:perform_path_change Delete $name \
		$item $i" -text Delete] -side left -expand 1 -fill both
	pack [button $w.insert -command "tkecl:gui_edit_one_path Insert $name \
		$item $i" -text Insert] -side left -expand 1 -fill both
	pack [button $w.cancel -text Cancel -command "destroy $w; set tkecl($name) [list $tkecl($name)]"] -side left -expand 1 -fill both
    }
    tkwait variable tkecl($name)

    if {[llength $tkecl($name)] == 0} {
	ec_rpc [list set_flag $name $tkecl($name)] {(()[])}
    } else {
	ec_rpc [list set_flag $name $tkecl($name)] {(()[S*])}
    }
    grab release $w
    focus $old
    destroy $w

}

proc tkecl:gui_edit_one_path {action name path i} {
    global tkecl

    set path [lindex [ec_rpc [list os_file_name $path _] (S_) ] 2]
    set new [tk_getDirectory -initialdir $path -title "$action a path"]
    if ![string match "" $new] {
	set new [lindex [ec_rpc [list os_file_name _ $new] (_S) ] 1]
	tkecl:perform_path_change $action $name $new $i
    } else {
	set tkecl($name) $tkecl($name) ;# make sure that tkwait does get its `changes'
    }
}

proc tkecl:perform_path_change {action name new i} {
    global tkecl

    switch -exact -- $action {
	Replace {
	    set tkecl($name) [lreplace $tkecl($name) $i $i $new]
	}
	Insert {
	    set tkecl($name) [linsert $tkecl($name) $i $new]
	}
	Delete {
	    set tkecl($name) [lreplace $tkecl($name) $i $i]
	}
    }
}

#---------------------------------------------------------------
# Change Output mode
#---------------------------------------------------------------
proc tkecl:Set_global_output_mode {name} {
    global tkecl

    set popmode .gstate.omodes
    if [winfo exists $popmode] {
	tkinspect:RaiseWindow $popmode
	return
    }

    toplevel $popmode
    wm title $popmode "Set Global Output Modes"

    set return [ec_rpc_check \
       "tracer_tcl:get_output_modes(Modes,Status,Set,UnSet,TO,TD,TS)"]
    set newmodes [tkecl:Set_output_mode $popmode $return]
    set tkecl($name) [string trim $newmodes \"]
    ec_rpc_check "set_flag(output_mode,$newmodes)"
}


proc tkecl:Set_output_mode {popmode return} {
    global outputmodes

    bind $popmode <Enter> {focus %W}
    foreach {f modes status descr unsetd triopts tridesc tristatus} [lindex $return 2] {
	set i -1 
	foreach m $modes s $status d $descr u $unsetd { 
	    set l $m
	    if [string match "." $m] {set m period} ;# catch special chars. here
	    set outputmodes($popmode.l$m) $s
	    set outputmodes($popmode.l$m,set) $d
	    set outputmodes($popmode.l$m,unset) $u
	    incr i
	    grid [checkbutton $popmode.c$m -onvalue 1 -offvalue 0 -text $l\
		    -anchor w -variable outputmodes($popmode.l$m) -command "tkecl:Change_output_options $m $popmode.l$m"] \
		    -sticky news -row $i -column 0
	    if {$s == 1} {
		set label $d
	    } else {
		set label $u
	    }
	    grid [label $popmode.l$m -text $label] -sticky w -row $i -column 1
	    bind $popmode <Key-$m> {
		regexp {^(.+)\.[^\.]+$} %W null parent
		set lw $parent.l%K
		if {$outputmodes($lw) == 1} {
		    set outputmodes($lw) 0
		} else {
		    set outputmodes($lw) 1
		}
		tkecl:Change_output_options %K $lw
	    }

#	    balloonhelp $popmode.c$m $d
	} ;# foreach m ...

	set trinames ""
	foreach tri0 $triopts tdes0 $tridesc s $tristatus {
	    incr i
	    set f [frame $popmode.c$i]
	    set tri [lrange $tri0 1 end] ;# drop functor
	    set tdes [lrange $tdes0 1 end]
	    set name ""
	    append name [lindex $tri 0] [lindex $tri 1]
	    lappend trinames $name
	    set j 0
	    set outputmodes($popmode,t$name) $s
	    set outputmodes($popmode,t$name,s) $tri
	    set outputmodes($popmode,t$name,d) $tdes
	    foreach mode $tri d $tdes {
		grid [radiobutton $f.b$mode -variable outputmodes($popmode,t$name) \
			-text $mode -value $mode -anchor w\
		        -command "tkecl:Change_output_trioptions $mode $name \
			    $popmode.l$name $popmode"] -row 0 -column $j
		incr j
		if [string match $mode $s] {
		    grid [label $popmode.l$name -text $d] -sticky w -row $i -column 1
		}
	    }
	    grid $f -sticky news -row $i -column 0
	}
	grid [button $popmode.end -command "destroy $popmode" -text Set] \
	        -sticky news -row [expr $i + 1] -column 0 -columnspan 2
#	        -sticky news -row [expr ($i/3) + 1] -column 0 -columnspan 3
    }
    tkwait window $popmode
    set newmodes "\""
    foreach m $modes {
	set l $m
	if [string match "." $m] {set m period} ;# catch special chars. here
	if {$outputmodes($popmode.l$m) == 1} {
	    append newmodes $l
	}
    }
    foreach name $trinames { ;# add in the tristate modes
	if {![string match $outputmodes($popmode,t$name) off]} {
	    append newmodes $outputmodes($popmode,t$name)
	}
    }
    return [append newmodes \"]

}

# update label for the simple output options
proc tkecl:Change_output_options {mode w} {
    global outputmodes

    ;# called after mode has been changed to new value
    if {$outputmodes($w) == 1} {
	$w configure -text $outputmodes($w,set)
    } else {
	$w configure -text $outputmodes($w,unset)
    }
}

# update label for the tri-state options
proc tkecl:Change_output_trioptions {selected name label w} {
    global outputmodes

    foreach opt $outputmodes($w,t$name,s) d $outputmodes($w,t$name,d) { 
	;# find the one that matches selected
	if [string match $selected $opt] {
	    $label configure -text $d
	}
    }
}


#----------------------------------------------------------------------
# Compile note pad
#----------------------------------------------------------------------

proc tkecl:compile_pad {w} {
    if [winfo exists $w.cpad] {
	tkinspect:RaiseWindow $w.cpad
	return
    }
    set pad [toplevel $w.cpad]
    wm title $pad "Compile scratch-pad"
    text $pad.t -bg white -yscrollcommand "$pad.vscroll set" -xscrollcommand "$pad.hscroll set" 
    set bbar [frame $pad.bbar]
    pack $bbar -side bottom -fill x
       pack [button $bbar.com -text "Compile All" -command "tkecl:do_compile_all $pad.t"] -side left -expand 1 -fill x
       pack [button $bbar.sel -text "Compile Selection" -command "tkecl:do_compile_sel $pad.t"] -side left -expand 1 -fill x
       pack [button $bbar.end -text Close -command "destroy $pad"] -side left -expand 1 -fill x
    pack [scrollbar $pad.vscroll -command "$pad.t yview"] -side right -fill y
    pack [scrollbar $pad.hscroll -command "$pad.t xview" -orient horizontal] -side bottom -fill x
    pack $pad.t -expand 1 -fill both
    bind $pad <Alt-h> "tkecl:Get_helpfileinfo scra $pad"
    balloonhelp $bbar "Type in (short) ECLiPSe code for compilation. Can compile everything in window, or only selection."
    focus $pad.t

}

proc tkecl:do_compile_all {t} {
    ec_rpc_check [list ":" tracer_tcl \
	    [list compile_string [$t get 1.0 end]]] (()(S))
}

proc tkecl:do_compile_sel {t} {
    foreach {start end} [$t tag ranges sel] {
       ec_rpc_check [list ":" tracer_tcl \
	    [list compile_string [$t get $start $end]]] (()(S))
    }
}

#----------------------------------------------------------------------
# Statistics display
#----------------------------------------------------------------------
proc tkecl:handle_statistics {} {
    global tkecl

    tkecl:create_stat_window
    set data [lindex [lindex [ec_rpc tracer_tcl:report_stats($tkecl(stats_interval),_)] 2] 2] 
    tkecl:display_stat $data
}

proc tkecl:display_stat {data} {
    global tkecl

    ;# colours are in pairs: dark and light versions
    if ![winfo exists .ec_stats] {
	return
    }

    set colours [list #00d040 #00f090 #c00000 #f00000 #c0c000 #ffff00 \
	    #b000b0 #f000f0  #c07000 #ff9000 #50d0b0 #a0ffe0 #000090 #0000ff]
    set cindex 0
    set h 85   ;# these are for the pie charts
    set w 85
    foreach item  $data {
        switch -exact -- [lindex $item 0] {
	    times  {
		set user [lindex $item 1]
		set real [lindex $item 2]
		foreach {gctime ngc gccol gcratio} [lrange [lindex $item 3] 1 end] {
		    break
		}
		set tframe .ec_stats.times
		set textf $tframe.text
		set pie $tframe.pie
		if ![winfo exists $tframe] {
		    pack [frame $tframe] -side top
		    pack [canvas $pie -width [expr $w + 10] -height [expr $h + 10]] -side left
		    pack [frame $textf] -side right
		    pack [frame $textf.times -relief ridge -borderwidth 3] -side top -padx 2 -pady 2
		    grid [label $textf.times.a -text "total time" -width 15 -anchor e] -row 1 -column 0
		    grid [label $textf.times.b -text "gc time" -width 15 -anchor e] -row 1 -column 1
		    grid [label $textf.times.c -text "\% user" -width 10 -anchor e] -row 1 -column 2
		    grid [label $textf.times.user -width 15 -anchor e] -row 2 -column 0
		    grid [label $textf.times.gc -width 15 -anchor e] -row 2 -column 1 -padx 2 -pady 2
		    grid [label $textf.times.userf -width 10 -anchor e] -row 2 -column 2 -padx 2 -pady 2
		    grid [label $textf.times.label -text "User CPU Time"] -row 0 -column 0 -columnspan 2 -sticky news
		    pack [frame $textf.gc -relief ridge -borderwidth 3] -side bottom
		    grid [label $textf.gc.a -text "total collected" -width 16 -anchor e] -row 1 -column 0
		    grid [label $textf.gc.b -text "\# gc" -width 9 -anchor e] -row 1 -column 1
		    grid [label $textf.gc.c -text "% recovered" -width 15 -anchor e] -row 1 -column 2
		    grid [label $textf.gc.col -width 16 -anchor e] -row 2 -column 0
		    grid [label $textf.gc.ngc -width 9 -anchor e] -row 2 -column 1
		    grid [label $textf.gc.ratio -width 15 -anchor e] -row 2 -column 2
		    grid [label $textf.gc.label -text "Garbage Collection"] -row 0 -column 0 -columnspan 3 -sticky news
		    set tkecl(stat,times,user) 0
		    set tkecl(stat,times,real) 0
		    balloonhelp $pie "Portion of total time spent on garbage collection with respect to total user CPU time"
		    balloonhelp $textf.gc "Garbage collection statistics"
		    balloonhelp $textf.times "Timing statistics"
		}
		    
		$textf.times.user configure -text "$user"
		$textf.times.gc configure -text "[expr round($gctime*100)/100.0]"
		$textf.times.userf configure -text \
		    "[expr round( ($user - $tkecl(stat,times,user)) / \
		    ($real - $tkecl(stat,times,real)) * 10000) / 100.0]"
		set tkecl(stat,times,user) $user 
		set tkecl(stat,times,real) $real

		$textf.gc.ngc configure -text "$ngc"
		$textf.gc.ratio configure -text "[expr round($gcratio*100)/100.0]"
		$textf.gc.col configure -text "$gccol"
		$pie create oval 10 10 $h $w -fill white
		if {$ngc != 0} {
		    tkecl:draw_pieslice $pie $h $w 0 [expr $gctime / double($user)] blue
		}
	    }

	    memory {
		set total [lindex $item 2]
		set mname [lindex $item 1]
		set ref   [lindex $item 3] 
		set sstart 0
		set mframe .ec_stats.$mname
		set pie $mframe.pie
		set textf $mframe.text
		if ![winfo exists $mframe] {
		    pack [frame $mframe -relief sunken -borderwidth 2] -side top
		    pack [canvas $pie -width [expr $w + 20] -height [expr $h + 10]] -side left
		    pack [frame $textf] -side right
		    pack [frame $textf.headings] -side top -expand 1 -fill x
		    grid [label $textf.headings.main -text "$mname"] -row 0 -column 0 -columnspan 4 -sticky news
		    grid [label $textf.headings.a -text area -width 10 -anchor e] -row 1 -column 0 -sticky news
		    grid [label $textf.headings.b -text alloc -width 11 -anchor e] -row 1 -column 1 -sticky news
		    grid [label $textf.headings.c -text used -width 11 -anchor e] -row 1 -column 2 -sticky news
		    grid [label $textf.headings.d -text free -width 11 -anchor e] -row 1 -column 3 -sticky news

		    balloonhelp $textf "Memory statistics (in bytes) for the $mname memory area"
		    balloonhelp $pie "Proportion of memory used/allocated in the $mname area with respect to $ref"

		}
		$pie create oval 10 10 $h $w -fill white
		
		foreach component [lrange $item 4 end] {
		    switch -exact -- [lindex $component 0] {
			stack {
			    foreach {cname alloc used} [lrange $component 1 end] {
				break
			    }
			    set usedf [expr $used / double($total)]
			    set free [expr $alloc - $used]
			    set freef [expr $free / double($total)]
			    set dcol [lindex $colours $cindex]
			    incr cindex 1
			    set lcol [lindex $colours $cindex]
			    incr cindex 1

			    set cframe $textf.$cname
			    if ![winfo exists $cframe] {
				pack [frame $cframe] -side bottom -expand 1 -fill x
				grid [label $cframe.name -text $cname -width 10 -anchor e] -row 0 -column 0 -sticky news
				grid [label $cframe.alloc -width 11 -anchor e] -row 0 -column 1 -sticky news
				grid [label $cframe.used -foreground $lcol -width 11 -anchor e] -row 0 -column 2 -sticky news 
				grid [label $cframe.free -foreground $dcol -width 11 -anchor e] -row 0 -column 3 -sticky news
			    }
			    $cframe.alloc configure -text "$alloc"
			    $cframe.free configure -text  "$free"
			    $cframe.used configure -text  "$used"


			    set sstart [tkecl:draw_pieslice $pie $h $w \
				    $sstart $usedf $lcol]
			    set sstart [tkecl:draw_pieslice $pie $h $w \
				    $sstart $freef $dcol]
			    if {$cindex >= [llength $colours]} {
				set cindex 0
			    }
			}


		    }
		}
		$pie create arc 10 10 $h $w -start $sstart -extent \
			[expr 360-$sstart] -style pieslice
	    }
	}
    }
}

proc tkecl:create_stat_window {} {
    if {![winfo exists .ec_stats]} {
	toplevel .ec_stats
	wm title .ec_stats "ECLiPSe statistics"
	wm resizable .ec_stats 0 0
	pack [frame .ec_stats.buttons] -side bottom -expand 1 -fill x
	pack [button .ec_stats.buttons.change -command "tkecl:change_stat_interval" -text "Change interval"] -side left -expand 1 -fill x
	pack [button .ec_stats.buttons.close -command "tkecl:kill_stat_window" -text "Close"] -side right -expand 1 -fill x
	bind .ec_stats <Alt-h> "tkecl:Get_helpfileinfo stat .ec_stats"

	balloonhelp .ec_stats.buttons.change "Change the time interval with which the statistics are updated"
	balloonhelp .ec_stats.buttons.close "Close this window and quit monitoring statistics"
    } else {
	tkinspect:RaiseWindow .ec_stats
    }
}


proc tkecl:change_stat_interval {} {
    global tkecl

    set tkecl(stats_interval1) $tkecl(stats_interval)
    set w .ec_stats.interval
    if {![winfo exists $w]} {
	toplevel $w
	wm title $w "Statistics Reporting Interval"
	pack [frame $w.f] -side top
	pack [label $w.f.l -text "New reporting interval (sec.)"] -side left
	pack [entry $w.f.e -relief sunken -width 10 -textvariable tkecl(stats_interval1)] -side right -expand 1 -fill both
	pack [button $w.set -text "Set" -command "tkecl:set_stat_interval $w"] -side left -fill x -expand 1
	pack [button $w.cancel -text "Cancel" -command "destroy $w"] -side left -fill x -expand 1
	bind $w.f.e <Return> "tkecl:set_stat_interval $w"
	focus $w.f.e

	balloonhelp $w "Change time interval at which the statistics are \
		updated in the statistics window.\nType in a positive number \
		and click `Set' to change, or `Cancel' to not change"
    } else {
	tkinspect:RaiseWindow $w
	focus $w.f.e
    }
}

proc tkecl:set_stat_interval {w} {
    global tkecl

    if [regexp {^([0-9]+[.][0-9]+)|([0-9]+)$} $tkecl(stats_interval1)] {
	set tkecl(stats_interval) $tkecl(stats_interval1)
	ec_rpc tracer_tcl:change_report_interval($tkecl(stats_interval))
	destroy $w
    } else {
	set tkecl(stats_interval1) $tkecl(stats_interval)
	bell 
    }
}

proc tkecl:kill_stat_window {} {
    ec_rpc "tracer_tcl:stop_report_stats"
    destroy .ec_stats
}

proc tkecl:draw_pieslice {c h w start share colour} {
    set extent [expr $share * 360]
    $c create arc 10 10 $h $w -start $start -extent $extent \
	    -style pieslice -fill $colour
    return [expr $extent + $start]
}

proc tkecl:handle_stats_report {stream} {
    tkecl:display_stat [ec_read_exdr ec_queue$stream]
}

#----------------------------------------------------------------------
# Grace-style term matrix display
#----------------------------------------------------------------------
proc tkecl:handle_mat_flush {stream} {
    global tkecl_displayvals

    set commandline [ec_read_exdr ec_queue$stream]
    set command [lindex $commandline 0]
    ;#puts "line-$commandline"
    set name [lindex $commandline 1] ;# name is id

    if {![winfo exists .ec_matdisplay$name]} {
	if [info exists tkecl_displayvals($name,killed)] {
	    ;# been kill explicitly, do not redisplay
	    return
	}
	if {[string match setup $command]} { ;# initial setup
	    foreach {title row col} [lrange $commandline 2 end] {
		set tkecl_displayvals($name,title) $title
		tkecl:setup_disptable $name $title $row $col
	    }
	    return
	} else { ;# window killed at Tcl end, redisplay window
	    set row $tkecl_displayvals($name,row)
	    set col $tkecl_displayvals($name,col)
	    set title $tkecl_displayvals($name,title)
	    tkecl:setup_disptable $name $title $row $col
	    .ec_matdisplay$name.b.cont configure -state disabled
	    ;# falls through to execute actual command
	}
    }
		
    switch -exact -- $command {
	setup {
	    tk_messageBox -type ok -message "Display matrix protocol error: trying to initialise existing matrix"
	}

	disp {

	    foreach {row col new ground back} [lrange $commandline 2 end] {
		if {$tkecl_displayvals($name,$row,$col,stop) == 1}  {
		    append id r $row c $col
		    set tkecl_displayvals($name,$row,$col,prev) \
			   [lindex [.ec_matdisplay$name.$id config -text] end]
		    if {$tkecl_displayvals($name,update) == 0 && \
			    [string match nonground $ground]} {
			return
		    }
		    .ec_matdisplay$name.$id config -text $new
		    if [string match $back back] {
			;#set tkecl_displayvals($name,back) 1
			set tkecl_displayvals($name,back) [list $row $col]
			set colour pink
		    } else {
			;#set tkecl_displayvals($name,back) 0
			set colour beige
		    }
		    .ec_matdisplay$name.$id config -foreground black
		    .ec_matdisplay$name.$id config -background $colour
		    .ec_matdisplay$name.b.cont configure -state active
		    tkwait variable tkecl_displayvals($name,cont)
		    set tkecl_displayvals($name,back) [list 0 0]
		    if [winfo exists .ec_matdisplay$name] {
			.ec_matdisplay$name.$id config -background lightgray
			.ec_matdisplay$name.b.cont configure -state disabled
		    }
		} elseif {(($tkecl_displayvals($name,update) == 1) ||
		    ![string match nonground $ground])} {
			append id r $row c $col
			set tkecl_displayvals($name,$row,$col,prev) \
			   [lindex [.ec_matdisplay$name.$id config -text] end]
			.ec_matdisplay$name.$id config -text $new
		    if [string match $back back] {
			.ec_matdisplay$name.$id config -foreground red
		    } else {
			.ec_matdisplay$name.$id config -foreground black
		    }
 		}
	    }

	}

	interact {
	    .ec_matdisplay$name.b.cont configure -state active
	    tkwait variable tkecl_displayvals($name,cont)
	    if [winfo exists .ec_matdisplay$name] {
		.ec_matdisplay$name.b.cont configure -state disabled
	    }
	}

	kill { 
	    destroy .ec_matdisplay$name
	    set tkecl_displayvals($name,killed) 1
	}
	    
	    
    }
}

proc tkecl:setup_disptable {name title row col} {
    global tkecl_displayvals

    set tkecl_displayvals($name,cont) 0
    ;#set tkecl_displayvals($name,back) 0
    set tkecl_displayvals($name,back) [list 0 0]
    set parent [toplevel .ec_matdisplay$name]
    wm title $parent "Term display for $title"
    set tkecl_displayvals($name,row) $row
    set tkecl_displayvals($name,col) $col
    set tkecl_displayvals($name,update) 1
    bind $parent <Button-1> "tkecl:display_popup $parent %W $name $row %X %Y"
    for {set i 1} {$i <= $row} {incr i 1} {
	grid [label $parent.r$i -text $i -relief groove -width 5 -fg red -bg lightblue] -row $i -column 0 -sticky news
    }
    for {set i 1} {$i <= $col} {incr i 1} {
	grid [label $parent.c$i -text $i -relief groove -width 15 -fg red -bg lightblue] -row 0 -column $i -sticky news
    }
    for {set i 1} {$i <= $row} {incr i 1} {
	for {set j 1} {$j <= $col} {incr j 1} {
	    set id ""
	    append id r $i c $j
	    grid [label $parent.$id -text "-- unknown --" -relief ridge -width 15] -row $i -column $j -sticky news
	    set tkecl_displayvals($name,$i,$j,stop) 0
	}
    }
    grid [frame $parent.b] -row [expr $row + 1] -column 0 -columnspan [expr $col + 1] -sticky news
    pack [button $parent.b.cont -text "Continue" -command \
	    "set tkecl_displayvals($name,cont) 1"] -side left -fill x -expand 1
    bind $parent.b.cont <Destroy> "set tkecl_displayvals($name,cont) 1" ;# continues
    pack [button $parent.b.kill -text "Kill display" -command "ec_rpc kill_display_matrix($title)"] -side right -fill x
    pack [checkbutton $parent.b.update -text "Update on ground" -variable \
	    tkecl_displayvals($name,update) -onvalue 0 -offvalue 1] \
	    -side right -fill x
    pack [button $parent.b.stop -text "stop all" -command \
	    "tkecl:all_mat_break 1 $name $row $col"] -side right -fill x
    pack [button $parent.b.go -text "stop none" -command \
	    "tkecl:all_mat_break 0 $name $row $col"] -side right -fill x

    for {set j 1} {$j <= $col} {incr j 1} {
	grid columnconfigure $parent $j -weight 1
    }

    for {set j 1} {$j <= [expr $row]} {incr j 1} {
	grid rowconfigure $parent $j -weight 1
    }
    balloonhelp $parent "Monitor changes on terms: each matrix cell represents\
	     a term and show its current value.\n Left click on cell to get \
	     options. Current and previous (pre-update) values are shown.\n \
             On break, changes due to forward execution shown in yellow,\
	    changes due to backtracking shown in pink."
    balloonhelp $parent.b.cont "Click to continue execution until next break-point.\n (if set, a break-point occurs when a cell is updated)"
    balloonhelp $parent.b.stop "Set break-points on all cells"
    balloonhelp $parent.b.go "Unset break-points on all cells"
    balloonhelp $parent.b.update "Control update events -- if set, only update when cell becomes ground.\n Otherwise, updates depends on make_display_matrix"
    balloonhelp $parent.b.kill "Click to kill this display matrix -- program will continue to run without the display matrix"
    bind $parent <Alt-h> "tkecl:Get_helpfileinfo disp $parent"
}


 
proc tkecl:all_mat_break {state name row col} {
    global tkecl_displayvals

    for {set i 1} {$i <= $row} {incr i 1} {
	for {set j 1} {$j <= $col} {incr j 1} {
	    set tkecl_displayvals($name,$i,$j,stop) $state
	}
    }
}

proc tkecl:display_popup {p w name nrow x y} {
    global tkecl_displayvals


    if [string match disabled [lindex [$p.b.cont configure -state] end]] {return}
    set widgetinfo [grid info $w]
    foreach {option value} $widgetinfo {
	set widget($option) $value
    }
    if {(![info exists widget(-row)] || $widget(-row) == 0 || $widget(-column) == 0)} {
	return
    }
    if [winfo exists $p.popup] {
	destroy $p.popup
    }
    set m [menu $p.popup -tearoff 0]
    $m add command -label "current: [lindex [$w configure -text] end]" ;#-state disabled
    $m add command -label "previous: $tkecl_displayvals($name,$widget(-row),$widget(-column),prev)" -state disabled
    $m add check -label "Break on updates" -onvalue 1 -offvalue 0 \
	    -variable tkecl_displayvals($name,$widget(-row),$widget(-column),stop)
    ;#if {$tkecl_displayvals($name,back) == 0} 
    foreach {brow bcol} $tkecl_displayvals($name,back) {break}
    if {$brow != $widget(-row) || $bcol != $widget(-column)} {
	    $m add command -label "Inspect this term" -command \
		    "tkinspect:Inspect_term_init display($name,$widget(-row),$widget(-column))"
    }

#    $m add command -label "row: $widget(-row) col: $widget(-column)"
    tk_popup $m $x $y
}

#---------------------------------------------------------------------
# Balloon Help Toggle
#---------------------------------------------------------------------

trace variable tkecl(balloonhelp) w tkecl:ToggleBalloonHelp

proc tkecl:ToggleBalloonHelp {name dummy op} {
    global tkecl

    if {$tkecl(balloonhelp) == 1} {
	balloonhelp enable
    } else {
	balloonhelp disable
    }
}

#----------------------------------------------------------------------
# Help Files procedures
#----------------------------------------------------------------------

# find the right help file given key (four letter unique id) and display
# help file as a subwindow of parent
proc tkecl:Get_helpfileinfo {key parent} {
    global tkecl

    set i [lsearch -glob $tkecl(helpfiles) $key]
    if {$i == -1} {
	tk_messageBox -type ok -message "Invalid topic name for help files"
	return [list $key "invalid"]
    }
    set topic [lindex $tkecl(helpfiles) [expr $i+1]]
    set filename [lindex $tkecl(helpfiles) [expr $i+2]]
    eval tkinspect:helpinfo [concat {$parent} [list $topic $filename $key]]
}

#----------------------------------------------------------------------
# Routines for handling initial user defaults
#----------------------------------------------------------------------

proc tkecl:read_defaults_file {filename} {
    global env

    set defaults ""
    set file_exists 0
    if [file exists $filename] {
	set file_exists 1
    } else {
	set filename [file join $env(HOME) $filename]  ;# check in homedir
	if [file exists $filename] { set file_exists 1 }
    }
    if {$file_exists == 1} {
	set fid [open $filename r]
	while {[gets $fid line] >= 0} {
	    if {[llength $line] != 2} {
		tk_messageBox -type ok -message "Not two items on line:$line in file $filename"
	    } else {
		set defaults [concat  $defaults $line]
	    }
	}
    }
    return $defaults
}


proc tkecl:get_defaults {type} {
    global tcl_platform

    ;# read in user defined defaults for type (tkeclipserc or tkeclipsetoolsrc)
    switch $tcl_platform(platform) {
	unix {
	    return [tkecl:read_defaults_file .$type]
	}
	
	windows {
	    package require registry
	    set regpath HKEY_CURRENT_USER\\Software\\IC-Parc\\ECLiPSe\\$type
	    registry set $regpath  ;# make sure the key is there
	    set names [registry values $regpath]
            set defaults ""

	    foreach {dname} $names {
		lappend defaults $dname [registry get $regpath $dname]
	    }

	    return $defaults
	}
    }
}

proc tkecl:set_tools_defaults {toolsdefaults} {
    global tkecl tkinspectvalues

    set validoptions [list balloonhelp trace_refresh_dg stats_interval defaultextension editor dgf_tracedonly dgf_spiedonly dgf_wakeonly tracer_prdepth inspect_prdepth inspect_nosymbols]

    tkinspect:RaiseWindow .  
    ;# ensure inspector is loaded so that defaults here would not be overridden
    foreach {dname dvalue} $toolsdefaults {
	if {[lsearch $validoptions $dname] != -1} {
	   switch -- $dname {
	     trace_refresh_dg -
	     dgf_tracedonly   -
	     dgf_spiedonly    -
	     dgf_wakeonly     -
	     balloonhelp {
 	        ;# all 0 or 1 options
	        if {$dvalue == 1 || $dvalue == 0} {
		    set tkecl($dname) $dvalue
		} else {
		   tk_messageBox -icon warning -message "$dvalue is an invalid value for $dname (0/1 expected)" -type ok
		}
	     }
	     tracer_prdepth {
		if [regexp {^[0-9]+$} $dvalue size] {
		    set tkecl(tracer_prdepth) $dvalue
		    ec_rpc "tracer_tcl:set_tracer_print_depth($tkecl(tracer_prdepth))" 
		} else {
 		    tk_messageBox -icon warning -message "$dvalue is an invalid valid for tracer_prdepth (positive integer expected" -type ok
	        }
	     }

             stats_interval {
	        if [regexp {^([0-9]+[.][0-9]+)|([0-9]+)$} $dvalue] {
		    set tkecl($dname) $dvalue
		    ec_rpc tracer_tcl:change_report_interval($tkecl(stats_interval))

		} else {
		    tk_messageBox -icon warning -message "$dvalue is an invalid value for stats_interval (number expected)"
		}
		  
	     }

	     editor      -
	     defaultextension {
		set tkecl($dname) $dvalue
	     }

	     inspect_nosymbols   {
	        if {$dvalue == 1 || $dvalue == 0} {
		    set tkinspectvalues(symbols) $dvalue
		} else {
		   tk_messageBox -icon warning -message "$dvalue is an invalid value for $dname (0/1 expected)" -type ok
		}
	     }
	          
	     inspect_prdepth   {
		if [regexp {^[0-9]+$} $dvalue size] {
		    set tkinspectvalues(prdepth) $dvalue
		} else {
		    tk_messageBox -icon warning -message "$dvalue is an invalid valid for inspect_prdepth (positive integer expected" -type ok
	        }
            }
       }
    } else {
       ;# not a valid option
       tk_messageBox -icon warning -message "$dname is not a valid tools default option" -type ok
    }
  }
	
}


#----------------------------------------------------------------------
# Initalise and create menu/toolbar
#----------------------------------------------------------------------

proc ec_tools_init {w} {
    global tkinspectvalues tkecl


# Init the Eclipse part (must be done after ec_init !!!)

    ec_rpc "ensure_loaded(library(development_support))"
    ec_rpc "ensure_loaded(library(tracer_tcl))"
    ec_queue_connect debug_traceline r tkecl:handle_trace_line
    ec_queue_connect debug_output r tkecl:handle_debug_output
    ec_queue_connect matrix_out_queue r tkecl:handle_mat_flush
    ec_queue_connect gui_dg_info r tkecl:handle_dg_print
    ec_queue_connect statistics_out_queue r tkecl:handle_stats_report

    ec_rpc {open(queue(""),write,gui_help_output)}

    set tkecl(toplevel_module) [lindex [ec_rpc_check get_flag(toplevel_module,_)] 2]
    set tkecl(predpropmodule) $tkecl(toplevel_module)
    set tkinspectvalues(tkinspect_in) [ec_queue_connect tkinspect_in w]
    set tkinspectvalues(tkinspect_out) [ec_queue_connect tkinspect_out r]

# Create the tools launcher menu and set up help files

    menu $w 
    $w add command -label "Compile Scratch-pad" -command "tkecl:compile_pad $w"
lappend tkecl(helpfiles) scra {Compile Scratch-Pad} scratchhelp.txt
    $w add command -label "Source File Manager" -command tkecl:popup_file_window
    lappend tkecl(helpfiles) file {Source Files Tool} sourcehelp.txt
    $w add command -label "Predicate Browser" -command tkecl:popup_pred_prop
    lappend tkecl(helpfiles) pred {Predicates Property Tool} predprophelp.txt
    $w add command -label "Source Viewer" -command tkecl:display_source
    lappend tkecl(helpfiles) sour {Source Viewer} sourceviewhelp.txt
    $w add separator
    $w add command -label "Delayed Goals" -command tkecl:popup_dg_window
    lappend tkecl(helpfiles) dela {Delayed Goals Viewer} delayhelp.txt
    $w add command -label "Tracer" -command tkecl:popup_tracer
    lappend tkecl(helpfiles) trac Tracer tracerhelp.txt
    $w add command -label "Inspector" -command "tkinspect:Inspect_term_init current"
    lappend tkecl(helpfiles) insp Inspector inspecthelp.txt
    $w add separator
    $w add command -label "Global Settings" -command tkecl:popup_global_state
    lappend tkecl(helpfiles) glob {Global Settings Tool} globalsethelp.txt
    $w add command -label "Statistics" -command tkecl:handle_statistics
    lappend tkecl(helpfiles) stat {Statistics Window} stathelp.txt
    $w add command -label "Simple Query" -command tkecl:rpc
    lappend tkecl(helpfiles) rpc {Simple Query Tool} rpchelp.txt
    $w add command -label "ECLiPSe Help" -command tkecl:popup_help_window
    lappend tkecl(helpfiles) help {ECLiPSe Help Tool} helphelp.txt
    $w add separator
    $w add check -label "Balloon Help" -variable tkecl(balloonhelp)
#    $w add command -label "Test" -command tkecl:test

    lappend tkecl(helpfiles) disp {Display Matrix} matdisplayhelp.txt

    tkecl:set_tools_defaults [tkecl:get_defaults tkeclipsetoolsrc]
    ;# set user defined defaults for tools

    return $w
}






