#! /bin/sh
# \
	exec wish8.0 $0 ${1+"$@"}

#
# ECLiPSe Development Environment
#
# Copyright 1999 IC-Parc, Imperial College, London
#
# $Id: tkeclipse.tcl,v 1.30.2.1 1999/08/19 16:54:53 js10 Exp $
#

#----------------------------------------------------------------------
# Find and load the eclipse package
#----------------------------------------------------------------------
set tkecl(version) 4.2

switch $tcl_platform(platform) {
    unix {
	set tkecl(ECLIPSEDIR) $env(ECLIPSEDIR)
    }
    windows {
	package require registry
	set tkecl(ECLIPSEDIR) [registry get \
	    HKEY_LOCAL_MACHINE\\SOFTWARE\\IC-Parc\\Eclipse\\$tkecl(version) ECLIPSEDIR]
    }
    default {
	error "$tcl_platform(platform) not supported"
	exit
    }
}

lappend auto_path [file join $tkecl(ECLIPSEDIR) lib_tcl]


# Display a splash window (as soon as possible)

wm title . "ECLiPSe Toplevel"
wm iconname . ECLiPSe
set tkecl(ec_image) [image create photo -format ppm -file \
	[file join $tkecl(ECLIPSEDIR) lib_tcl Images eclipse_logo.ppm]]
pack [label .splash -image $tkecl(ec_image) -relief raised] -padx 5 -pady 5
update


#----------------------------------------------------------------------
# Load packages and initialise global settings
#----------------------------------------------------------------------

package require eclipse
package require eclipse_tools
package require AllWidgets

set tkecl(ec_state) Initialising
set tkecl(goal) {}
set tkecl(error_output_flag) 0

#----------------------------------------------------------------------
# GUI toplevel
#----------------------------------------------------------------------

proc tkecl:run_goal {} {
    global tkecl

    switch $tkecl(ec_state) {
	More -
	Yes	{
	    ec_queue_write toplevel_in [ec_tcl2exdr end ()]
	}
	Stopped -
	"Running..." {
	    return
	}
    }
    .query.goal_entry selection range 0 end
    lappend tkecl(history) $tkecl(goal)
    .query.goal_entry add $tkecl(goal)
    set tkecl(historypos) -1
    if [winfo exists .history] {
	.history.box insert 0 $tkecl(goal)
    }
    ec_queue_write toplevel_in [ec_tcl2exdr [list call $tkecl(goal)] (S)]
    tkecl:resume_query 
}

proc tkecl:more_goal {} {
    global tkecl

    if {$tkecl(ec_state) == "More"} {
	ec_queue_write toplevel_in [ec_tcl2exdr more ()]
	tkecl:resume_query 
    }
}


# we are about to call or backtrack into the main query
# and expect a result back on toplevel_out/answer_out

proc tkecl:resume_query {} {
    global tkecl

    set tkecl(ec_state) "Running..."
    set oldcursor [. cget -cursor]
    . configure -cursor watch
    .query.buttons.run configure -state disabled
    .query.buttons.make configure -state disabled
    .query.buttons.more configure -state disabled
    tkecl:activate_abort
    tkecl:remove_current_highlights
    if [winfo exists .ec_dg] {
	.ec_dg.text tag remove highlight 1.0 end
    }
    update

    # resume async (if this is supported) to keep the interface active
    if [catch {ec_resume 1} res] {
        .pane.stdio.tout insert end $res errorcolour
    }
 
    set tkecl(ec_state) [ec_read_exdr $tkecl(toplevel_out_channel)]
    if {$tkecl(ec_state) == "More"} {
	.query.buttons.more configure -state active
    }
    . configure -cursor $oldcursor
    .query.buttons.run configure -state active
    .query.buttons.make configure -state active
    tkecl:disable_abort
    if [winfo exists .ec_dg] {
	tkecl:refresh_dg
    }
    update
}

#----------------------------------------------------------------------
# History
#----------------------------------------------------------------------

proc tkecl:select_history {dir} {
    global tkecl

    set size [llength $tkecl(history)]
    if {$tkecl(historypos) == -1} {
	set tkecl(historypos) [expr $size - 1]
    }

    if [string match up $dir] {
	;# directions can only be up or down
	if {$tkecl(historypos) > 0} {
	    incr tkecl(historypos) -1
	    set tkecl(goal) [lindex $tkecl(history) $tkecl(historypos)]
	} else {
	    bell
	}
    } else { ;# move down
	if {$tkecl(historypos) < $size} {
	    incr tkecl(historypos) 1
	    set tkecl(goal) [lindex $tkecl(history) $tkecl(historypos)]
	} else {
	    bell
	}
    }
}

#----------------------------------------------------------------------
# Error notification
#----------------------------------------------------------------------

proc tkecl:error_to_window {Window stream} {
    global tkecl
    if !$tkecl(error_output_flag) {
	# make sure we keep seeing the first error
	$Window see end
	set tkecl(error_output_flag) 1
    }
    tkecl:output_to_window errorcolour $Window $stream
}

proc tkecl:output_to_window {Tag Window Stream} {
    global tkecl
    set data [ec_queue_read $Stream 1000]
    while {$data != ""} {
	$Window insert end $data $Tag
	set data [ec_queue_read $Stream 1000]
    }
    if !$tkecl(error_output_flag) {
	# while no errors, keep scrolling and show end of output
	$Window see end
    }
}

proc tkecl:CreateImage {name format} {
    global tkecl
    return [image create photo -format $format -file [file join $tkecl(ECLIPSEDIR) lib_tcl Images $name.$format]]
}

proc tkecl:Update_current_module {name dummy op} {
    global tkecl

    ec_rpc [list set_flag toplevel_module $tkecl(toplevel_module)] (()())
}

#----------------------------------------------------------------------
# About ECLiPSe
#----------------------------------------------------------------------

proc tkecl:About {} {
    global tkecl

    set w .tkecl_about

    if [winfo exists $w] {return}
    foreach {name date} [lrange [lindex [ec_rpc_check \
	    "sepia_kernel:sepia_version_banner(Name, Date)"] 2] 1 end] {
	toplevel $w
	wm title $w "About this Eclipse"
	wm resizable $w 0 0
	set t [frame $w.f] 
	pack [label $t.ic -image [tkecl:CreateImage iccrest gif]] -side left
	pack [label $t.icp -image [tkecl:CreateImage ic-parc gif]] -side top
#	pack [label $t.ec -image $tkecl(ec_image)] -side left
	pack [label $t.n -text "$name (Tcl/Tk GUI)"] -side top
	pack $t -side top -padx 10 -pady 10
	pack [label $w.blah -text {
ECLiPSe is not public domain software. To use it, you or your institution
or company is obliged to sign a licence agreement with IC-Parc,
Imperial College, London SW7 2AZ, Fax +44 171 5948432.
See http://www.icparc.ic.ac.uk/eclipse for details.
	}] -side top -expand 1 -fill x -padx 10 -pady 10
	pack [button $w.ok -text OK -command "destroy $w"] \
		-ipady 10 -padx 10 -pady 10 -side bottom -fill x -expand 1
    }
}

#----------------------------------------------------------------------
# Selecting a query's output
#----------------------------------------------------------------------

proc tkecl:Select_query_outputs {w other} {
    set left [$w mark previous current]
    set right [$w mark next current]

    while {![regexp -- {^q[0-9]+$} $left]} {
	if [string match "" $left] {
	    ;# got to left edge
	    set left 1.0
	    break
	}
	;# repeat until a qN mark found
	set left [$w mark previous $left]
    }

    while {![regexp -- {^q[0-9]+$} $right]} {
	if [string match "" $right] {
	    ;# got to right edge
	    set right end
	    break
	}
	;# repeat until a qN mark found
	set right [$w mark next $right]
    }

    set notsame 1
    foreach {oldl oldr} [$w tag ranges qsel] {
	if {($oldl == [$w index $left] && $oldr == [$w index $right])} {
	    set notsame 0
	} else {
	    set notsame 1
	}
    }
    $w tag remove qsel 1.0 end
    if {$notsame} {
	$w tag add qsel $left $right
	$w see "$right -1 lines"
    }

    $other tag remove qsel 1.0 end
    if {$notsame} {
	$other tag add qsel $left $right
	$other see "$right -1 lines"
    }
    return -code break
}

#triple click selects all earlier queries
proc tkecl:Select_earlier_queries {w other} {
    set right [$w mark next current]

    while {![regexp -- {^q[0-9]+$} $right]} {
	if [string match "" $right] {
	    ;# got to right edge
	    set right end
	    break
	}
	;# repeat until a qN mark found
	set right [$w mark next $right]
    }

    $w tag remove qsel 1.0 end
    $w tag add qsel 1.0 $right
    # no need to see right; should already be there because of double match

    $other tag remove qsel 1.0 end
    $other tag add qsel 1.0 $right

    return -code break
}

proc tkecl:toplevel_keypress {keycode} {

    switch -exact -- $keycode {
	"\x3" {  ;#^C -- allow default handling for window copy operation
	    return 0
	}
	"\x7f" -
	"\b" {  ;# delete and backspace
	    foreach t {.pane.stdio.tout .pane.answer.tout} {
		foreach {left right} [$t tag ranges qsel] {
		    $t delete $left $right
		}
	    }
	    return -code break
	}
	default {
	    return -code break
	}
    }
}

#----------------------------------------------------------------------
# Make the existing outputs in stdio and answer windows non-current 
#----------------------------------------------------------------------

proc tkecl:remove_current_highlights {} {
    global tkecl

    .pane.stdio.tout tag remove highlight 1.0 end
    .pane.stdio.tout tag remove errorcolour 1.0 end
    .pane.answer.tout tag remove highlight 1.0 end
    .pane.answer.tout tag remove errorcolour 1.0 end
    .pane.answer.tout tag remove successcolour 1.0 end
    .pane.stdio.tout mark set q$tkecl(nquery) "end -1 chars"
    .pane.stdio.tout mark gravity q$tkecl(nquery) left
    .pane.answer.tout mark set q$tkecl(nquery) "end -1 chars"
    .pane.answer.tout mark gravity q$tkecl(nquery) left
    incr tkecl(nquery) 1
    set tkecl(error_output_flag) 0
}

#----------------------------------------------------------------------
# Start of toplevel initialisation code
#----------------------------------------------------------------------
lappend tkecl(helpfiles) topl "TkECLiPSe Toplevel" toplevelhelp.txt

#----------------------------------------------------------------------
# Make the toplevel window
#----------------------------------------------------------------------

menu .mbar
. config -menu .mbar
.mbar add cascade -label "File" -menu .mbar.file -underline 0
menu .mbar.file
.mbar.file add command -label "Compile ..." -command {tkecl:remove_current_highlights; tkecl:compile_popup}
.mbar.file add command -label "Edit ..." -command tkecl:edit_popup
.mbar.file add command -label Make -command {tkecl:remove_current_highlights ;\
	ec_rpc "make,flush(output),flush(error)"}
.mbar.file add separator
.mbar.file add command -label Exit -command {ec_rpc "event(152)"; ec_cleanup; exit}

.mbar add cascade -label "Tools" -underline 0 -menu .mbar.windows

.mbar add cascade -label "Help" -menu .mbar.help -underline 0
menu .mbar.help
.mbar.help add command -label "About this ECLiPSe ..." -command tkecl:About
.mbar.help add separator
.mbar.help add check -label "Balloon Help" -variable tkecl(balloonhelp)
.mbar.help add separator

#----------------------------------------------------------------------
frame .query -relief groove -bd 3
#----------------------------------------------------------------------
label .query.label -text "Query Entry"
#label .query.module -textvariable tkecl(toplevel_module)
combobox .query.module -click single -listheight 6 -width 10 -editable 0 \
	-postcommand {tkecl:combo_add_modules .query.module} \
	-textvariable tkecl(toplevel_module)
label .query.colon -text ":"

trace variable tkecl(toplevel_module) w tkecl:Update_current_module
frame .query.buttons
button .query.buttons.make -text "make" -command \
	{tkecl:remove_current_highlights; ec_rpc "make,flush(output),flush(error)"}
button .query.buttons.run -text "run" -command tkecl:run_goal
button .query.buttons.more -text "more" -command tkecl:more_goal

frame .query.buttons.abort

#entry .query.goal_entry -bg white -width 80 -textvariable tkecl(goal)
combobox .query.goal_entry -click single -listheight 6 -bg white -width 65 \
	-textvariable tkecl(goal)
bind .query.goal_entry <Return> {tkecl:run_goal}
bind .query.goal_entry <Button-3> {tkecl:popup_history}
bind .query.goal_entry <Key-Up> {tkecl:select_history up}
bind .query.goal_entry <Key-Down> {tkecl:select_history down}

label .query.buttons.status -bg white -relief sunken -width 20 -textvariable tkecl(ec_state)

#----------------------------------------------------------------------
# Answer binding window and output/error window
# they are together in a frame and managed by the pane-manager
#----------------------------------------------------------------------
frame .pane -height 12c

frame .pane.answer -relief groove -bd 3
scrollbar .pane.answer.vscroll -command ".pane.answer.tout yview"
scrollbar .pane.answer.hscroll -command ".pane.answer.tout xview" -orient horizontal
#text .pane.answer.tout -bg white -height 15 -yscrollcommand ".pane.answer.vscroll set" -wrap none -xscrollcommand ".pane.answer.hscroll set"
text .pane.answer.tout -bg white -width 80 -yscrollcommand ".pane.answer.vscroll set" -wrap none -xscrollcommand ".pane.answer.hscroll set"
label .pane.answer.label -text "Results"
.pane.answer.tout tag configure highlight -foreground blue
.pane.answer.tout tag configure errorcolour -foreground red
.pane.answer.tout tag configure successcolour -foreground #00b000
.pane.answer.tout tag configure qsel -background lightblue
bind .pane.answer.tout <Any-Key> "tkecl:toplevel_keypress %A"
# allow ^C to work as copy in window
bind .pane.answer.tout <ButtonRelease-2> {break} 
bind .pane.answer.tout <Double-Button-1> "tkecl:Select_query_outputs .pane.answer.tout .pane.stdio.tout"
bind .pane.answer.tout <Triple-Button-1> "tkecl:Select_earlier_queries .pane.answer.tout .pane.stdio.tout"

#pack .pane.answer.vscroll -side left -fill y
#pack .pane.answer.label -side top -fill x
#pack .pane.answer.hscroll -side bottom -fill x
#pack .pane.answer.tout -side bottom -expand 1 -fill both

pack .pane.answer.label -side top -fill x
pack .pane.answer.vscroll -side left -fill y
pack .pane.answer.hscroll -side bottom -fill x
pack .pane.answer.tout -expand 1 -fill both


frame .pane.stdio -relief groove -bd 3
scrollbar .pane.stdio.vscroll -command ".pane.stdio.tout yview"
scrollbar .pane.stdio.hscroll -command ".pane.stdio.tout xview" -orient horizontal
text .pane.stdio.tout -width 80 -bg white -height 15 -yscrollcommand ".pane.stdio.vscroll set" -wrap none -xscrollcommand ".pane.stdio.hscroll set"
.pane.stdio.tout tag configure highlight -foreground blue
.pane.stdio.tout tag configure errorcolour -foreground red
.pane.stdio.tout tag configure nohandlercolour -foreground green
.pane.stdio.tout tag configure qsel -background lightblue
label .pane.stdio.label -text "Output and Error Messages"

pack .pane.stdio.label -side top -fill x
pack .pane.stdio.vscroll -side left -fill y
pack .pane.stdio.hscroll -side bottom -fill x
pack .pane.stdio.tout -expand 1 -fill both
bind .pane.stdio.tout <Any-Key> "tkecl:toplevel_keypress %A"
bind .pane.stdio.tout <ButtonRelease-2> {break}
bind .pane.stdio.tout <Double-Button-1> "tkecl:Select_query_outputs .pane.stdio.tout .pane.answer.tout"
bind .pane.stdio.tout <Triple-Button-1> "tkecl:Select_earlier_queries .pane.stdio.tout .pane.answer.tout"

pane .pane.answer .pane.stdio -orient vertical -initfrac [list 0.35 0.65]


#----------------------------------------------------------------------
# Pack the toplevel window
#----------------------------------------------------------------------
pack forget .splash
destroy .splash

#pack .mbar -side top -fill x
pack .query -side top -fill x
pack .pane -side top -fill both -expand 1
#pack .mbar.file .mbar.windows -side left

pack .query.buttons.run .query.buttons.more -side left -expand 1 -fill x
pack .query.buttons.status -side left -fill y
pack .query.buttons.make -side left -expand 1 -fill x
pack .query.buttons.abort -side left -expand 1 -fill x
pack .query.label -side top -fill x -expand 1
pack .query.buttons -side bottom -fill x -expand 1
pack .query.module -side left
pack .query.colon -side left
pack .query.goal_entry -side left -expand 1 -fill x
focus [.query.goal_entry subwidget entry]


#----------------------------------------------------------------------
# The abort button
#
# On Unix, the abort button is implemented using a separate process
# in order to allow aborts while eclipse is running. The button cannot
# simply be disabled since it is in the other process. Instead we
# temporarily replace it with the abort_dummy button.
#----------------------------------------------------------------------

switch $tcl_platform(platform) {
    unix {
	button .query.buttons.abort.abort_dummy -text "interrupt"
	.query.buttons.abort.abort_dummy configure -state disabled
	frame .query.buttons.abort.abort_frame -container true
	pack .query.buttons.abort.abort_frame -expand 1 -fill both
	update
	exec [info nameofexecutable] \
	    [file join $env(ECLIPSEDIR) lib_tcl tkabortbutton] \
	    -use [winfo id .query.buttons.abort.abort_frame] [pid] &

	proc tkecl:disable_abort {} {
	    pack forget .query.buttons.abort.abort_frame
	    pack .query.buttons.abort.abort_dummy -expand 1 -fill both
	}
	proc tkecl:activate_abort {} {
	    pack forget .query.buttons.abort.abort_dummy
	    pack .query.buttons.abort.abort_frame -expand 1 -fill both
	}
    }
    windows {
	button .query.buttons.abort.abort_button -text interrupt -command "ec_post_event int"
	pack .query.buttons.abort.abort_button -expand 1 -fill both

	proc tkecl:disable_abort {} {
	    .query.buttons.abort.abort_button configure -state disabled
	}
	proc tkecl:activate_abort {} {
	    .query.buttons.abort.abort_button configure -state active
	}
    }
}


proc tkecl:stop_request_handler {stream} {
    global tkecl
    set event [ec_read_exdr ec_queue$stream]
    if ![winfo exists .ec_stop_continue_box] {
        # We don't use a tk_messageBox or tk_dialog because they are modal.
	toplevel .ec_stop_continue_box
	wm title .ec_stop_continue_box "ECLiPSe interrupt"
	label .ec_stop_continue_box.msg -relief raised -height 3 -width 50 \
		-text "Execution interrupted - do you want to abort?"
	button .ec_stop_continue_box.abort -text "Yes, abort" \
	    -command {set tkecl(stop_continue) abort; destroy .ec_stop_continue_box}
	button .ec_stop_continue_box.cont -text "No, continue" \
	    -command {set tkecl(stop_continue) cont; destroy .ec_stop_continue_box}
	button .ec_stop_continue_box.creep -text "Continue in creep mode" \
	    -command {set tkecl(stop_continue) creep; destroy .ec_stop_continue_box}
	pack .ec_stop_continue_box.msg -side top -fill both -expand 1
	pack .ec_stop_continue_box.abort -side left -expand 1 -pady 3m -padx 3m
	pack .ec_stop_continue_box.cont -side left -expand 1 -pady 3m -padx 3m
	pack .ec_stop_continue_box.creep -side left -expand 1 -pady 3m -padx 3m

	switch [lindex [ec_rpc get_flag(debugging,_)] 2] {
	    nodebug { .ec_stop_continue_box.creep configure -state disabled }
	}

	tkwait variable tkecl(stop_continue)
	switch $tkecl(stop_continue) {
	    abort {ec_post_event abort}
	    creep {ec_rpc sepia_kernel:trace_mode(0,0)}
	}
    }
}

#------------------------------------------------------------------------
# default settings
#------------------------------------------------------------------------

proc tkecl:toplevel_init {} {

    tkecl:set_toplevel_defaults [tkecl:get_defaults tkeclipserc]

    ec_init

    ec_tools_init .mbar.windows

}

proc tkecl:set_toplevel_defaults {toplevdefaults} {
    global tclecl

    set validtoplev [list globalsize localsize]
    foreach {dname dvalue} $toplevdefaults {
	if {[lsearch $validtoplev $dname] != -1} {
	    switch -exact -- $dname {
		globalsize -
		localsize {
		    if [regexp {^[0-9]+$} $dvalue size] { 
			;# make sure it is a number!
			ec_set_option $dname [expr $dvalue * 1048576] ;# in megabytes 
		    } else {
		    tk_messageBox -icon warning -message "$dname parameter: $dvalue should be a number" -type ok
		    }
		}
		
		default {set tkecl($dname) $dvalue }
	    }
	} else { 
	    ;# not a valid option
	    tk_messageBox -icon warning -message "$dname is not a valid toplevel default option" -type ok
	}
    }

}


#----------------------------------------------------------------------
# Balloon Help
#----------------------------------------------------------------------

balloonhelp .query.goal_entry "Query entry - type query in here (terminating `.' optional). <Ret> or run to execute \n Up and down arrows moves through previous queries, <Tab> for query completion\n Left-click arrow on right-hand side for selecting previous queries (non-duplicated)\n Right-click popups history window (duplicated, most recent first)"
balloonhelp .query.buttons.status "Status after executing last query (Yes/More/No/Abort)"
balloonhelp .query.buttons.more "Click to try find more solutions"
balloonhelp .query.buttons.run "Click to start execution of query in query entry.\n Note: Execution will restart even if there are pending solutions to an executing query."
balloonhelp .query.buttons.make "Click to recompile modified files"
balloonhelp .query.buttons.abort "Click to interrupt executing query"
balloonhelp .query.module "Module in which the query will be executed\nClick arrow on right to select a different module"
balloonhelp .pane.answer.label "Results window - results (top-level bindings and status after execution) are displayed here.\n Results for the most recent query are in blue.\n\
Double left click to match a query's outputs, triple click to highlight all earlier queries' output."
balloonhelp .pane.stdio.label "Output and Error Messages from Eclipse are displayed here.\n Most recent outputs are in blue, error messages are in red.\n\
Double left click to match a query's outputs, triple click to highlight all earlier queries' output."
balloonhelp .pane.__h1 "Press and drag mouse button 1 to adjust Results and Output window sizes"
# bind . <Alt-h> "tkecl:Get_helpfileinfo topl {}" get help menu

#----------------------------------------------------------------------
# Initialise and start eclipse toplevel
#----------------------------------------------------------------------

#ec_set_option io 0
set tkecl(historypos) -1
set tkecl(nquery) 0

tkecl:toplevel_init ;# initialise and set defaults

foreach {key topic filename} $tkecl(helpfiles) {
    .mbar.help add command -label $topic -command "tkecl:Get_helpfileinfo $key {}"
}

ec_set_queue_handler output r "tkecl:output_to_window highlight .pane.stdio.tout"
ec_set_queue_handler error r "tkecl:error_to_window .pane.stdio.tout"
ec_queue_connect gui_interrupt_request r tkecl:stop_request_handler

# ensure_loaded rather than use_module: we don't want to import
ec_post_goal "ensure_loaded(library(toplevel))"
ec_resume
ec_set_queue_handler answer_out r "ec_stream_to_window highlight .pane.answer.tout"
set tkecl(toplevel_out_channel) [ec_queue_connect toplevel_out r]

if {![string match $tkecl(version) [lindex [ec_rpc_check "get_flag(version, V)"] 2]]} {
    tk_messageBox -icon warning -message "Version differences detected between Tcl and ECLiPSe codes" -type ok
}

ec_post_goal "gui_toplevel"
tkecl:resume_query 


